// [AsmJit]
// Complete x86/x64 JIT and Remote Assembler for C++.
//
// [License]
// Zlib - See LICENSE.md file in the package.

// [Export]
#define ASMJIT_EXPORTS

// [Guard]
#include "../build.h"
#if defined(ASMJIT_BUILD_X86) || defined(ASMJIT_BUILD_X64)

// [Dependencies]
#include "../x86/x86inst.h"

// [Api-Begin]
#include "../apibegin.h"

namespace asmjit {

// ============================================================================
// [Enums (Internal)]
// ============================================================================

//! \internal
enum {
  // REX/VEX.
  kX86InstTable_L__ = (0) << kX86InstOpCode_L_Shift,  // L is operand-based or unspecified.
  kX86InstTable_L_I = (0) << kX86InstOpCode_L_Shift,  // L is ignored (LIG).
  kX86InstTable_L_0 = (0) << kX86InstOpCode_L_Shift,  // L has to be zero.
  kX86InstTable_L_L = (1) << kX86InstOpCode_L_Shift,  // L has to be set.

  kX86InstTable_W__ = (0) << kX86InstOpCode_W_Shift,  // W is operand-based or unspecified.
  kX86InstTable_W_I = (0) << kX86InstOpCode_W_Shift,  // W is ignored (WIG).
  kX86InstTable_W_0 = (0) << kX86InstOpCode_W_Shift,  // W has to be zero.
  kX86InstTable_W_W = (1) << kX86InstOpCode_W_Shift,  // W has to be set.

  // EVEX.
  kX86InstTable_E__ = (0) << kX86InstOpCode_EW_Shift, // EVEX.W is operand-based or unspecified.
  kX86InstTable_E_I = (0) << kX86InstOpCode_EW_Shift, // EVEX.W is ignored (WIG).
  kX86InstTable_E_0 = (0) << kX86InstOpCode_EW_Shift, // EVEX.W has to be zero.
  kX86InstTable_E_1 = (1) << kX86InstOpCode_EW_Shift  // EVEX.W has to be set.
};

//! \internal
//!
//! Combined flags.
enum X86InstOpInternal {
  kX86InstOpI        = kX86InstOpImm,

  kX86InstOpL        = kX86InstOpLabel,
  kX86InstOpLImm     = kX86InstOpLabel | kX86InstOpImm,

  kX86InstOpGwb      = kX86InstOpGw    | kX86InstOpGb,
  kX86InstOpGqd      = kX86InstOpGq    | kX86InstOpGd,
  kX86InstOpGqdw     = kX86InstOpGq    | kX86InstOpGd | kX86InstOpGw,
  kX86InstOpGqdwb    = kX86InstOpGq    | kX86InstOpGd | kX86InstOpGw | kX86InstOpGb,

  kX86InstOpGbMem    = kX86InstOpGb    | kX86InstOpMem,
  kX86InstOpGwMem    = kX86InstOpGw    | kX86InstOpMem,
  kX86InstOpGdMem    = kX86InstOpGd    | kX86InstOpMem,
  kX86InstOpGqMem    = kX86InstOpGq    | kX86InstOpMem,
  kX86InstOpGwbMem   = kX86InstOpGwb   | kX86InstOpMem,
  kX86InstOpGqdMem   = kX86InstOpGqd   | kX86InstOpMem,
  kX86InstOpGqdwMem  = kX86InstOpGqdw  | kX86InstOpMem,
  kX86InstOpGqdwbMem = kX86InstOpGqdwb | kX86InstOpMem,

  kX86InstOpFpMem    = kX86InstOpFp    | kX86InstOpMem,
  kX86InstOpMmMem    = kX86InstOpMm    | kX86InstOpMem,
  kX86InstOpKMem     = kX86InstOpK     | kX86InstOpMem,
  kX86InstOpXmmMem   = kX86InstOpXmm   | kX86InstOpMem,
  kX86InstOpYmmMem   = kX86InstOpYmm   | kX86InstOpMem,
  kX86InstOpZmmMem   = kX86InstOpZmm   | kX86InstOpMem,

  kX86InstOpMmXmm    = kX86InstOpMm    | kX86InstOpXmm,
  kX86InstOpMmXmmMem = kX86InstOpMmXmm | kX86InstOpMem,

  kX86InstOpXy       = kX86InstOpXmm   | kX86InstOpYmm,
  kX86InstOpXyMem    = kX86InstOpXy    | kX86InstOpMem,

  kX86InstOpXyz      = kX86InstOpXy    | kX86InstOpZmm,
  kX86InstOpXyzMem   = kX86InstOpXyz   | kX86InstOpMem
};

//! \internal
//!
//! X86/X64 Instruction AVX-512 flags (combined).
ASMJIT_ENUM(X86InstFlagsInternal) {
  // FPU.
  kX86InstFlagMem2_4           = kX86InstFlagMem2        | kX86InstFlagMem4,
  kX86InstFlagMem2_4_8         = kX86InstFlagMem2_4      | kX86InstFlagMem8,
  kX86InstFlagMem4_8           = kX86InstFlagMem4        | kX86InstFlagMem8,
  kX86InstFlagMem4_8_10        = kX86InstFlagMem4_8      | kX86InstFlagMem10
};

// ============================================================================
// [Macros]
// ============================================================================

// Undefined. Used to distinguish (visually) between zero and unused data.
#define U 0

// Instruction opcodes.
#define O_000000(op, o, L, W, EvexW) (kX86InstOpCode_PP_00 | kX86InstOpCode_MM_00   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_000F00(op, o, L, W, EvexW) (kX86InstOpCode_PP_00 | kX86InstOpCode_MM_0F   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_000F01(op, o, L, W, EvexW) (kX86InstOpCode_PP_00 | kX86InstOpCode_MM_0F01 | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_000F0F(op, o, L, W, EvexW) (kX86InstOpCode_PP_00 | kX86InstOpCode_MM_0F   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_000F38(op, o, L, W, EvexW) (kX86InstOpCode_PP_00 | kX86InstOpCode_MM_0F38 | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_000F3A(op, o, L, W, EvexW) (kX86InstOpCode_PP_00 | kX86InstOpCode_MM_0F3A | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_660000(op, o, L, W, EvexW) (kX86InstOpCode_PP_66 | kX86InstOpCode_MM_00   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_660F00(op, o, L, W, EvexW) (kX86InstOpCode_PP_66 | kX86InstOpCode_MM_0F   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_660F38(op, o, L, W, EvexW) (kX86InstOpCode_PP_66 | kX86InstOpCode_MM_0F38 | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_660F3A(op, o, L, W, EvexW) (kX86InstOpCode_PP_66 | kX86InstOpCode_MM_0F3A | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_9B0000(op, o, L, W, EvexW) (kX86InstOpCode_PP_9B | kX86InstOpCode_MM_00   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_F20000(op, o, L, W, EvexW) (kX86InstOpCode_PP_F2 | kX86InstOpCode_MM_00   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_F20F00(op, o, L, W, EvexW) (kX86InstOpCode_PP_F2 | kX86InstOpCode_MM_0F   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_F20F38(op, o, L, W, EvexW) (kX86InstOpCode_PP_F2 | kX86InstOpCode_MM_0F38 | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_F20F3A(op, o, L, W, EvexW) (kX86InstOpCode_PP_F2 | kX86InstOpCode_MM_0F3A | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_F30000(op, o, L, W, EvexW) (kX86InstOpCode_PP_F3 | kX86InstOpCode_MM_00   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_F30F00(op, o, L, W, EvexW) (kX86InstOpCode_PP_F3 | kX86InstOpCode_MM_0F   | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_F30F38(op, o, L, W, EvexW) (kX86InstOpCode_PP_F3 | kX86InstOpCode_MM_0F38 | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_F30F3A(op, o, L, W, EvexW) (kX86InstOpCode_PP_F3 | kX86InstOpCode_MM_0F3A | (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)

#define O_00_M08(op, o, L, W, EvexW) (kX86InstOpCode_PP_00 | kX86InstOpCode_MM_01000| (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_00_M09(op, o, L, W, EvexW) (kX86InstOpCode_PP_00 | kX86InstOpCode_MM_01001| (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)
#define O_66_M03(op, o, L, W, EvexW) (kX86InstOpCode_PP_66 | kX86InstOpCode_MM_00011| (0x##op) | ((o) << kX86InstOpCode_O_Shift) | kX86InstTable_L_##L | kX86InstTable_W_##W | kX86InstTable_E_##EvexW)

#define O_00_X(op, o) (kX86InstOpCode_PP_00 | (0x##op) | ((o) << kX86InstOpCode_O_Shift))
#define O_9B_X(op, o) (kX86InstOpCode_PP_9B | (0x##op) | ((o) << kX86InstOpCode_O_Shift))

#define F(flag) kX86InstFlag##flag           // Instruction Base Flag(s) `F(...)`.
#define A(flag) kX86InstFlagAvx512##flag     // Instruction AVX-512 flag(s) `A(...)`.
#define O(flag) kX86InstOp##flag             // Instruction operand flag(s) `O(...)`.
#define EF(eflags) 0                         // Instruction EFLAGS `EF(OSZAPCDX)`.
#define Enc(iEnc) kX86InstEncoding##iEnc     // Instruction Encoding `Enc(...)`.

// Defines an X86/X64 instruction.
#define INST(id, name, encoding, opcode0, opcode1, instFlags, eflags, writeIndex, writeSize, op0, op1, op2, op3, op4) \
  { 0, kX86InstId##id##_ExtendedIndex, opcode0 }

// ============================================================================
// [asmjit::X86Inst]
// ============================================================================

// ${X86InstData:Begin}
// ------------------- Automatically generated, do not edit -------------------
#if !defined(ASMJIT_DISABLE_NAMES)
static const char _x86InstNameData[] =
  "\0" "adc\0" "adcx\0" "adox\0" "bextr\0" "blcfill\0" "blci\0" "blcic\0"
  "blcmsk\0" "blcs\0" "blsfill\0" "blsi\0" "blsic\0" "blsmsk\0" "blsr\0"
  "bsf\0" "bsr\0" "bswap\0" "bt\0" "btc\0" "btr\0" "bts\0" "bzhi\0" "call\0"
  "cbw\0" "cdq\0" "cdqe\0" "clc\0" "cld\0" "clflush\0" "clflushopt\0" "cmc\0"
  "cmova\0" "cmovae\0" "cmovc\0" "cmovg\0" "cmovge\0" "cmovl\0" "cmovle\0"
  "cmovna\0" "cmovnae\0" "cmovnc\0" "cmovng\0" "cmovnge\0" "cmovnl\0"
  "cmovnle\0" "cmovno\0" "cmovnp\0" "cmovns\0" "cmovnz\0" "cmovo\0" "cmovp\0"
  "cmovpe\0" "cmovpo\0" "cmovs\0" "cmovz\0" "cmp\0" "cmpxchg\0" "cmpxchg16b\0"
  "cmpxchg8b\0" "cpuid\0" "cqo\0" "crc32\0" "cvtpd2pi\0" "cvtpi2pd\0"
  "cvtpi2ps\0" "cvtps2pi\0" "cvttpd2pi\0" "cvttps2pi\0" "cwd\0" "cwde\0"
  "daa\0" "das\0" "enter\0" "f2xm1\0" "fabs\0" "faddp\0" "fbld\0" "fbstp\0"
  "fchs\0" "fclex\0" "fcmovb\0" "fcmovbe\0" "fcmove\0" "fcmovnb\0" "fcmovnbe\0"
  "fcmovne\0" "fcmovnu\0" "fcmovu\0" "fcom\0" "fcomi\0" "fcomip\0" "fcomp\0"
  "fcompp\0" "fcos\0" "fdecstp\0" "fdiv\0" "fdivp\0" "fdivr\0" "fdivrp\0"
  "femms\0" "ffree\0" "fiadd\0" "ficom\0" "ficomp\0" "fidiv\0" "fidivr\0"
  "fild\0" "fimul\0" "fincstp\0" "finit\0" "fist\0" "fistp\0" "fisttp\0"
  "fisub\0" "fisubr\0" "fld\0" "fld1\0" "fldcw\0" "fldenv\0" "fldl2e\0"
  "fldl2t\0" "fldlg2\0" "fldln2\0" "fldpi\0" "fldz\0" "fmulp\0" "fnclex\0"
  "fninit\0" "fnop\0" "fnsave\0" "fnstcw\0" "fnstenv\0" "fnstsw\0" "fpatan\0"
  "fprem\0" "fprem1\0" "fptan\0" "frndint\0" "frstor\0" "fsave\0" "fscale\0"
  "fsin\0" "fsincos\0" "fsqrt\0" "fst\0" "fstcw\0" "fstenv\0" "fstp\0"
  "fstsw\0" "fsubp\0" "fsubrp\0" "ftst\0" "fucom\0" "fucomi\0" "fucomip\0"
  "fucomp\0" "fucompp\0" "fwait\0" "fxam\0" "fxch\0" "fxrstor\0" "fxrstor64\0"
  "fxsave\0" "fxsave64\0" "fxtract\0" "fyl2x\0" "fyl2xp1\0" "inc\0" "insertq\0"
  "ja\0" "jae\0" "jb\0" "jbe\0" "jc\0" "je\0" "jecxz\0" "jg\0" "jge\0" "jl\0"
  "jle\0" "jmp\0" "jna\0" "jnae\0" "jnb\0" "jnbe\0" "jnc\0" "jne\0" "jng\0"
  "jnge\0" "jnl\0" "jnle\0" "jno\0" "jnp\0" "jns\0" "jnz\0" "jo\0" "jp\0"
  "jpe\0" "jpo\0" "js\0" "jz\0" "lahf\0" "lea\0" "leave\0" "lfence\0" "lzcnt\0"
  "mfence\0" "monitor\0" "mov_ptr\0" "movdq2q\0" "movnti\0" "movntq\0"
  "movntsd\0" "movntss\0" "movq2dq\0" "movsx\0" "movsxd\0" "movzx\0" "mulx\0"
  "mwait\0" "neg\0" "not\0" "pause\0" "pavgusb\0" "pdep\0" "pext\0" "pf2id\0"
  "pf2iw\0" "pfacc\0" "pfadd\0" "pfcmpeq\0" "pfcmpge\0" "pfcmpgt\0" "pfmax\0"
  "pfmin\0" "pfmul\0" "pfnacc\0" "pfpnacc\0" "pfrcp\0" "pfrcpit1\0"
  "pfrcpit2\0" "pfrsqit1\0" "pfrsqrt\0" "pfsub\0" "pfsubr\0" "pi2fd\0"
  "pi2fw\0" "pmulhrw\0" "pop\0" "popa\0" "popcnt\0" "popf\0" "prefetch\0"
  "prefetch3dnow\0" "prefetchw\0" "prefetchwt1\0" "pshufw\0" "pswapd\0"
  "push\0" "pusha\0" "pushf\0" "rcl\0" "rcr\0" "rdfsbase\0" "rdgsbase\0"
  "rdrand\0" "rdseed\0" "rdtsc\0" "rdtscp\0" "rep lods_b\0" "rep lods_d\0"
  "rep lods_q\0" "rep lods_w\0" "rep movs_b\0" "rep movs_d\0" "rep movs_q\0"
  "rep movs_w\0" "rep stos_b\0" "rep stos_d\0" "rep stos_q\0" "rep stos_w\0"
  "repe cmps_b\0" "repe cmps_d\0" "repe cmps_q\0" "repe cmps_w\0"
  "repe scas_b\0" "repe scas_d\0" "repe scas_q\0" "repe scas_w\0"
  "repne cmps_b\0" "repne cmps_d\0" "repne cmps_q\0" "repne cmps_w\0"
  "repne scas_b\0" "repne scas_d\0" "repne scas_q\0" "repne scas_w\0" "ret\0"
  "rol\0" "ror\0" "rorx\0" "sahf\0" "sal\0" "sar\0" "sarx\0" "sbb\0" "seta\0"
  "setae\0" "setb\0" "setbe\0" "setc\0" "sete\0" "setg\0" "setge\0" "setl\0"
  "setle\0" "setna\0" "setnae\0" "setnb\0" "setnbe\0" "setnc\0" "setne\0"
  "setng\0" "setnge\0" "setnl\0" "setnle\0" "setno\0" "setnp\0" "setns\0"
  "setnz\0" "seto\0" "setp\0" "setpe\0" "setpo\0" "sets\0" "setz\0" "sfence\0"
  "sha1msg1\0" "sha1msg2\0" "sha1nexte\0" "sha1rnds4\0" "sha256msg1\0"
  "sha256msg2\0" "sha256rnds2\0" "shl\0" "shlx\0" "shr\0" "shrd\0" "shrx\0"
  "stc\0" "t1mskc\0" "tzcnt\0" "tzmsk\0" "ud2\0" "vaddpd\0" "vaddps\0"
  "vaddsd\0" "vaddss\0" "vaddsubpd\0" "vaddsubps\0" "vaesdec\0" "vaesdeclast\0"
  "vaesenc\0" "vaesenclast\0" "vaesimc\0" "vaeskeygenassist\0" "vandnpd\0"
  "vandnps\0" "vandpd\0" "vandps\0" "vblendpd\0" "vblendps\0" "vblendvpd\0"
  "vblendvps\0" "vbroadcastf128\0" "vbroadcasti128\0" "vbroadcastsd\0"
  "vbroadcastss\0" "vcmppd\0" "vcmpps\0" "vcmpsd\0" "vcmpss\0" "vcomisd\0"
  "vcomiss\0" "vcvtdq2pd\0" "vcvtdq2ps\0" "vcvtpd2dq\0" "vcvtpd2ps\0"
  "vcvtph2ps\0" "vcvtps2dq\0" "vcvtps2pd\0" "vcvtps2ph\0" "vcvtsd2si\0"
  "vcvtsd2ss\0" "vcvtsi2sd\0" "vcvtsi2ss\0" "vcvtss2sd\0" "vcvtss2si\0"
  "vcvttpd2dq\0" "vcvttps2dq\0" "vcvttsd2si\0" "vcvttss2si\0" "vdivpd\0"
  "vdivps\0" "vdivsd\0" "vdivss\0" "vdppd\0" "vdpps\0" "vextractf128\0"
  "vextracti128\0" "vextractps\0" "vfmadd132pd\0" "vfmadd132ps\0"
  "vfmadd132sd\0" "vfmadd132ss\0" "vfmadd213pd\0" "vfmadd213ps\0"
  "vfmadd213sd\0" "vfmadd213ss\0" "vfmadd231pd\0" "vfmadd231ps\0"
  "vfmadd231sd\0" "vfmadd231ss\0" "vfmaddpd\0" "vfmaddps\0" "vfmaddsd\0"
  "vfmaddss\0" "vfmaddsub132pd\0" "vfmaddsub132ps\0" "vfmaddsub213pd\0"
  "vfmaddsub213ps\0" "vfmaddsub231pd\0" "vfmaddsub231ps\0" "vfmaddsubpd\0"
  "vfmaddsubps\0" "vfmsub132pd\0" "vfmsub132ps\0" "vfmsub132sd\0"
  "vfmsub132ss\0" "vfmsub213pd\0" "vfmsub213ps\0" "vfmsub213sd\0"
  "vfmsub213ss\0" "vfmsub231pd\0" "vfmsub231ps\0" "vfmsub231sd\0"
  "vfmsub231ss\0" "vfmsubadd132pd\0" "vfmsubadd132ps\0" "vfmsubadd213pd\0"
  "vfmsubadd213ps\0" "vfmsubadd231pd\0" "vfmsubadd231ps\0" "vfmsubaddpd\0"
  "vfmsubaddps\0" "vfmsubpd\0" "vfmsubps\0" "vfmsubsd\0" "vfmsubss\0"
  "vfnmadd132pd\0" "vfnmadd132ps\0" "vfnmadd132sd\0" "vfnmadd132ss\0"
  "vfnmadd213pd\0" "vfnmadd213ps\0" "vfnmadd213sd\0" "vfnmadd213ss\0"
  "vfnmadd231pd\0" "vfnmadd231ps\0" "vfnmadd231sd\0" "vfnmadd231ss\0"
  "vfnmaddpd\0" "vfnmaddps\0" "vfnmaddsd\0" "vfnmaddss\0" "vfnmsub132pd\0"
  "vfnmsub132ps\0" "vfnmsub132sd\0" "vfnmsub132ss\0" "vfnmsub213pd\0"
  "vfnmsub213ps\0" "vfnmsub213sd\0" "vfnmsub213ss\0" "vfnmsub231pd\0"
  "vfnmsub231ps\0" "vfnmsub231sd\0" "vfnmsub231ss\0" "vfnmsubpd\0"
  "vfnmsubps\0" "vfnmsubsd\0" "vfnmsubss\0" "vfrczpd\0" "vfrczps\0" "vfrczsd\0"
  "vfrczss\0" "vgatherdpd\0" "vgatherdps\0" "vgatherqpd\0" "vgatherqps\0"
  "vhaddpd\0" "vhaddps\0" "vhsubpd\0" "vhsubps\0" "vinsertf128\0"
  "vinserti128\0" "vinsertps\0" "vlddqu\0" "vldmxcsr\0" "vmaskmovdqu\0"
  "vmaskmovpd\0" "vmaskmovps\0" "vmaxpd\0" "vmaxps\0" "vmaxsd\0" "vmaxss\0"
  "vminpd\0" "vminps\0" "vminsd\0" "vminss\0" "vmovapd\0" "vmovaps\0" "vmovd\0"
  "vmovddup\0" "vmovdqa\0" "vmovdqu\0" "vmovhlps\0" "vmovhpd\0" "vmovhps\0"
  "vmovlhps\0" "vmovlpd\0" "vmovlps\0" "vmovmskpd\0" "vmovmskps\0" "vmovntdq\0"
  "vmovntdqa\0" "vmovntpd\0" "vmovntps\0" "vmovq\0" "vmovsd\0" "vmovshdup\0"
  "vmovsldup\0" "vmovss\0" "vmovupd\0" "vmovups\0" "vmpsadbw\0" "vmulpd\0"
  "vmulps\0" "vmulsd\0" "vmulss\0" "vorpd\0" "vorps\0" "vpabsb\0" "vpabsd\0"
  "vpabsw\0" "vpackssdw\0" "vpacksswb\0" "vpackusdw\0" "vpackuswb\0" "vpaddb\0"
  "vpaddd\0" "vpaddq\0" "vpaddsb\0" "vpaddsw\0" "vpaddusb\0" "vpaddusw\0"
  "vpaddw\0" "vpalignr\0" "vpand\0" "vpandn\0" "vpavgb\0" "vpavgw\0"
  "vpblendd\0" "vpblendvb\0" "vpblendw\0" "vpbroadcastb\0" "vpbroadcastd\0"
  "vpbroadcastq\0" "vpbroadcastw\0" "vpclmulqdq\0" "vpcmov\0" "vpcmpeqb\0"
  "vpcmpeqd\0" "vpcmpeqq\0" "vpcmpeqw\0" "vpcmpestri\0" "vpcmpestrm\0"
  "vpcmpgtb\0" "vpcmpgtd\0" "vpcmpgtq\0" "vpcmpgtw\0" "vpcmpistri\0"
  "vpcmpistrm\0" "vpcomb\0" "vpcomd\0" "vpcomq\0" "vpcomub\0" "vpcomud\0"
  "vpcomuq\0" "vpcomuw\0" "vpcomw\0" "vperm2f128\0" "vperm2i128\0" "vpermd\0"
  "vpermil2pd\0" "vpermil2ps\0" "vpermilpd\0" "vpermilps\0" "vpermpd\0"
  "vpermps\0" "vpermq\0" "vpextrb\0" "vpextrd\0" "vpextrq\0" "vpextrw\0"
  "vpgatherdd\0" "vpgatherdq\0" "vpgatherqd\0" "vpgatherqq\0" "vphaddbd\0"
  "vphaddbq\0" "vphaddbw\0" "vphaddd\0" "vphadddq\0" "vphaddsw\0" "vphaddubd\0"
  "vphaddubq\0" "vphaddubw\0" "vphaddudq\0" "vphadduwd\0" "vphadduwq\0"
  "vphaddw\0" "vphaddwd\0" "vphaddwq\0" "vphminposuw\0" "vphsubbw\0"
  "vphsubd\0" "vphsubdq\0" "vphsubsw\0" "vphsubw\0" "vphsubwd\0" "vpinsrb\0"
  "vpinsrd\0" "vpinsrq\0" "vpinsrw\0" "vpmacsdd\0" "vpmacsdqh\0" "vpmacsdql\0"
  "vpmacssdd\0" "vpmacssdqh\0" "vpmacssdql\0" "vpmacsswd\0" "vpmacssww\0"
  "vpmacswd\0" "vpmacsww\0" "vpmadcsswd\0" "vpmadcswd\0" "vpmaddubsw\0"
  "vpmaddwd\0" "vpmaskmovd\0" "vpmaskmovq\0" "vpmaxsb\0" "vpmaxsd\0"
  "vpmaxsw\0" "vpmaxub\0" "vpmaxud\0" "vpmaxuw\0" "vpminsb\0" "vpminsd\0"
  "vpminsw\0" "vpminub\0" "vpminud\0" "vpminuw\0" "vpmovmskb\0" "vpmovsxbd\0"
  "vpmovsxbq\0" "vpmovsxbw\0" "vpmovsxdq\0" "vpmovsxwd\0" "vpmovsxwq\0"
  "vpmovzxbd\0" "vpmovzxbq\0" "vpmovzxbw\0" "vpmovzxdq\0" "vpmovzxwd\0"
  "vpmovzxwq\0" "vpmuldq\0" "vpmulhrsw\0" "vpmulhuw\0" "vpmulhw\0" "vpmulld\0"
  "vpmullw\0" "vpmuludq\0" "vpor\0" "vpperm\0" "vprotb\0" "vprotd\0" "vprotq\0"
  "vprotw\0" "vpsadbw\0" "vpshab\0" "vpshad\0" "vpshaq\0" "vpshaw\0" "vpshlb\0"
  "vpshld\0" "vpshlq\0" "vpshlw\0" "vpshufb\0" "vpshufd\0" "vpshufhw\0"
  "vpshuflw\0" "vpsignb\0" "vpsignd\0" "vpsignw\0" "vpslld\0" "vpslldq\0"
  "vpsllq\0" "vpsllvd\0" "vpsllvq\0" "vpsllw\0" "vpsrad\0" "vpsravd\0"
  "vpsraw\0" "vpsrld\0" "vpsrldq\0" "vpsrlq\0" "vpsrlvd\0" "vpsrlvq\0"
  "vpsrlw\0" "vpsubb\0" "vpsubd\0" "vpsubq\0" "vpsubsb\0" "vpsubsw\0"
  "vpsubusb\0" "vpsubusw\0" "vpsubw\0" "vptest\0" "vpunpckhbw\0" "vpunpckhdq\0"
  "vpunpckhqdq\0" "vpunpckhwd\0" "vpunpcklbw\0" "vpunpckldq\0" "vpunpcklqdq\0"
  "vpunpcklwd\0" "vpxor\0" "vrcpps\0" "vrcpss\0" "vroundpd\0" "vroundps\0"
  "vroundsd\0" "vroundss\0" "vrsqrtps\0" "vrsqrtss\0" "vshufpd\0" "vshufps\0"
  "vsqrtpd\0" "vsqrtps\0" "vsqrtsd\0" "vsqrtss\0" "vstmxcsr\0" "vsubpd\0"
  "vsubps\0" "vsubsd\0" "vsubss\0" "vtestpd\0" "vtestps\0" "vucomisd\0"
  "vucomiss\0" "vunpckhpd\0" "vunpckhps\0" "vunpcklpd\0" "vunpcklps\0"
  "vxorpd\0" "vxorps\0" "vzeroall\0" "vzeroupper\0" "wrfsbase\0" "wrgsbase\0"
  "xadd\0" "xgetbv\0" "xsaveopt\0" "xsaveopt64\0" "xsetbv";
// ----------------------------------------------------------------------------

static const uint16_t _x86InstNameIndex[] = {
  0, 1, 5, 640, 3418, 3430, 3652, 3662, 3157, 3169, 10, 2391, 2399, 2411, 2419,
  2431, 2439, 1644, 4511, 2456, 2464, 2472, 2479, 15, 21, 29, 34, 40, 47, 2486,
  2495, 2504, 2514, 52, 60, 65, 71, 78, 83, 87, 91, 97, 100, 104, 108, 112,
  117, 122, 126, 130, 135, 139, 143, 151, 162, 166, 172, 497, 504, 179, 512,
  185, 191, 198, 204, 211, 218, 519, 527, 226, 536, 233, 240, 248, 255, 263,
  270, 277, 284, 291, 297, 303, 310, 317, 323, 329, 2580, 2587, 1902, 1915,
  1928, 1941, 2594, 2601, 333, 341, 352, 6267, 6276, 362, 368, 372, 2624, 2634,
  2644, 378, 2654, 387, 396, 2674, 2684, 405, 2704, 2714, 2724, 2734, 2744,
  2754, 2764, 414, 2775, 424, 2786, 2797, 434, 438, 443, 447, 2394, 659, 2808,
  2815, 2822, 2829, 2836, 2842, 627, 451, 2874, 4916, 457, 463, 1382, 468, 474,
  479, 485, 490, 496, 503, 511, 518, 526, 535, 543, 551, 558, 563, 569, 576,
  582, 589, 594, 602, 607, 613, 619, 626, 632, 638, 644, 650, 657, 663, 670,
  675, 681, 689, 695, 700, 706, 713, 719, 726, 730, 735, 741, 748, 755, 762,
  769, 776, 782, 1424, 787, 793, 800, 807, 812, 819, 826, 834, 841, 848, 854,
  861, 867, 875, 882, 888, 895, 900, 908, 914, 918, 924, 931, 936, 1486, 942,
  1492, 948, 955, 960, 966, 973, 981, 988, 996, 1002, 1007, 1012, 1020, 1030,
  1037, 1046, 1054, 1060, 3941, 3949, 3957, 3965, 658, 676, 1068, 3997, 1072,
  871, 1080, 1083, 1087, 1090, 1094, 1097, 1106, 1109, 1113, 1116, 1124, 1128,
  1133, 1137, 1142, 1146, 1150, 1154, 1159, 1163, 1168, 1172, 1176, 1180, 1184,
  1187, 1190, 1194, 1198, 1201, 1100, 1120, 1204, 4007, 4014, 1209, 1213, 1219,
  1672, 1683, 1694, 1705, 1226, 4023, 5362, 4057, 4064, 5381, 4078, 1232, 4085,
  4092, 5429, 4106, 1239, 4624, 1247, 4113, 4121, 505, 5355, 4135, 1255, 4144,
  4027, 4160, 4169, 4177, 4185, 4194, 4202, 4210, 4220, 4230, 4239, 1263, 4249,
  4258, 1270, 1277, 1285, 5366, 1293, 1716, 1727, 1738, 1749, 4273, 4280, 4290,
  4300, 1301, 1307, 4307, 4315, 1314, 4323, 677, 4332, 4339, 4346, 4353, 1320,
  1325, 1331, 808, 1335, 1017, 6325, 6332, 4372, 4379, 4386, 4393, 4403, 4413,
  4423, 4433, 4440, 4447, 4454, 4462, 4470, 4479, 4488, 4495, 4504, 4510, 1339,
  4517, 1345, 4524, 4540, 4550, 4611, 4629, 4638, 4647, 4656, 4665, 4676, 4687,
  4696, 4705, 4714, 4723, 4734, 1353, 1358, 4899, 4907, 4915, 4923, 1363, 1369,
  1375, 1381, 1387, 1395, 1403, 1411, 1417, 1423, 1429, 1436, 1444, 1450, 1459,
  1468, 1477, 1485, 1491, 5002, 5019, 5088, 5114, 5135, 5152, 5161, 1498, 1504,
  5178, 5186, 5194, 5202, 5330, 5341, 5372, 5380, 5388, 5396, 5404, 5412, 5420,
  5428, 5436, 5444, 5452, 5460, 5468, 5478, 5488, 5498, 5508, 5518, 5528, 5538,
  5548, 5558, 5568, 5578, 5588, 5598, 5606, 1510, 5616, 5625, 5633, 5641, 5649,
  1518, 1522, 1527, 1534, 5658, 1539, 1548, 1562, 1572, 4324, 5762, 5770, 5778,
  5787, 1584, 5796, 5804, 5812, 5820, 5827, 5835, 5858, 5865, 5880, 5887, 5894,
  5902, 5925, 5932, 5939, 5946, 5953, 5961, 5969, 5978, 5987, 1591, 5994, 6001,
  6012, 6023, 6035, 6046, 6057, 6068, 6080, 1598, 1603, 1609, 6091, 1615, 6097,
  6104, 1619, 1623, 1632, 1641, 1648, 1655, 1661, 1668, 1679, 1690, 1701, 1712,
  1723, 1734, 1745, 1756, 1767, 1778, 1789, 1800, 1812, 1824, 1836, 1848, 1860,
  1872, 1884, 1896, 1909, 1922, 1935, 1948, 1961, 1974, 1987, 2000, 2004, 2008,
  2012, 6111, 6120, 6129, 6138, 6147, 6156, 2017, 2022, 2026, 2030, 2035, 1954,
  1967, 1980, 1993, 2039, 2044, 2050, 2055, 2061, 2066, 2071, 2076, 2082, 2087,
  2093, 2099, 2106, 2112, 2119, 2125, 2131, 2137, 2144, 2150, 2157, 2163, 2169,
  2175, 2181, 2186, 2191, 2197, 2203, 2208, 2213, 2220, 2229, 2238, 2248, 2258,
  2269, 2280, 2292, 5742, 2296, 2301, 2305, 2310, 6165, 6173, 6181, 6148, 6197,
  6157, 2315, 4580, 6213, 1760, 1771, 1782, 1793, 715, 3160, 3172, 3848, 3858,
  2319, 5995, 2326, 2332, 6266, 6275, 2338, 6284, 6294, 6304, 6314, 2342, 2349,
  2356, 2363, 2370, 2380, 2390, 2398, 2410, 2418, 2430, 2438, 2455, 2463, 2471,
  2478, 2485, 2494, 2503, 2513, 2523, 2538, 2553, 2566, 2579, 2586, 2593, 2600,
  2607, 2615, 2623, 2633, 2643, 2653, 2663, 2673, 2683, 2693, 2703, 2713, 2723,
  2733, 2743, 2753, 2763, 2774, 2785, 2796, 2807, 2814, 2821, 2828, 2835, 2841,
  2847, 2860, 2873, 2884, 2896, 2908, 2920, 2932, 2944, 2956, 2968, 2980, 2992,
  3004, 3016, 3028, 3037, 3046, 3055, 3064, 3079, 3094, 3109, 3124, 3139, 3154,
  3166, 3178, 3190, 3202, 3214, 3226, 3238, 3250, 3262, 3274, 3286, 3298, 3310,
  3322, 3337, 3352, 3367, 3382, 3397, 3412, 3424, 3436, 3445, 3454, 3463, 3472,
  3485, 3498, 3511, 3524, 3537, 3550, 3563, 3576, 3589, 3602, 3615, 3628, 3638,
  3648, 3658, 3668, 3681, 3694, 3707, 3720, 3733, 3746, 3759, 3772, 3785, 3798,
  3811, 3824, 3834, 3844, 3854, 3864, 3872, 3880, 3888, 3896, 3907, 3918, 3929,
  3940, 3948, 3956, 3964, 3972, 3984, 3996, 4006, 4013, 4022, 4034, 4045, 4056,
  4063, 4070, 4077, 4084, 4091, 4098, 4105, 4112, 4120, 4128, 4134, 4143, 4151,
  4159, 4168, 4176, 4184, 4193, 4201, 4209, 4219, 4229, 4238, 4248, 4257, 4266,
  4272, 4279, 4289, 4299, 4306, 4314, 4322, 4331, 4338, 4345, 4352, 4359, 4365,
  4371, 4378, 4385, 4392, 4402, 4412, 4422, 4432, 4439, 4446, 4453, 4461, 4469,
  4478, 4487, 4494, 4503, 4509, 4516, 4523, 4530, 4539, 4549, 4558, 4571, 4584,
  4597, 4610, 4621, 4628, 4637, 4646, 4655, 4664, 4675, 4686, 4695, 4704, 4713,
  4722, 4733, 4744, 4751, 4758, 4765, 4773, 4781, 4789, 4797, 4804, 4815, 4826,
  4833, 4844, 4855, 4865, 4875, 4883, 4891, 4898, 4906, 4914, 4922, 4930, 4941,
  4952, 4963, 4974, 4983, 4992, 5001, 5009, 5018, 5027, 5037, 5047, 5057, 5067,
  5077, 5087, 5095, 5104, 5113, 5125, 5134, 5142, 5151, 5160, 5168, 5177, 5185,
  5193, 5201, 5209, 5218, 5228, 5238, 5248, 5259, 5270, 5280, 5290, 5299, 5308,
  5319, 5329, 5340, 5349, 5360, 5371, 5379, 5387, 5395, 5403, 5411, 5419, 5427,
  5435, 5443, 5451, 5459, 5467, 5477, 5487, 5497, 5507, 5517, 5527, 5537, 5547,
  5557, 5567, 5577, 5587, 5597, 5605, 5615, 5624, 5632, 5640, 5648, 5657, 5662,
  5669, 5676, 5683, 5690, 5697, 5705, 5712, 5719, 5726, 5733, 5740, 5747, 5754,
  5761, 5769, 5777, 5786, 5795, 5803, 5811, 5819, 5826, 5834, 5841, 5849, 5857,
  5864, 5871, 5879, 5886, 5893, 5901, 5908, 5916, 5924, 5931, 5938, 5945, 5952,
  5960, 5968, 5977, 5986, 5993, 6000, 6011, 6022, 6034, 6045, 6056, 6067, 6079,
  6090, 6096, 6103, 6110, 6119, 6128, 6137, 6146, 6155, 6164, 6172, 6180, 6188,
  6196, 6204, 6212, 6221, 6228, 6235, 6242, 6249, 6257, 6265, 6274, 6283, 6293,
  6303, 6313, 6323, 6330, 6337, 6346, 6357, 6366, 6375, 336, 6380, 6092, 6324,
  6331, 1013, 1021, 1031, 1038, 6387, 6396, 6407
};

enum X86InstAlphaIndex {
  kX86InstAlphaIndexFirst = 'a',
  kX86InstAlphaIndexLast = 'z',
  kX86InstAlphaIndexInvalid = 0xFFFF
};

static const uint16_t _x86InstAlphaIndex[26] = {
  kX86InstIdAdc,
  kX86InstIdBextr,
  kX86InstIdCall,
  kX86InstIdDaa,
  kX86InstIdEmms,
  kX86InstIdF2xm1,
  0xFFFF,
  kX86InstIdHaddpd,
  kX86InstIdIdiv,
  kX86InstIdJa,
  0xFFFF,
  kX86InstIdLahf,
  kX86InstIdMaskmovdqu,
  kX86InstIdNeg,
  kX86InstIdOr,
  kX86InstIdPabsb,
  0xFFFF,
  kX86InstIdRcl,
  kX86InstIdSahf,
  kX86InstIdT1mskc,
  kX86InstIdUcomisd,
  kX86InstIdVaddpd,
  kX86InstIdWrfsbase,
  kX86InstIdXadd,
  0xFFFF,
  0xFFFF
};
#endif // !ASMJIT_DISABLE_NAMES
// ----------------------------------------------------------------------------

// ------------------- Automatically generated, do not edit -------------------
const X86InstExtendedInfo _x86InstExtendedInfo[] = {
  { Enc(None)            , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(None)                            , U                    },
  { Enc(X86Arith)        , 0 , 0 , 0x20, 0x3F, 0, { O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                  }, F(RW)|F(Lock)                      , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x20, 0x20, 0, { O(Gqd)            , O(GqdMem)         , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86Arith)        , 0 , 0 , 0x00, 0x3F, 0, { O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                  }, F(RW)|F(Lock)                      , U                    },
  { Enc(SimdRm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x01, 0x01, 0, { O(Gqd)            , O(GqdMem)         , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdRm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                  }, F(WO)                              , U                    },
  { Enc(AvxRvm_OptW)     , 0 , 0 , 0x00, 0x3F, 0, { O(Gqd)            , O(Gqd)            , O(GqdMem)         , U                 , U                  }, F(RW)                              , U                    },
  { Enc(AvxRmv_OptW)     , 0 , 0 , 0x00, 0x3F, 0, { O(Gqd)            , O(GqdMem)         , O(Gqd)            , U                 , U                  }, F(RW)                              , U                    },
  { Enc(XopVm_OptW)      , 0 , 0 , 0x00, 0x3F, 0, { O(Gqd)            , O(GqdMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdRm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(AvxVm_OptW)      , 0 , 0 , 0x00, 0x3F, 0, { O(Gqd)            , O(GqdMem)         , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x00, 0x3F, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86BSwap)        , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , U                 , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86BTest)        , 0 , 0 , 0x00, 0x3B, 0, { O(GqdwMem)        , O(Gqdw)|O(Imm)    , U                 , U                 , U                  }, F(RO)                              , O_000F00(BA,4,_,_,_) },
  { Enc(X86BTest)        , 0 , 0 , 0x00, 0x3B, 0, { O(GqdwMem)        , O(Gqdw)|O(Imm)    , U                 , U                 , U                  }, F(RW)|F(Lock)                      , O_000F00(BA,7,_,_,_) },
  { Enc(X86BTest)        , 0 , 0 , 0x00, 0x3B, 0, { O(GqdwMem)        , O(Gqdw)|O(Imm)    , U                 , U                 , U                  }, F(RW)|F(Lock)                      , O_000F00(BA,6,_,_,_) },
  { Enc(X86BTest)        , 0 , 0 , 0x00, 0x3B, 0, { O(GqdwMem)        , O(Gqdw)|O(Imm)    , U                 , U                 , U                  }, F(RW)|F(Lock)                      , O_000F00(BA,5,_,_,_) },
  { Enc(X86Call)         , 0 , 0 , 0x00, 0x00, 0, { O(GqdMem)|O(LImm) , U                 , U                 , U                 , U                  }, F(RW)|F(Flow)|F(Volatile)          , O_000000(E8,U,_,_,_) },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x20, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)                        , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x40, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)                        , U                    },
  { Enc(X86M)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(RO)|F(Volatile)                  , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x20, 0x20, 0, { U                 , U                 , U                 , U                 , U                  }, F(None)                            , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x24, 0x00, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x20, 0x00, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x04, 0x00, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x07, 0x00, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x03, 0x00, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x01, 0x00, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x10, 0x00, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86RegRm)        , 0 , 0 , 0x02, 0x00, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86Arith)        , 0 , 0 , 0x00, 0x3F, 0, { O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                  }, F(RO)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x40, 0x3F, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Op_66H)       , 0 , 0 , 0x40, 0x3F, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86RmReg)        , 0 , 0 , 0x00, 0x3F, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Lock)|F(Special)           , U                    },
  { Enc(X86M)            , 0 , 0 , 0x00, 0x04, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(RW)|F(Lock)|F(Special)           , U                    },
  { Enc(SimdRm)          , 0 , 0 , 0x00, 0x3F, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(RO)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Crc)          , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , O(GqdwbMem)       , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdRm)          , 0 , 16, 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRm)          , 0 , 8 , 0x00, 0x00, 0, { O(Mm)             , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRm)          , 0 , 16, 0x00, 0x00, 0, { O(Xmm)            , O(MmMem)          , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRm)          , 0 , 8 , 0x00, 0x00, 0, { O(Xmm)            , O(MmMem)          , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRm_Q)        , 0 , 8 , 0x00, 0x00, 0, { O(Gqd)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRm)          , 0 , 4 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRm_Q)        , 0 , 8 , 0x00, 0x00, 0, { O(Xmm)            , O(GqdMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRm_Q)        , 0 , 4 , 0x00, 0x00, 0, { O(Xmm)            , O(GqdMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRm)          , 0 , 8 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x28, 0x3F, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86IncDec)       , 0 , 0 , 0x00, 0x1F, 0, { O(GqdwbMem)       , U                 , U                 , U                 , U                  }, F(RW)|F(Lock)                      , O_000000(48,U,_,_,_) },
  { Enc(X86Rm_B)         , 0 , 0 , 0x00, 0x3F, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)                        , U                    },
  { Enc(X86Enter)        , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)|F(Special)             , U                    },
  { Enc(SimdExtract)     , 0 , 8 , 0x00, 0x00, 0, { O(GqdMem)         , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdExtrq)       , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)|O(Imm)     , O(None)|O(Imm)    , U                 , U                  }, F(RW)                              , O_660F00(78,0,_,_,_) },
  { Enc(FpuOp)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuArith)        , 0 , 0 , 0x00, 0x00, 0, { O(FpMem)          , O(Fp)             , U                 , U                 , U                  }, F(Fp)|F(Mem4_8)                    , U                    },
  { Enc(FpuRDef)         , 0 , 0 , 0x00, 0x00, 0, { O(Fp)             , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(X86M)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuR)            , 0 , 0 , 0x20, 0x00, 0, { O(Fp)             , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuR)            , 0 , 0 , 0x24, 0x00, 0, { O(Fp)             , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuR)            , 0 , 0 , 0x04, 0x00, 0, { O(Fp)             , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuR)            , 0 , 0 , 0x10, 0x00, 0, { O(Fp)             , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuCom)          , 0 , 0 , 0x00, 0x00, 0, { O(Fp)|O(Mem)      , O(Fp)             , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuR)            , 0 , 0 , 0x00, 0x3F, 0, { O(Fp)             , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuR)            , 0 , 0 , 0x00, 0x00, 0, { O(Fp)             , U                 , U                 , U                 , U                  }, F(Fp)                              , U                    },
  { Enc(FpuM)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)|F(Mem2_4)                    , U                    },
  { Enc(FpuM)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)|F(Mem2_4_8)                  , O_000000(DF,5,_,_,_) },
  { Enc(FpuM)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)|F(Mem2_4_8)                  , O_000000(DF,7,_,_,_) },
  { Enc(FpuM)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)|F(Mem2_4_8)                  , O_000000(DD,1,_,_,_) },
  { Enc(FpuFldFst)       , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)|F(Mem4_8_10)                 , O_000000(DB,5,_,_,_) },
  { Enc(FpuStsw)         , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)                              , O_00_X(DFE0,U)       },
  { Enc(FpuFldFst)       , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)|F(Mem4_8)                    , U                    },
  { Enc(FpuFldFst)       , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)|F(Mem4_8_10)                 , O_000000(DB,7,_,_,_) },
  { Enc(FpuStsw)         , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Fp)                              , O_9B_X(DFE0,U)       },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Fp)|F(Volatile)                  , U                    },
  { Enc(X86Rm_B)         , 0 , 0 , 0x00, 0x3F, 0, { 0                 , 0                 , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86Imul)         , 0 , 0 , 0x00, 0x3F, 0, { 0                 , 0                 , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86IncDec)       , 0 , 0 , 0x00, 0x1F, 0, { O(GqdwbMem)       , U                 , U                 , U                 , U                  }, F(RW)|F(Lock)                      , O_000000(40,U,_,_,_) },
  { Enc(SimdInsertq)     , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(None)|O(Imm)    , O(None)|O(Imm)    , U                  }, F(RW)                              , O_F20F00(78,U,_,_,_) },
  { Enc(X86Int)          , 0 , 0 , 0x00, 0x80, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)                        , U                    },
  { Enc(X86Jcc)          , 0 , 0 , 0x24, 0x00, 0, { O(Label)          , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , U                    },
  { Enc(X86Jcc)          , 0 , 0 , 0x20, 0x00, 0, { O(Label)          , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , U                    },
  { Enc(X86Jcc)          , 0 , 0 , 0x04, 0x00, 0, { O(Label)          , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , U                    },
  { Enc(X86Jcc)          , 0 , 0 , 0x07, 0x00, 0, { O(Label)          , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , U                    },
  { Enc(X86Jcc)          , 0 , 0 , 0x03, 0x00, 0, { O(Label)          , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , U                    },
  { Enc(X86Jcc)          , 0 , 0 , 0x01, 0x00, 0, { O(Label)          , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , U                    },
  { Enc(X86Jcc)          , 0 , 0 , 0x10, 0x00, 0, { O(Label)          , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , U                    },
  { Enc(X86Jcc)          , 0 , 0 , 0x02, 0x00, 0, { O(Label)          , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , U                    },
  { Enc(X86Jecxz)        , 0 , 0 , 0x00, 0x00, 0, { O(Gqdw)           , O(Label)          , U                 , U                 , U                  }, F(Flow)|F(Volatile)|F(Special)     , U                    },
  { Enc(X86Jmp)          , 0 , 0 , 0x00, 0x00, 0, { O(Label)|O(Imm)   , U                 , U                 , U                 , U                  }, F(Flow)|F(Volatile)                , O_000000(E9,U,_,_,_) },
  { Enc(X86Op)           , 0 , 0 , 0x3E, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(SimdRm)          , 0 , 16, 0x00, 0x00, 0, { O(Xmm)            , O(Mem)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Lea)          , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , O(Mem)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)|F(Special)             , U                    },
  { Enc(X86Fence)        , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)                        , U                    },
  { Enc(X86Op)           , 0 , 1 , 0x40, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(WO)|F(Special)                   , U                    },
  { Enc(X86Op)           , 0 , 4 , 0x40, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(WO)|F(Special)                   , U                    },
  { Enc(X86Op)           , 0 , 8 , 0x40, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(WO)|F(Special)                   , U                    },
  { Enc(X86Op_66H)       , 0 , 2 , 0x40, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(WO)|F(Special)                   , U                    },
  { Enc(SimdRm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(SimdRm)          , 0 , 0 , 0x00, 0x00, 0, { O(Mm)             , O(Mm)             , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86Fence)        , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RO)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Mov)          , 0 , 0 , 0x00, 0x00, 0, { O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86MovPtr)       , 0 , 0 , 0x00, 0x00, 0, { O(Gqdwb)          , O(Imm)            , U                 , U                 , U                  }, F(WO)|F(Special)                   , O_000000(A2,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , O_660F00(29,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , O_000F00(29,U,_,_,_) },
  { Enc(SimdMovBe)       , 0 , 0 , 0x00, 0x00, 0, { O(GqdwMem)        , O(GqdwMem)        , U                 , U                 , U                  }, F(WO)                              , O_000F38(F1,U,_,_,_) },
  { Enc(SimdMovD)        , 0 , 16, 0x00, 0x00, 0, { O(Gd)|O(MmXmmMem) , O(Gd)|O(MmXmmMem) , U                 , U                 , U                  }, F(WO)                              , O_000F00(7E,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdMov)         , 0 , 8 , 0x00, 0x00, 0, { O(Mm)             , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , O_660F00(7F,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , O_F30F00(7F,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 8 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdMov)         , 8 , 8 , 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(RW)                              , O_660F00(17,U,_,_,_) },
  { Enc(SimdMov)         , 8 , 8 , 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(RW)                              , O_000F00(17,U,_,_,_) },
  { Enc(SimdMov)         , 8 , 8 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdMov)         , 0 , 8 , 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , O_660F00(13,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 8 , 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , O_000F00(13,U,_,_,_) },
  { Enc(SimdMovNoRexW)   , 0 , 8 , 0x00, 0x00, 0, { O(Gqd)            , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(Mem)            , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , O_660F00(E7,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(Xmm)            , O(Mem)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdMov)         , 0 , 8 , 0x00, 0x00, 0, { O(Mem)            , O(Gqd)            , U                 , U                 , U                  }, F(WO)                              , O_000F00(C3,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(Mem)            , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , O_660F00(2B,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(Mem)            , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , O_000F00(2B,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 8 , 0x00, 0x00, 0, { O(Mem)            , O(Mm)             , U                 , U                 , U                  }, F(WO)                              , O_000F00(E7,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 8 , 0x00, 0x00, 0, { O(Mem)            , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , O_F20F00(2B,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 4 , 0x00, 0x00, 0, { O(Mem)            , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , O_F30F00(2B,U,_,_,_) },
  { Enc(SimdMovQ)        , 0 , 16, 0x00, 0x00, 0, { O(Gq)|O(MmXmmMem) , O(Gq)|O(MmXmmMem) , U                 , U                 , U                  }, F(WO)                              , O_000F00(7E,U,_,W,_) },
  { Enc(SimdRm)          , 0 , 16, 0x00, 0x00, 0, { O(Xmm)            , O(Mm)             , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(WO)|F(Special)                   , U                    },
  { Enc(X86Op_66H)       , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(WO)|F(Special)                   , U                    },
  { Enc(SimdMov)         , 0 , 8 , 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)|F(ZeroIfMem)                 , O_F20F00(11,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 4 , 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)|F(ZeroIfMem)                 , O_F30F00(11,U,_,_,_) },
  { Enc(X86MovsxMovzx)   , 0 , 0 , 0x00, 0x00, 0, { O(Gqdw)           , O(GwbMem)         , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Movsxd)       , 0 , 0 , 0x00, 0x00, 0, { O(Gq)             , O(GdMem)          , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , O_660F00(11,U,_,_,_) },
  { Enc(SimdMov)         , 0 , 16, 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)                              , O_000F00(11,U,_,_,_) },
  { Enc(AvxRvm_OptW)     , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , O(Gqd)            , O(GqdMem)         , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86Rm_B)         , 0 , 0 , 0x00, 0x3F, 0, { O(GqdwbMem)       , U                 , U                 , U                 , U                  }, F(RW)|F(Lock)                      , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(None)                            , U                    },
  { Enc(X86Rm_B)         , 0 , 0 , 0x00, 0x00, 0, { O(GqdwbMem)       , U                 , U                 , U                 , U                  }, F(RW)|F(Lock)                      , U                    },
  { Enc(SimdRm_P)        , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdRmi_P)       , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)       , O(Imm)            , U                 , U                  }, F(RW)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(Simd3dNow)       , 0 , 0 , 0x00, 0x00, 0, { O(Mm)             , O(MmMem)          , U                 , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdRmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                  }, F(WO)|F(Special)                   , U                    },
  { Enc(AvxRvm_OptW)     , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , O(Gqd)            , O(GqdMem)         , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdExtract)     , 0 , 8 , 0x00, 0x00, 0, { O(Gd)|O(Gb)|O(Mem), O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdExtract)     , 0 , 8 , 0x00, 0x00, 0, { O(GdMem)          , O(Xmm)            , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdPextrw)      , 0 , 8 , 0x00, 0x00, 0, { O(GdMem)          , O(MmXmm)          , U                 , U                 , U                  }, F(WO)                              , O_000F3A(15,U,_,_,_) },
  { Enc(Simd3dNow)       , 0 , 8 , 0x00, 0x00, 0, { O(Mm)             , O(MmMem)          , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(GdMem)          , O(Imm)            , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdRmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(GqMem)          , O(Imm)            , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdRmi_P)       , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(GdMem)          , O(Imm)            , U                 , U                  }, F(RW)                              , U                    },
  { Enc(SimdRm_PQ)       , 0 , 8 , 0x00, 0x00, 0, { O(Gqd)            , O(MmXmm)          , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Pop)          , 0 , 0 , 0x00, 0x00, 0, { 0                 , U                 , U                 , U                 , U                  }, F(WO)|F(Volatile)|F(Special)       , O_000000(58,U,_,_,_) },
  { Enc(X86RegRm)        , 0 , 0 , 0x00, 0x3F, 0, { O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0xFF, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)|F(Special)             , U                    },
  { Enc(X86Prefetch)     , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , O(Imm)            , U                 , U                 , U                  }, F(RO)|F(Volatile)                  , U                    },
  { Enc(X86M)            , 0 , 0 , 0x00, 0x3F, 0, { O(Mem)            , O(Imm)            , U                 , U                 , U                  }, F(RO)|F(Volatile)                  , U                    },
  { Enc(SimdRmi)         , 0 , 16, 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRmi_P)       , 0 , 8 , 0x00, 0x00, 0, { O(Mm)             , O(MmMem)          , O(Imm)            , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRmRi_P)      , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                  }, F(RW)                              , O_000F00(72,6,_,_,_) },
  { Enc(SimdRmRi)        , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Imm)            , U                 , U                 , U                  }, F(RW)                              , O_660F00(73,7,_,_,_) },
  { Enc(SimdRmRi_P)      , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                  }, F(RW)                              , O_000F00(73,6,_,_,_) },
  { Enc(SimdRmRi_P)      , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                  }, F(RW)                              , O_000F00(71,6,_,_,_) },
  { Enc(SimdRmRi_P)      , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                  }, F(RW)                              , O_000F00(72,4,_,_,_) },
  { Enc(SimdRmRi_P)      , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                  }, F(RW)                              , O_000F00(71,4,_,_,_) },
  { Enc(SimdRmRi_P)      , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                  }, F(RW)                              , O_000F00(72,2,_,_,_) },
  { Enc(SimdRmRi)        , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Imm)            , U                 , U                 , U                  }, F(RW)                              , O_660F00(73,3,_,_,_) },
  { Enc(SimdRmRi_P)      , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                  }, F(RW)                              , O_000F00(73,2,_,_,_) },
  { Enc(SimdRmRi_P)      , 0 , 0 , 0x00, 0x00, 0, { O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                  }, F(RW)                              , O_000F00(71,2,_,_,_) },
  { Enc(X86Push)         , 0 , 0 , 0x00, 0x00, 0, { 0                 , U                 , U                 , U                 , U                  }, F(RO)|F(Volatile)|F(Special)       , O_000000(50,U,_,_,_) },
  { Enc(X86Op)           , 0 , 0 , 0xFF, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)|F(Special)             , U                    },
  { Enc(X86Rot)          , 0 , 0 , 0x20, 0x21, 0, { O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86Rm)           , 0 , 8 , 0x00, 0x00, 0, { O(Gqd)            , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Rm)           , 0 , 8 , 0x00, 0x3F, 0, { O(Gqdw)           , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(WO)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Rep)          , 0 , 0 , 0x40, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Rep)          , 0 , 0 , 0x40, 0x00, 0, { O(Mem)            , O(Mem)            , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Rep)          , 0 , 0 , 0x40, 0x3F, 0, { O(Mem)            , O(Mem)            , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Ret)          , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Rot)          , 0 , 0 , 0x00, 0x21, 0, { O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(AvxRmi_OptW)     , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , O(GqdMem)         , O(Imm)            , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRmi)         , 0 , 8 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                  }, F(WO)                              , U                    },
  { Enc(SimdRmi)         , 0 , 4 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x3E, 0, { U                 , U                 , U                 , U                 , U                  }, F(RO)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Rot)          , 0 , 0 , 0x00, 0x3F, 0, { O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(AvxRmv_OptW)     , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , O(GqdMem)         , O(Gqd)            , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Set)          , 0 , 1 , 0x24, 0x00, 0, { O(GbMem)          , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Set)          , 0 , 1 , 0x20, 0x00, 0, { O(GbMem)          , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Set)          , 0 , 1 , 0x04, 0x00, 0, { O(GbMem)          , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Set)          , 0 , 1 , 0x07, 0x00, 0, { O(GbMem)          , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Set)          , 0 , 1 , 0x03, 0x00, 0, { O(GbMem)          , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Set)          , 0 , 1 , 0x01, 0x00, 0, { O(GbMem)          , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Set)          , 0 , 1 , 0x10, 0x00, 0, { O(GbMem)          , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86Set)          , 0 , 1 , 0x02, 0x00, 0, { O(GbMem)          , U                 , U                 , U                 , U                  }, F(WO)                              , U                    },
  { Enc(X86ShldShrd)     , 0 , 0 , 0x00, 0x3F, 0, { O(GqdwbMem)       , O(Gb)             , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86ShldShrd)     , 0 , 0 , 0x00, 0x3F, 0, { O(GqdwbMem)       , O(Gqdwb)          , U                 , U                 , U                  }, F(RW)|F(Special)                   , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x20, 0, { U                 , U                 , U                 , U                 , U                  }, F(None)                            , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x00, 0x40, 0, { U                 , U                 , U                 , U                 , U                  }, F(None)                            , U                    },
  { Enc(X86M)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Volatile)                        , U                    },
  { Enc(X86Op)           , 0 , 0 , 0x40, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Op_66H)       , 0 , 0 , 0x40, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(RW)|F(Volatile)|F(Special)       , U                    },
  { Enc(X86Test)         , 0 , 0 , 0x00, 0x3F, 0, { O(GqdwbMem)       , O(Gqdwb)|O(Imm)   , U                 , U                 , U                  }, F(RO)                              , O_000000(F6,U,_,_,_) },
  { Enc(AvxRvm_OptL)     , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm)           , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRmi)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmi_OptL)    , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmr_OptL)    , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Xy)             , O(XyMem)          , O(Xy)             , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm)           , 0 , 0 , 0x00, 0x00, 0, { O(Ymm)            , O(Mem)            , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm)           , 0 , 0 , 0x00, 0x00, 0, { O(Ymm)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm_OptL)      , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XmmMem)         , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm)           , 0 , 0 , 0x00, 0x3F, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(RO)|F(Avx)                       , U                    },
  { Enc(AvxRm_OptL)      , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm)           , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxMri_OptL)     , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(Xy)             , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm)           , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(GqdMem)         , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm_OptL)      , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxMri)          , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(Ymm)            , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxMri)          , 0 , 0 , 0x00, 0x00, 0, { O(GqdMem)         , O(Xmm)            , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvm_OptL)     , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                  }, F(RW)|F(Avx)                       , U                    },
  { Enc(AvxRvm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                  }, F(RW)|F(Avx)                       , U                    },
  { Enc(Fma4_OptL)       , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(Fma4)            , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(XopRm_OptL)      , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(XopRm)           , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxGather)       , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Mem)            , O(Xy)             , U                 , U                  }, F(RW)|F(Avx)                       , U                    },
  { Enc(AvxGatherEx)     , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Mem)            , O(Xmm)            , U                 , U                  }, F(RW)|F(Avx)                       , U                    },
  { Enc(AvxRvmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Ymm)            , O(Ymm)            , O(XmmMem)         , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRm_OptL)      , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Mem)            , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxM)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(RO)|F(Avx)|F(Volatile)           , U                    },
  { Enc(AvxRm)           , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , U                 , U                 , U                  }, F(RO)|F(Avx)|F(Special)            , U                    },
  { Enc(AvxRvmMvr_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(Xy)             , O(XyMem)          , U                 , U                  }, F(RW)|F(Avx)                       , O_660F38(2F,U,_,_,_) },
  { Enc(AvxRvmMvr_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(Xy)             , O(XyMem)          , U                 , U                  }, F(RW)|F(Avx)                       , O_660F38(2E,U,_,_,_) },
  { Enc(AvxRmMr_OptL)    , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(29,U,_,_,_) },
  { Enc(AvxRmMr_OptL)    , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , O_000F00(29,U,_,_,_) },
  { Enc(AvxMovDQ)        , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(7E,U,_,_,_) },
  { Enc(AvxRmMr_OptL)    , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(7F,U,_,_,_) },
  { Enc(AvxRmMr_OptL)    , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , O_F30F00(7F,U,_,_,_) },
  { Enc(AvxRvm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(Xmm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmMr)        , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(Xmm)            , O(Mem)            , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(17,U,_,_,_) },
  { Enc(AvxRvmMr)        , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(Xmm)            , O(Mem)            , U                 , U                  }, F(WO)|F(Avx)                       , O_000F00(17,U,_,_,_) },
  { Enc(AvxRvmMr)        , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(Xmm)            , O(Mem)            , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(13,U,_,_,_) },
  { Enc(AvxRvmMr)        , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(Xmm)            , O(Mem)            , U                 , U                  }, F(WO)|F(Avx)                       , O_000F00(13,U,_,_,_) },
  { Enc(AvxRm_OptL)      , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , O(Xy)             , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxMr_OptL)      , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , O(Xy)             , U                 , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxMovSsSd)      , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(XmmMem)         , O(Xmm)            , U                 , U                  }, F(WO)|F(Avx)                       , O_F20F00(11,U,_,_,_) },
  { Enc(AvxMovSsSd)      , 0 , 0 , 0x00, 0x00, 0, { O(XmmMem)         , O(Xmm)            , O(Xmm)            , U                 , U                  }, F(WO)|F(Avx)                       , O_F30F00(11,U,_,_,_) },
  { Enc(AvxRmMr_OptL)    , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(11,U,_,_,_) },
  { Enc(AvxRmMr_OptL)    , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(XyMem)          , U                 , U                 , U                  }, F(WO)|F(Avx)                       , O_000F00(11,U,_,_,_) },
  { Enc(AvxRvmr)         , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Xy)             , O(XyMem)          , O(Xy)             , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(XopRvrmRvmr_OptL), 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRmi)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)|F(Special)            , U                    },
  { Enc(XopRvmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Ymm)            , O(Ymm)            , O(YmmMem)         , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvm)          , 0 , 0 , 0x00, 0x00, 0, { O(Ymm)            , O(Ymm)            , O(YmmMem)         , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvrmRvmr_OptL), 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmRmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F3A(05,U,_,_,_) },
  { Enc(AvxRvmRmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F3A(04,U,_,_,_) },
  { Enc(AvxRmi)          , 0 , 0 , 0x00, 0x00, 0, { O(Ymm)            , O(YmmMem)         , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxMri)          , 0 , 0 , 0x00, 0x00, 0, { O(GqdwbMem)       , O(Xmm)            , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxMri)          , 0 , 0 , 0x00, 0x00, 0, { O(GqMem)          , O(Xmm)            , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxMri)          , 0 , 0 , 0x00, 0x00, 0, { O(GqdwMem)        , O(Xmm)            , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(GqdwbMem)       , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(GqdMem)         , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(GqMem)          , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmi)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(GqdwMem)        , O(Imm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(XopRvmr)         , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmMvr_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(XyMem)          , O(Xy)             , O(XyMem)          , U                 , U                  }, F(WO)|F(Avx)                       , O_660F38(8E,U,_,_,_) },
  { Enc(XopRvrmRvmr)     , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(XopRvmRmi)       , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(XmmMem)|O(Imm)  , U                 , U                  }, F(WO)|F(Avx)                       , O_00_M08(C0,U,_,_,_) },
  { Enc(XopRvmRmi)       , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(XmmMem)|O(Imm)  , U                 , U                  }, F(WO)|F(Avx)                       , O_00_M08(C2,U,_,_,_) },
  { Enc(XopRvmRmi)       , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(XmmMem)|O(Imm)  , U                 , U                  }, F(WO)|F(Avx)                       , O_00_M08(C3,U,_,_,_) },
  { Enc(XopRvmRmi)       , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(XmmMem)|O(Imm)  , U                 , U                  }, F(WO)|F(Avx)                       , O_00_M08(C1,U,_,_,_) },
  { Enc(XopRvmRmv)       , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRmi_OptL)     , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmVmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(72,6,_,_,_) },
  { Enc(AvxVmi_OptL)     , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                  }, F(WO)|F(Avx)                       , U                    },
  { Enc(AvxRvmVmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(73,6,_,_,_) },
  { Enc(AvxRvmVmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(71,6,_,_,_) },
  { Enc(AvxRvmVmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(72,4,_,_,_) },
  { Enc(AvxRvmVmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(71,4,_,_,_) },
  { Enc(AvxRvmVmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(72,2,_,_,_) },
  { Enc(AvxRvmVmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(73,2,_,_,_) },
  { Enc(AvxRvmVmi_OptL)  , 0 , 0 , 0x00, 0x00, 0, { O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                  }, F(WO)|F(Avx)                       , O_660F00(71,2,_,_,_) },
  { Enc(AvxRm_OptL)      , 0 , 0 , 0x00, 0x3F, 0, { O(Xy)             , O(XyMem)          , U                 , U                 , U                  }, F(RO)|F(Avx)                       , U                    },
  { Enc(AvxM)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(Volatile)|F(Avx)                 , U                    },
  { Enc(AvxOp)           , 0 , 0 , 0x00, 0x00, 0, { U                 , U                 , U                 , U                 , U                  }, F(Volatile)|F(Avx)                 , U                    },
  { Enc(X86Rm)           , 0 , 0 , 0x00, 0x00, 0, { O(Gqd)            , U                 , U                 , U                 , U                  }, F(RO)|F(Volatile)                  , U                    },
  { Enc(X86Xadd)         , 0 , 0 , 0x00, 0x3F, 0, { O(GqdwbMem)       , O(Gqdwb)          , U                 , U                 , U                  }, F(RW)|F(Xchg)|F(Lock)              , U                    },
  { Enc(X86Xchg)         , 0 , 0 , 0x00, 0x00, 0, { O(GqdwbMem)       , O(Gqdwb)          , U                 , U                 , U                  }, F(RW)|F(Xchg)|F(Lock)              , U                    },
  { Enc(SimdRm)          , 0 , 0 , 0x00, 0x00, 0, { O(Xmm)            , O(XmmMem)         , U                 , U                 , U                  }, F(RW)|F(None)                      , U                    },
  { Enc(X86M)            , 0 , 0 , 0x00, 0x00, 0, { O(Mem)            , U                 , U                 , U                 , U                  }, F(RO)|F(Volatile)|F(Special)       , U                    }
};
// ----------------------------------------------------------------------------

// ------------------- Automatically generated, do not edit -------------------
enum X86InstData_ExtendedIndex {
    kX86InstIdNone_ExtendedIndex = 0,
    kX86InstIdAdc_ExtendedIndex = 1,
    kX86InstIdAdcx_ExtendedIndex = 2,
    kX86InstIdAdd_ExtendedIndex = 3,
    kX86InstIdAddpd_ExtendedIndex = 4,
    kX86InstIdAddps_ExtendedIndex = 4,
    kX86InstIdAddsd_ExtendedIndex = 4,
    kX86InstIdAddss_ExtendedIndex = 4,
    kX86InstIdAddsubpd_ExtendedIndex = 4,
    kX86InstIdAddsubps_ExtendedIndex = 4,
    kX86InstIdAdox_ExtendedIndex = 5,
    kX86InstIdAesdec_ExtendedIndex = 4,
    kX86InstIdAesdeclast_ExtendedIndex = 4,
    kX86InstIdAesenc_ExtendedIndex = 4,
    kX86InstIdAesenclast_ExtendedIndex = 4,
    kX86InstIdAesimc_ExtendedIndex = 6,
    kX86InstIdAeskeygenassist_ExtendedIndex = 7,
    kX86InstIdAnd_ExtendedIndex = 3,
    kX86InstIdAndn_ExtendedIndex = 8,
    kX86InstIdAndnpd_ExtendedIndex = 4,
    kX86InstIdAndnps_ExtendedIndex = 4,
    kX86InstIdAndpd_ExtendedIndex = 4,
    kX86InstIdAndps_ExtendedIndex = 4,
    kX86InstIdBextr_ExtendedIndex = 9,
    kX86InstIdBlcfill_ExtendedIndex = 10,
    kX86InstIdBlci_ExtendedIndex = 10,
    kX86InstIdBlcic_ExtendedIndex = 10,
    kX86InstIdBlcmsk_ExtendedIndex = 10,
    kX86InstIdBlcs_ExtendedIndex = 10,
    kX86InstIdBlendpd_ExtendedIndex = 11,
    kX86InstIdBlendps_ExtendedIndex = 11,
    kX86InstIdBlendvpd_ExtendedIndex = 12,
    kX86InstIdBlendvps_ExtendedIndex = 12,
    kX86InstIdBlsfill_ExtendedIndex = 10,
    kX86InstIdBlsi_ExtendedIndex = 13,
    kX86InstIdBlsic_ExtendedIndex = 10,
    kX86InstIdBlsmsk_ExtendedIndex = 13,
    kX86InstIdBlsr_ExtendedIndex = 13,
    kX86InstIdBsf_ExtendedIndex = 14,
    kX86InstIdBsr_ExtendedIndex = 14,
    kX86InstIdBswap_ExtendedIndex = 15,
    kX86InstIdBt_ExtendedIndex = 16,
    kX86InstIdBtc_ExtendedIndex = 17,
    kX86InstIdBtr_ExtendedIndex = 18,
    kX86InstIdBts_ExtendedIndex = 19,
    kX86InstIdBzhi_ExtendedIndex = 9,
    kX86InstIdCall_ExtendedIndex = 20,
    kX86InstIdCbw_ExtendedIndex = 21,
    kX86InstIdCdq_ExtendedIndex = 21,
    kX86InstIdCdqe_ExtendedIndex = 21,
    kX86InstIdClc_ExtendedIndex = 22,
    kX86InstIdCld_ExtendedIndex = 23,
    kX86InstIdClflush_ExtendedIndex = 24,
    kX86InstIdClflushopt_ExtendedIndex = 24,
    kX86InstIdCmc_ExtendedIndex = 25,
    kX86InstIdCmova_ExtendedIndex = 26,
    kX86InstIdCmovae_ExtendedIndex = 27,
    kX86InstIdCmovb_ExtendedIndex = 27,
    kX86InstIdCmovbe_ExtendedIndex = 26,
    kX86InstIdCmovc_ExtendedIndex = 27,
    kX86InstIdCmove_ExtendedIndex = 28,
    kX86InstIdCmovg_ExtendedIndex = 29,
    kX86InstIdCmovge_ExtendedIndex = 30,
    kX86InstIdCmovl_ExtendedIndex = 30,
    kX86InstIdCmovle_ExtendedIndex = 29,
    kX86InstIdCmovna_ExtendedIndex = 26,
    kX86InstIdCmovnae_ExtendedIndex = 27,
    kX86InstIdCmovnb_ExtendedIndex = 27,
    kX86InstIdCmovnbe_ExtendedIndex = 26,
    kX86InstIdCmovnc_ExtendedIndex = 27,
    kX86InstIdCmovne_ExtendedIndex = 28,
    kX86InstIdCmovng_ExtendedIndex = 29,
    kX86InstIdCmovnge_ExtendedIndex = 30,
    kX86InstIdCmovnl_ExtendedIndex = 30,
    kX86InstIdCmovnle_ExtendedIndex = 29,
    kX86InstIdCmovno_ExtendedIndex = 31,
    kX86InstIdCmovnp_ExtendedIndex = 32,
    kX86InstIdCmovns_ExtendedIndex = 33,
    kX86InstIdCmovnz_ExtendedIndex = 28,
    kX86InstIdCmovo_ExtendedIndex = 31,
    kX86InstIdCmovp_ExtendedIndex = 32,
    kX86InstIdCmovpe_ExtendedIndex = 32,
    kX86InstIdCmovpo_ExtendedIndex = 32,
    kX86InstIdCmovs_ExtendedIndex = 33,
    kX86InstIdCmovz_ExtendedIndex = 28,
    kX86InstIdCmp_ExtendedIndex = 34,
    kX86InstIdCmppd_ExtendedIndex = 11,
    kX86InstIdCmpps_ExtendedIndex = 11,
    kX86InstIdCmpsB_ExtendedIndex = 35,
    kX86InstIdCmpsD_ExtendedIndex = 35,
    kX86InstIdCmpsQ_ExtendedIndex = 35,
    kX86InstIdCmpsW_ExtendedIndex = 36,
    kX86InstIdCmpsd_ExtendedIndex = 11,
    kX86InstIdCmpss_ExtendedIndex = 11,
    kX86InstIdCmpxchg_ExtendedIndex = 37,
    kX86InstIdCmpxchg16b_ExtendedIndex = 38,
    kX86InstIdCmpxchg8b_ExtendedIndex = 38,
    kX86InstIdComisd_ExtendedIndex = 39,
    kX86InstIdComiss_ExtendedIndex = 39,
    kX86InstIdCpuid_ExtendedIndex = 40,
    kX86InstIdCqo_ExtendedIndex = 21,
    kX86InstIdCrc32_ExtendedIndex = 41,
    kX86InstIdCvtdq2pd_ExtendedIndex = 42,
    kX86InstIdCvtdq2ps_ExtendedIndex = 42,
    kX86InstIdCvtpd2dq_ExtendedIndex = 42,
    kX86InstIdCvtpd2pi_ExtendedIndex = 43,
    kX86InstIdCvtpd2ps_ExtendedIndex = 42,
    kX86InstIdCvtpi2pd_ExtendedIndex = 44,
    kX86InstIdCvtpi2ps_ExtendedIndex = 45,
    kX86InstIdCvtps2dq_ExtendedIndex = 42,
    kX86InstIdCvtps2pd_ExtendedIndex = 42,
    kX86InstIdCvtps2pi_ExtendedIndex = 43,
    kX86InstIdCvtsd2si_ExtendedIndex = 46,
    kX86InstIdCvtsd2ss_ExtendedIndex = 47,
    kX86InstIdCvtsi2sd_ExtendedIndex = 48,
    kX86InstIdCvtsi2ss_ExtendedIndex = 49,
    kX86InstIdCvtss2sd_ExtendedIndex = 50,
    kX86InstIdCvtss2si_ExtendedIndex = 46,
    kX86InstIdCvttpd2dq_ExtendedIndex = 42,
    kX86InstIdCvttpd2pi_ExtendedIndex = 43,
    kX86InstIdCvttps2dq_ExtendedIndex = 42,
    kX86InstIdCvttps2pi_ExtendedIndex = 43,
    kX86InstIdCvttsd2si_ExtendedIndex = 46,
    kX86InstIdCvttss2si_ExtendedIndex = 46,
    kX86InstIdCwd_ExtendedIndex = 21,
    kX86InstIdCwde_ExtendedIndex = 21,
    kX86InstIdDaa_ExtendedIndex = 51,
    kX86InstIdDas_ExtendedIndex = 51,
    kX86InstIdDec_ExtendedIndex = 52,
    kX86InstIdDiv_ExtendedIndex = 53,
    kX86InstIdDivpd_ExtendedIndex = 4,
    kX86InstIdDivps_ExtendedIndex = 4,
    kX86InstIdDivsd_ExtendedIndex = 4,
    kX86InstIdDivss_ExtendedIndex = 4,
    kX86InstIdDppd_ExtendedIndex = 11,
    kX86InstIdDpps_ExtendedIndex = 11,
    kX86InstIdEmms_ExtendedIndex = 54,
    kX86InstIdEnter_ExtendedIndex = 55,
    kX86InstIdExtractps_ExtendedIndex = 56,
    kX86InstIdExtrq_ExtendedIndex = 57,
    kX86InstIdF2xm1_ExtendedIndex = 58,
    kX86InstIdFabs_ExtendedIndex = 58,
    kX86InstIdFadd_ExtendedIndex = 59,
    kX86InstIdFaddp_ExtendedIndex = 60,
    kX86InstIdFbld_ExtendedIndex = 61,
    kX86InstIdFbstp_ExtendedIndex = 61,
    kX86InstIdFchs_ExtendedIndex = 58,
    kX86InstIdFclex_ExtendedIndex = 58,
    kX86InstIdFcmovb_ExtendedIndex = 62,
    kX86InstIdFcmovbe_ExtendedIndex = 63,
    kX86InstIdFcmove_ExtendedIndex = 64,
    kX86InstIdFcmovnb_ExtendedIndex = 62,
    kX86InstIdFcmovnbe_ExtendedIndex = 63,
    kX86InstIdFcmovne_ExtendedIndex = 64,
    kX86InstIdFcmovnu_ExtendedIndex = 65,
    kX86InstIdFcmovu_ExtendedIndex = 65,
    kX86InstIdFcom_ExtendedIndex = 66,
    kX86InstIdFcomi_ExtendedIndex = 67,
    kX86InstIdFcomip_ExtendedIndex = 67,
    kX86InstIdFcomp_ExtendedIndex = 66,
    kX86InstIdFcompp_ExtendedIndex = 58,
    kX86InstIdFcos_ExtendedIndex = 58,
    kX86InstIdFdecstp_ExtendedIndex = 58,
    kX86InstIdFdiv_ExtendedIndex = 59,
    kX86InstIdFdivp_ExtendedIndex = 60,
    kX86InstIdFdivr_ExtendedIndex = 59,
    kX86InstIdFdivrp_ExtendedIndex = 60,
    kX86InstIdFemms_ExtendedIndex = 68,
    kX86InstIdFfree_ExtendedIndex = 69,
    kX86InstIdFiadd_ExtendedIndex = 70,
    kX86InstIdFicom_ExtendedIndex = 70,
    kX86InstIdFicomp_ExtendedIndex = 70,
    kX86InstIdFidiv_ExtendedIndex = 70,
    kX86InstIdFidivr_ExtendedIndex = 70,
    kX86InstIdFild_ExtendedIndex = 71,
    kX86InstIdFimul_ExtendedIndex = 70,
    kX86InstIdFincstp_ExtendedIndex = 58,
    kX86InstIdFinit_ExtendedIndex = 58,
    kX86InstIdFist_ExtendedIndex = 70,
    kX86InstIdFistp_ExtendedIndex = 72,
    kX86InstIdFisttp_ExtendedIndex = 73,
    kX86InstIdFisub_ExtendedIndex = 70,
    kX86InstIdFisubr_ExtendedIndex = 70,
    kX86InstIdFld_ExtendedIndex = 74,
    kX86InstIdFld1_ExtendedIndex = 58,
    kX86InstIdFldcw_ExtendedIndex = 61,
    kX86InstIdFldenv_ExtendedIndex = 61,
    kX86InstIdFldl2e_ExtendedIndex = 58,
    kX86InstIdFldl2t_ExtendedIndex = 58,
    kX86InstIdFldlg2_ExtendedIndex = 58,
    kX86InstIdFldln2_ExtendedIndex = 58,
    kX86InstIdFldpi_ExtendedIndex = 58,
    kX86InstIdFldz_ExtendedIndex = 58,
    kX86InstIdFmul_ExtendedIndex = 59,
    kX86InstIdFmulp_ExtendedIndex = 60,
    kX86InstIdFnclex_ExtendedIndex = 58,
    kX86InstIdFninit_ExtendedIndex = 58,
    kX86InstIdFnop_ExtendedIndex = 58,
    kX86InstIdFnsave_ExtendedIndex = 61,
    kX86InstIdFnstcw_ExtendedIndex = 61,
    kX86InstIdFnstenv_ExtendedIndex = 61,
    kX86InstIdFnstsw_ExtendedIndex = 75,
    kX86InstIdFpatan_ExtendedIndex = 58,
    kX86InstIdFprem_ExtendedIndex = 58,
    kX86InstIdFprem1_ExtendedIndex = 58,
    kX86InstIdFptan_ExtendedIndex = 58,
    kX86InstIdFrndint_ExtendedIndex = 58,
    kX86InstIdFrstor_ExtendedIndex = 61,
    kX86InstIdFsave_ExtendedIndex = 61,
    kX86InstIdFscale_ExtendedIndex = 58,
    kX86InstIdFsin_ExtendedIndex = 58,
    kX86InstIdFsincos_ExtendedIndex = 58,
    kX86InstIdFsqrt_ExtendedIndex = 58,
    kX86InstIdFst_ExtendedIndex = 76,
    kX86InstIdFstcw_ExtendedIndex = 61,
    kX86InstIdFstenv_ExtendedIndex = 61,
    kX86InstIdFstp_ExtendedIndex = 77,
    kX86InstIdFstsw_ExtendedIndex = 78,
    kX86InstIdFsub_ExtendedIndex = 59,
    kX86InstIdFsubp_ExtendedIndex = 60,
    kX86InstIdFsubr_ExtendedIndex = 59,
    kX86InstIdFsubrp_ExtendedIndex = 60,
    kX86InstIdFtst_ExtendedIndex = 58,
    kX86InstIdFucom_ExtendedIndex = 60,
    kX86InstIdFucomi_ExtendedIndex = 67,
    kX86InstIdFucomip_ExtendedIndex = 67,
    kX86InstIdFucomp_ExtendedIndex = 60,
    kX86InstIdFucompp_ExtendedIndex = 58,
    kX86InstIdFwait_ExtendedIndex = 79,
    kX86InstIdFxam_ExtendedIndex = 58,
    kX86InstIdFxch_ExtendedIndex = 69,
    kX86InstIdFxrstor_ExtendedIndex = 61,
    kX86InstIdFxrstor64_ExtendedIndex = 61,
    kX86InstIdFxsave_ExtendedIndex = 61,
    kX86InstIdFxsave64_ExtendedIndex = 61,
    kX86InstIdFxtract_ExtendedIndex = 58,
    kX86InstIdFyl2x_ExtendedIndex = 58,
    kX86InstIdFyl2xp1_ExtendedIndex = 58,
    kX86InstIdHaddpd_ExtendedIndex = 4,
    kX86InstIdHaddps_ExtendedIndex = 4,
    kX86InstIdHsubpd_ExtendedIndex = 4,
    kX86InstIdHsubps_ExtendedIndex = 4,
    kX86InstIdIdiv_ExtendedIndex = 80,
    kX86InstIdImul_ExtendedIndex = 81,
    kX86InstIdInc_ExtendedIndex = 82,
    kX86InstIdInsertps_ExtendedIndex = 11,
    kX86InstIdInsertq_ExtendedIndex = 83,
    kX86InstIdInt_ExtendedIndex = 84,
    kX86InstIdJa_ExtendedIndex = 85,
    kX86InstIdJae_ExtendedIndex = 86,
    kX86InstIdJb_ExtendedIndex = 86,
    kX86InstIdJbe_ExtendedIndex = 85,
    kX86InstIdJc_ExtendedIndex = 86,
    kX86InstIdJe_ExtendedIndex = 87,
    kX86InstIdJg_ExtendedIndex = 88,
    kX86InstIdJge_ExtendedIndex = 89,
    kX86InstIdJl_ExtendedIndex = 89,
    kX86InstIdJle_ExtendedIndex = 88,
    kX86InstIdJna_ExtendedIndex = 85,
    kX86InstIdJnae_ExtendedIndex = 86,
    kX86InstIdJnb_ExtendedIndex = 86,
    kX86InstIdJnbe_ExtendedIndex = 85,
    kX86InstIdJnc_ExtendedIndex = 86,
    kX86InstIdJne_ExtendedIndex = 87,
    kX86InstIdJng_ExtendedIndex = 88,
    kX86InstIdJnge_ExtendedIndex = 89,
    kX86InstIdJnl_ExtendedIndex = 89,
    kX86InstIdJnle_ExtendedIndex = 88,
    kX86InstIdJno_ExtendedIndex = 90,
    kX86InstIdJnp_ExtendedIndex = 91,
    kX86InstIdJns_ExtendedIndex = 92,
    kX86InstIdJnz_ExtendedIndex = 87,
    kX86InstIdJo_ExtendedIndex = 90,
    kX86InstIdJp_ExtendedIndex = 91,
    kX86InstIdJpe_ExtendedIndex = 91,
    kX86InstIdJpo_ExtendedIndex = 91,
    kX86InstIdJs_ExtendedIndex = 92,
    kX86InstIdJz_ExtendedIndex = 87,
    kX86InstIdJecxz_ExtendedIndex = 93,
    kX86InstIdJmp_ExtendedIndex = 94,
    kX86InstIdLahf_ExtendedIndex = 95,
    kX86InstIdLddqu_ExtendedIndex = 96,
    kX86InstIdLdmxcsr_ExtendedIndex = 24,
    kX86InstIdLea_ExtendedIndex = 97,
    kX86InstIdLeave_ExtendedIndex = 98,
    kX86InstIdLfence_ExtendedIndex = 99,
    kX86InstIdLodsB_ExtendedIndex = 100,
    kX86InstIdLodsD_ExtendedIndex = 101,
    kX86InstIdLodsQ_ExtendedIndex = 102,
    kX86InstIdLodsW_ExtendedIndex = 103,
    kX86InstIdLzcnt_ExtendedIndex = 14,
    kX86InstIdMaskmovdqu_ExtendedIndex = 104,
    kX86InstIdMaskmovq_ExtendedIndex = 105,
    kX86InstIdMaxpd_ExtendedIndex = 4,
    kX86InstIdMaxps_ExtendedIndex = 4,
    kX86InstIdMaxsd_ExtendedIndex = 4,
    kX86InstIdMaxss_ExtendedIndex = 4,
    kX86InstIdMfence_ExtendedIndex = 106,
    kX86InstIdMinpd_ExtendedIndex = 4,
    kX86InstIdMinps_ExtendedIndex = 4,
    kX86InstIdMinsd_ExtendedIndex = 4,
    kX86InstIdMinss_ExtendedIndex = 4,
    kX86InstIdMonitor_ExtendedIndex = 107,
    kX86InstIdMov_ExtendedIndex = 108,
    kX86InstIdMovPtr_ExtendedIndex = 109,
    kX86InstIdMovapd_ExtendedIndex = 110,
    kX86InstIdMovaps_ExtendedIndex = 111,
    kX86InstIdMovbe_ExtendedIndex = 112,
    kX86InstIdMovd_ExtendedIndex = 113,
    kX86InstIdMovddup_ExtendedIndex = 114,
    kX86InstIdMovdq2q_ExtendedIndex = 115,
    kX86InstIdMovdqa_ExtendedIndex = 116,
    kX86InstIdMovdqu_ExtendedIndex = 117,
    kX86InstIdMovhlps_ExtendedIndex = 118,
    kX86InstIdMovhpd_ExtendedIndex = 119,
    kX86InstIdMovhps_ExtendedIndex = 120,
    kX86InstIdMovlhps_ExtendedIndex = 121,
    kX86InstIdMovlpd_ExtendedIndex = 122,
    kX86InstIdMovlps_ExtendedIndex = 123,
    kX86InstIdMovmskpd_ExtendedIndex = 124,
    kX86InstIdMovmskps_ExtendedIndex = 124,
    kX86InstIdMovntdq_ExtendedIndex = 125,
    kX86InstIdMovntdqa_ExtendedIndex = 126,
    kX86InstIdMovnti_ExtendedIndex = 127,
    kX86InstIdMovntpd_ExtendedIndex = 128,
    kX86InstIdMovntps_ExtendedIndex = 129,
    kX86InstIdMovntq_ExtendedIndex = 130,
    kX86InstIdMovntsd_ExtendedIndex = 131,
    kX86InstIdMovntss_ExtendedIndex = 132,
    kX86InstIdMovq_ExtendedIndex = 133,
    kX86InstIdMovq2dq_ExtendedIndex = 134,
    kX86InstIdMovsB_ExtendedIndex = 135,
    kX86InstIdMovsD_ExtendedIndex = 135,
    kX86InstIdMovsQ_ExtendedIndex = 135,
    kX86InstIdMovsW_ExtendedIndex = 136,
    kX86InstIdMovsd_ExtendedIndex = 137,
    kX86InstIdMovshdup_ExtendedIndex = 42,
    kX86InstIdMovsldup_ExtendedIndex = 42,
    kX86InstIdMovss_ExtendedIndex = 138,
    kX86InstIdMovsx_ExtendedIndex = 139,
    kX86InstIdMovsxd_ExtendedIndex = 140,
    kX86InstIdMovupd_ExtendedIndex = 141,
    kX86InstIdMovups_ExtendedIndex = 142,
    kX86InstIdMovzx_ExtendedIndex = 139,
    kX86InstIdMpsadbw_ExtendedIndex = 11,
    kX86InstIdMul_ExtendedIndex = 80,
    kX86InstIdMulpd_ExtendedIndex = 4,
    kX86InstIdMulps_ExtendedIndex = 4,
    kX86InstIdMulsd_ExtendedIndex = 4,
    kX86InstIdMulss_ExtendedIndex = 4,
    kX86InstIdMulx_ExtendedIndex = 143,
    kX86InstIdMwait_ExtendedIndex = 107,
    kX86InstIdNeg_ExtendedIndex = 144,
    kX86InstIdNop_ExtendedIndex = 145,
    kX86InstIdNot_ExtendedIndex = 146,
    kX86InstIdOr_ExtendedIndex = 3,
    kX86InstIdOrpd_ExtendedIndex = 4,
    kX86InstIdOrps_ExtendedIndex = 4,
    kX86InstIdPabsb_ExtendedIndex = 147,
    kX86InstIdPabsd_ExtendedIndex = 147,
    kX86InstIdPabsw_ExtendedIndex = 147,
    kX86InstIdPackssdw_ExtendedIndex = 147,
    kX86InstIdPacksswb_ExtendedIndex = 147,
    kX86InstIdPackusdw_ExtendedIndex = 4,
    kX86InstIdPackuswb_ExtendedIndex = 147,
    kX86InstIdPaddb_ExtendedIndex = 147,
    kX86InstIdPaddd_ExtendedIndex = 147,
    kX86InstIdPaddq_ExtendedIndex = 147,
    kX86InstIdPaddsb_ExtendedIndex = 147,
    kX86InstIdPaddsw_ExtendedIndex = 147,
    kX86InstIdPaddusb_ExtendedIndex = 147,
    kX86InstIdPaddusw_ExtendedIndex = 147,
    kX86InstIdPaddw_ExtendedIndex = 147,
    kX86InstIdPalignr_ExtendedIndex = 148,
    kX86InstIdPand_ExtendedIndex = 147,
    kX86InstIdPandn_ExtendedIndex = 147,
    kX86InstIdPause_ExtendedIndex = 149,
    kX86InstIdPavgb_ExtendedIndex = 147,
    kX86InstIdPavgusb_ExtendedIndex = 150,
    kX86InstIdPavgw_ExtendedIndex = 147,
    kX86InstIdPblendvb_ExtendedIndex = 12,
    kX86InstIdPblendw_ExtendedIndex = 11,
    kX86InstIdPclmulqdq_ExtendedIndex = 11,
    kX86InstIdPcmpeqb_ExtendedIndex = 147,
    kX86InstIdPcmpeqd_ExtendedIndex = 147,
    kX86InstIdPcmpeqq_ExtendedIndex = 4,
    kX86InstIdPcmpeqw_ExtendedIndex = 147,
    kX86InstIdPcmpestri_ExtendedIndex = 151,
    kX86InstIdPcmpestrm_ExtendedIndex = 151,
    kX86InstIdPcmpgtb_ExtendedIndex = 147,
    kX86InstIdPcmpgtd_ExtendedIndex = 147,
    kX86InstIdPcmpgtq_ExtendedIndex = 4,
    kX86InstIdPcmpgtw_ExtendedIndex = 147,
    kX86InstIdPcmpistri_ExtendedIndex = 151,
    kX86InstIdPcmpistrm_ExtendedIndex = 151,
    kX86InstIdPdep_ExtendedIndex = 152,
    kX86InstIdPext_ExtendedIndex = 152,
    kX86InstIdPextrb_ExtendedIndex = 153,
    kX86InstIdPextrd_ExtendedIndex = 154,
    kX86InstIdPextrq_ExtendedIndex = 56,
    kX86InstIdPextrw_ExtendedIndex = 155,
    kX86InstIdPf2id_ExtendedIndex = 156,
    kX86InstIdPf2iw_ExtendedIndex = 156,
    kX86InstIdPfacc_ExtendedIndex = 150,
    kX86InstIdPfadd_ExtendedIndex = 150,
    kX86InstIdPfcmpeq_ExtendedIndex = 150,
    kX86InstIdPfcmpge_ExtendedIndex = 150,
    kX86InstIdPfcmpgt_ExtendedIndex = 150,
    kX86InstIdPfmax_ExtendedIndex = 150,
    kX86InstIdPfmin_ExtendedIndex = 150,
    kX86InstIdPfmul_ExtendedIndex = 150,
    kX86InstIdPfnacc_ExtendedIndex = 150,
    kX86InstIdPfpnacc_ExtendedIndex = 150,
    kX86InstIdPfrcp_ExtendedIndex = 156,
    kX86InstIdPfrcpit1_ExtendedIndex = 150,
    kX86InstIdPfrcpit2_ExtendedIndex = 150,
    kX86InstIdPfrsqit1_ExtendedIndex = 150,
    kX86InstIdPfrsqrt_ExtendedIndex = 150,
    kX86InstIdPfsub_ExtendedIndex = 150,
    kX86InstIdPfsubr_ExtendedIndex = 150,
    kX86InstIdPhaddd_ExtendedIndex = 147,
    kX86InstIdPhaddsw_ExtendedIndex = 147,
    kX86InstIdPhaddw_ExtendedIndex = 147,
    kX86InstIdPhminposuw_ExtendedIndex = 4,
    kX86InstIdPhsubd_ExtendedIndex = 147,
    kX86InstIdPhsubsw_ExtendedIndex = 147,
    kX86InstIdPhsubw_ExtendedIndex = 147,
    kX86InstIdPi2fd_ExtendedIndex = 156,
    kX86InstIdPi2fw_ExtendedIndex = 156,
    kX86InstIdPinsrb_ExtendedIndex = 157,
    kX86InstIdPinsrd_ExtendedIndex = 157,
    kX86InstIdPinsrq_ExtendedIndex = 158,
    kX86InstIdPinsrw_ExtendedIndex = 159,
    kX86InstIdPmaddubsw_ExtendedIndex = 147,
    kX86InstIdPmaddwd_ExtendedIndex = 147,
    kX86InstIdPmaxsb_ExtendedIndex = 4,
    kX86InstIdPmaxsd_ExtendedIndex = 4,
    kX86InstIdPmaxsw_ExtendedIndex = 147,
    kX86InstIdPmaxub_ExtendedIndex = 147,
    kX86InstIdPmaxud_ExtendedIndex = 4,
    kX86InstIdPmaxuw_ExtendedIndex = 4,
    kX86InstIdPminsb_ExtendedIndex = 4,
    kX86InstIdPminsd_ExtendedIndex = 4,
    kX86InstIdPminsw_ExtendedIndex = 147,
    kX86InstIdPminub_ExtendedIndex = 147,
    kX86InstIdPminud_ExtendedIndex = 4,
    kX86InstIdPminuw_ExtendedIndex = 4,
    kX86InstIdPmovmskb_ExtendedIndex = 160,
    kX86InstIdPmovsxbd_ExtendedIndex = 42,
    kX86InstIdPmovsxbq_ExtendedIndex = 42,
    kX86InstIdPmovsxbw_ExtendedIndex = 42,
    kX86InstIdPmovsxdq_ExtendedIndex = 42,
    kX86InstIdPmovsxwd_ExtendedIndex = 42,
    kX86InstIdPmovsxwq_ExtendedIndex = 42,
    kX86InstIdPmovzxbd_ExtendedIndex = 42,
    kX86InstIdPmovzxbq_ExtendedIndex = 42,
    kX86InstIdPmovzxbw_ExtendedIndex = 42,
    kX86InstIdPmovzxdq_ExtendedIndex = 42,
    kX86InstIdPmovzxwd_ExtendedIndex = 42,
    kX86InstIdPmovzxwq_ExtendedIndex = 42,
    kX86InstIdPmuldq_ExtendedIndex = 4,
    kX86InstIdPmulhrsw_ExtendedIndex = 147,
    kX86InstIdPmulhrw_ExtendedIndex = 150,
    kX86InstIdPmulhuw_ExtendedIndex = 147,
    kX86InstIdPmulhw_ExtendedIndex = 147,
    kX86InstIdPmulld_ExtendedIndex = 4,
    kX86InstIdPmullw_ExtendedIndex = 147,
    kX86InstIdPmuludq_ExtendedIndex = 147,
    kX86InstIdPop_ExtendedIndex = 161,
    kX86InstIdPopa_ExtendedIndex = 98,
    kX86InstIdPopcnt_ExtendedIndex = 162,
    kX86InstIdPopf_ExtendedIndex = 163,
    kX86InstIdPor_ExtendedIndex = 147,
    kX86InstIdPrefetch_ExtendedIndex = 164,
    kX86InstIdPrefetch3dNow_ExtendedIndex = 24,
    kX86InstIdPrefetchw_ExtendedIndex = 165,
    kX86InstIdPrefetchwt1_ExtendedIndex = 165,
    kX86InstIdPsadbw_ExtendedIndex = 147,
    kX86InstIdPshufb_ExtendedIndex = 147,
    kX86InstIdPshufd_ExtendedIndex = 166,
    kX86InstIdPshufhw_ExtendedIndex = 166,
    kX86InstIdPshuflw_ExtendedIndex = 166,
    kX86InstIdPshufw_ExtendedIndex = 167,
    kX86InstIdPsignb_ExtendedIndex = 147,
    kX86InstIdPsignd_ExtendedIndex = 147,
    kX86InstIdPsignw_ExtendedIndex = 147,
    kX86InstIdPslld_ExtendedIndex = 168,
    kX86InstIdPslldq_ExtendedIndex = 169,
    kX86InstIdPsllq_ExtendedIndex = 170,
    kX86InstIdPsllw_ExtendedIndex = 171,
    kX86InstIdPsrad_ExtendedIndex = 172,
    kX86InstIdPsraw_ExtendedIndex = 173,
    kX86InstIdPsrld_ExtendedIndex = 174,
    kX86InstIdPsrldq_ExtendedIndex = 175,
    kX86InstIdPsrlq_ExtendedIndex = 176,
    kX86InstIdPsrlw_ExtendedIndex = 177,
    kX86InstIdPsubb_ExtendedIndex = 147,
    kX86InstIdPsubd_ExtendedIndex = 147,
    kX86InstIdPsubq_ExtendedIndex = 147,
    kX86InstIdPsubsb_ExtendedIndex = 147,
    kX86InstIdPsubsw_ExtendedIndex = 147,
    kX86InstIdPsubusb_ExtendedIndex = 147,
    kX86InstIdPsubusw_ExtendedIndex = 147,
    kX86InstIdPsubw_ExtendedIndex = 147,
    kX86InstIdPswapd_ExtendedIndex = 156,
    kX86InstIdPtest_ExtendedIndex = 39,
    kX86InstIdPunpckhbw_ExtendedIndex = 147,
    kX86InstIdPunpckhdq_ExtendedIndex = 147,
    kX86InstIdPunpckhqdq_ExtendedIndex = 4,
    kX86InstIdPunpckhwd_ExtendedIndex = 147,
    kX86InstIdPunpcklbw_ExtendedIndex = 147,
    kX86InstIdPunpckldq_ExtendedIndex = 147,
    kX86InstIdPunpcklqdq_ExtendedIndex = 4,
    kX86InstIdPunpcklwd_ExtendedIndex = 147,
    kX86InstIdPush_ExtendedIndex = 178,
    kX86InstIdPusha_ExtendedIndex = 98,
    kX86InstIdPushf_ExtendedIndex = 179,
    kX86InstIdPxor_ExtendedIndex = 147,
    kX86InstIdRcl_ExtendedIndex = 180,
    kX86InstIdRcpps_ExtendedIndex = 42,
    kX86InstIdRcpss_ExtendedIndex = 47,
    kX86InstIdRcr_ExtendedIndex = 180,
    kX86InstIdRdfsbase_ExtendedIndex = 181,
    kX86InstIdRdgsbase_ExtendedIndex = 181,
    kX86InstIdRdrand_ExtendedIndex = 182,
    kX86InstIdRdseed_ExtendedIndex = 182,
    kX86InstIdRdtsc_ExtendedIndex = 183,
    kX86InstIdRdtscp_ExtendedIndex = 183,
    kX86InstIdRepLodsB_ExtendedIndex = 184,
    kX86InstIdRepLodsD_ExtendedIndex = 184,
    kX86InstIdRepLodsQ_ExtendedIndex = 184,
    kX86InstIdRepLodsW_ExtendedIndex = 184,
    kX86InstIdRepMovsB_ExtendedIndex = 185,
    kX86InstIdRepMovsD_ExtendedIndex = 185,
    kX86InstIdRepMovsQ_ExtendedIndex = 185,
    kX86InstIdRepMovsW_ExtendedIndex = 185,
    kX86InstIdRepStosB_ExtendedIndex = 184,
    kX86InstIdRepStosD_ExtendedIndex = 184,
    kX86InstIdRepStosQ_ExtendedIndex = 184,
    kX86InstIdRepStosW_ExtendedIndex = 184,
    kX86InstIdRepeCmpsB_ExtendedIndex = 186,
    kX86InstIdRepeCmpsD_ExtendedIndex = 186,
    kX86InstIdRepeCmpsQ_ExtendedIndex = 186,
    kX86InstIdRepeCmpsW_ExtendedIndex = 186,
    kX86InstIdRepeScasB_ExtendedIndex = 186,
    kX86InstIdRepeScasD_ExtendedIndex = 186,
    kX86InstIdRepeScasQ_ExtendedIndex = 186,
    kX86InstIdRepeScasW_ExtendedIndex = 186,
    kX86InstIdRepneCmpsB_ExtendedIndex = 186,
    kX86InstIdRepneCmpsD_ExtendedIndex = 186,
    kX86InstIdRepneCmpsQ_ExtendedIndex = 186,
    kX86InstIdRepneCmpsW_ExtendedIndex = 186,
    kX86InstIdRepneScasB_ExtendedIndex = 186,
    kX86InstIdRepneScasD_ExtendedIndex = 186,
    kX86InstIdRepneScasQ_ExtendedIndex = 186,
    kX86InstIdRepneScasW_ExtendedIndex = 186,
    kX86InstIdRet_ExtendedIndex = 187,
    kX86InstIdRol_ExtendedIndex = 188,
    kX86InstIdRor_ExtendedIndex = 188,
    kX86InstIdRorx_ExtendedIndex = 189,
    kX86InstIdRoundpd_ExtendedIndex = 166,
    kX86InstIdRoundps_ExtendedIndex = 166,
    kX86InstIdRoundsd_ExtendedIndex = 190,
    kX86InstIdRoundss_ExtendedIndex = 191,
    kX86InstIdRsqrtps_ExtendedIndex = 42,
    kX86InstIdRsqrtss_ExtendedIndex = 47,
    kX86InstIdSahf_ExtendedIndex = 192,
    kX86InstIdSal_ExtendedIndex = 193,
    kX86InstIdSar_ExtendedIndex = 193,
    kX86InstIdSarx_ExtendedIndex = 194,
    kX86InstIdSbb_ExtendedIndex = 1,
    kX86InstIdScasB_ExtendedIndex = 35,
    kX86InstIdScasD_ExtendedIndex = 35,
    kX86InstIdScasQ_ExtendedIndex = 35,
    kX86InstIdScasW_ExtendedIndex = 36,
    kX86InstIdSeta_ExtendedIndex = 195,
    kX86InstIdSetae_ExtendedIndex = 196,
    kX86InstIdSetb_ExtendedIndex = 196,
    kX86InstIdSetbe_ExtendedIndex = 195,
    kX86InstIdSetc_ExtendedIndex = 196,
    kX86InstIdSete_ExtendedIndex = 197,
    kX86InstIdSetg_ExtendedIndex = 198,
    kX86InstIdSetge_ExtendedIndex = 199,
    kX86InstIdSetl_ExtendedIndex = 199,
    kX86InstIdSetle_ExtendedIndex = 198,
    kX86InstIdSetna_ExtendedIndex = 195,
    kX86InstIdSetnae_ExtendedIndex = 196,
    kX86InstIdSetnb_ExtendedIndex = 196,
    kX86InstIdSetnbe_ExtendedIndex = 195,
    kX86InstIdSetnc_ExtendedIndex = 196,
    kX86InstIdSetne_ExtendedIndex = 197,
    kX86InstIdSetng_ExtendedIndex = 198,
    kX86InstIdSetnge_ExtendedIndex = 199,
    kX86InstIdSetnl_ExtendedIndex = 199,
    kX86InstIdSetnle_ExtendedIndex = 198,
    kX86InstIdSetno_ExtendedIndex = 200,
    kX86InstIdSetnp_ExtendedIndex = 201,
    kX86InstIdSetns_ExtendedIndex = 202,
    kX86InstIdSetnz_ExtendedIndex = 197,
    kX86InstIdSeto_ExtendedIndex = 200,
    kX86InstIdSetp_ExtendedIndex = 201,
    kX86InstIdSetpe_ExtendedIndex = 201,
    kX86InstIdSetpo_ExtendedIndex = 201,
    kX86InstIdSets_ExtendedIndex = 202,
    kX86InstIdSetz_ExtendedIndex = 197,
    kX86InstIdSfence_ExtendedIndex = 99,
    kX86InstIdSha1msg1_ExtendedIndex = 4,
    kX86InstIdSha1msg2_ExtendedIndex = 4,
    kX86InstIdSha1nexte_ExtendedIndex = 4,
    kX86InstIdSha1rnds4_ExtendedIndex = 11,
    kX86InstIdSha256msg1_ExtendedIndex = 4,
    kX86InstIdSha256msg2_ExtendedIndex = 4,
    kX86InstIdSha256rnds2_ExtendedIndex = 4,
    kX86InstIdShl_ExtendedIndex = 193,
    kX86InstIdShld_ExtendedIndex = 203,
    kX86InstIdShlx_ExtendedIndex = 194,
    kX86InstIdShr_ExtendedIndex = 193,
    kX86InstIdShrd_ExtendedIndex = 204,
    kX86InstIdShrx_ExtendedIndex = 194,
    kX86InstIdShufpd_ExtendedIndex = 11,
    kX86InstIdShufps_ExtendedIndex = 11,
    kX86InstIdSqrtpd_ExtendedIndex = 42,
    kX86InstIdSqrtps_ExtendedIndex = 42,
    kX86InstIdSqrtsd_ExtendedIndex = 50,
    kX86InstIdSqrtss_ExtendedIndex = 47,
    kX86InstIdStc_ExtendedIndex = 205,
    kX86InstIdStd_ExtendedIndex = 206,
    kX86InstIdStmxcsr_ExtendedIndex = 207,
    kX86InstIdStosB_ExtendedIndex = 208,
    kX86InstIdStosD_ExtendedIndex = 208,
    kX86InstIdStosQ_ExtendedIndex = 208,
    kX86InstIdStosW_ExtendedIndex = 209,
    kX86InstIdSub_ExtendedIndex = 3,
    kX86InstIdSubpd_ExtendedIndex = 4,
    kX86InstIdSubps_ExtendedIndex = 4,
    kX86InstIdSubsd_ExtendedIndex = 4,
    kX86InstIdSubss_ExtendedIndex = 4,
    kX86InstIdT1mskc_ExtendedIndex = 10,
    kX86InstIdTest_ExtendedIndex = 210,
    kX86InstIdTzcnt_ExtendedIndex = 162,
    kX86InstIdTzmsk_ExtendedIndex = 10,
    kX86InstIdUcomisd_ExtendedIndex = 39,
    kX86InstIdUcomiss_ExtendedIndex = 39,
    kX86InstIdUd2_ExtendedIndex = 145,
    kX86InstIdUnpckhpd_ExtendedIndex = 4,
    kX86InstIdUnpckhps_ExtendedIndex = 4,
    kX86InstIdUnpcklpd_ExtendedIndex = 4,
    kX86InstIdUnpcklps_ExtendedIndex = 4,
    kX86InstIdVaddpd_ExtendedIndex = 211,
    kX86InstIdVaddps_ExtendedIndex = 211,
    kX86InstIdVaddsd_ExtendedIndex = 212,
    kX86InstIdVaddss_ExtendedIndex = 212,
    kX86InstIdVaddsubpd_ExtendedIndex = 211,
    kX86InstIdVaddsubps_ExtendedIndex = 211,
    kX86InstIdVaesdec_ExtendedIndex = 212,
    kX86InstIdVaesdeclast_ExtendedIndex = 212,
    kX86InstIdVaesenc_ExtendedIndex = 212,
    kX86InstIdVaesenclast_ExtendedIndex = 212,
    kX86InstIdVaesimc_ExtendedIndex = 213,
    kX86InstIdVaeskeygenassist_ExtendedIndex = 214,
    kX86InstIdVandnpd_ExtendedIndex = 211,
    kX86InstIdVandnps_ExtendedIndex = 211,
    kX86InstIdVandpd_ExtendedIndex = 211,
    kX86InstIdVandps_ExtendedIndex = 211,
    kX86InstIdVblendpd_ExtendedIndex = 215,
    kX86InstIdVblendps_ExtendedIndex = 215,
    kX86InstIdVblendvpd_ExtendedIndex = 216,
    kX86InstIdVblendvps_ExtendedIndex = 216,
    kX86InstIdVbroadcastf128_ExtendedIndex = 217,
    kX86InstIdVbroadcasti128_ExtendedIndex = 217,
    kX86InstIdVbroadcastsd_ExtendedIndex = 218,
    kX86InstIdVbroadcastss_ExtendedIndex = 219,
    kX86InstIdVcmppd_ExtendedIndex = 215,
    kX86InstIdVcmpps_ExtendedIndex = 215,
    kX86InstIdVcmpsd_ExtendedIndex = 220,
    kX86InstIdVcmpss_ExtendedIndex = 220,
    kX86InstIdVcomisd_ExtendedIndex = 221,
    kX86InstIdVcomiss_ExtendedIndex = 221,
    kX86InstIdVcvtdq2pd_ExtendedIndex = 219,
    kX86InstIdVcvtdq2ps_ExtendedIndex = 222,
    kX86InstIdVcvtpd2dq_ExtendedIndex = 223,
    kX86InstIdVcvtpd2ps_ExtendedIndex = 223,
    kX86InstIdVcvtph2ps_ExtendedIndex = 219,
    kX86InstIdVcvtps2dq_ExtendedIndex = 222,
    kX86InstIdVcvtps2pd_ExtendedIndex = 219,
    kX86InstIdVcvtps2ph_ExtendedIndex = 224,
    kX86InstIdVcvtsd2si_ExtendedIndex = 225,
    kX86InstIdVcvtsd2ss_ExtendedIndex = 212,
    kX86InstIdVcvtsi2sd_ExtendedIndex = 226,
    kX86InstIdVcvtsi2ss_ExtendedIndex = 226,
    kX86InstIdVcvtss2sd_ExtendedIndex = 212,
    kX86InstIdVcvtss2si_ExtendedIndex = 225,
    kX86InstIdVcvttpd2dq_ExtendedIndex = 227,
    kX86InstIdVcvttps2dq_ExtendedIndex = 222,
    kX86InstIdVcvttsd2si_ExtendedIndex = 225,
    kX86InstIdVcvttss2si_ExtendedIndex = 225,
    kX86InstIdVdivpd_ExtendedIndex = 211,
    kX86InstIdVdivps_ExtendedIndex = 211,
    kX86InstIdVdivsd_ExtendedIndex = 212,
    kX86InstIdVdivss_ExtendedIndex = 212,
    kX86InstIdVdppd_ExtendedIndex = 220,
    kX86InstIdVdpps_ExtendedIndex = 215,
    kX86InstIdVextractf128_ExtendedIndex = 228,
    kX86InstIdVextracti128_ExtendedIndex = 228,
    kX86InstIdVextractps_ExtendedIndex = 229,
    kX86InstIdVfmadd132pd_ExtendedIndex = 230,
    kX86InstIdVfmadd132ps_ExtendedIndex = 230,
    kX86InstIdVfmadd132sd_ExtendedIndex = 231,
    kX86InstIdVfmadd132ss_ExtendedIndex = 231,
    kX86InstIdVfmadd213pd_ExtendedIndex = 230,
    kX86InstIdVfmadd213ps_ExtendedIndex = 230,
    kX86InstIdVfmadd213sd_ExtendedIndex = 231,
    kX86InstIdVfmadd213ss_ExtendedIndex = 231,
    kX86InstIdVfmadd231pd_ExtendedIndex = 230,
    kX86InstIdVfmadd231ps_ExtendedIndex = 230,
    kX86InstIdVfmadd231sd_ExtendedIndex = 231,
    kX86InstIdVfmadd231ss_ExtendedIndex = 231,
    kX86InstIdVfmaddpd_ExtendedIndex = 232,
    kX86InstIdVfmaddps_ExtendedIndex = 232,
    kX86InstIdVfmaddsd_ExtendedIndex = 233,
    kX86InstIdVfmaddss_ExtendedIndex = 233,
    kX86InstIdVfmaddsub132pd_ExtendedIndex = 230,
    kX86InstIdVfmaddsub132ps_ExtendedIndex = 230,
    kX86InstIdVfmaddsub213pd_ExtendedIndex = 230,
    kX86InstIdVfmaddsub213ps_ExtendedIndex = 230,
    kX86InstIdVfmaddsub231pd_ExtendedIndex = 230,
    kX86InstIdVfmaddsub231ps_ExtendedIndex = 230,
    kX86InstIdVfmaddsubpd_ExtendedIndex = 232,
    kX86InstIdVfmaddsubps_ExtendedIndex = 232,
    kX86InstIdVfmsub132pd_ExtendedIndex = 230,
    kX86InstIdVfmsub132ps_ExtendedIndex = 230,
    kX86InstIdVfmsub132sd_ExtendedIndex = 231,
    kX86InstIdVfmsub132ss_ExtendedIndex = 231,
    kX86InstIdVfmsub213pd_ExtendedIndex = 230,
    kX86InstIdVfmsub213ps_ExtendedIndex = 230,
    kX86InstIdVfmsub213sd_ExtendedIndex = 231,
    kX86InstIdVfmsub213ss_ExtendedIndex = 231,
    kX86InstIdVfmsub231pd_ExtendedIndex = 230,
    kX86InstIdVfmsub231ps_ExtendedIndex = 230,
    kX86InstIdVfmsub231sd_ExtendedIndex = 231,
    kX86InstIdVfmsub231ss_ExtendedIndex = 231,
    kX86InstIdVfmsubadd132pd_ExtendedIndex = 230,
    kX86InstIdVfmsubadd132ps_ExtendedIndex = 230,
    kX86InstIdVfmsubadd213pd_ExtendedIndex = 230,
    kX86InstIdVfmsubadd213ps_ExtendedIndex = 230,
    kX86InstIdVfmsubadd231pd_ExtendedIndex = 230,
    kX86InstIdVfmsubadd231ps_ExtendedIndex = 230,
    kX86InstIdVfmsubaddpd_ExtendedIndex = 232,
    kX86InstIdVfmsubaddps_ExtendedIndex = 232,
    kX86InstIdVfmsubpd_ExtendedIndex = 232,
    kX86InstIdVfmsubps_ExtendedIndex = 232,
    kX86InstIdVfmsubsd_ExtendedIndex = 233,
    kX86InstIdVfmsubss_ExtendedIndex = 233,
    kX86InstIdVfnmadd132pd_ExtendedIndex = 230,
    kX86InstIdVfnmadd132ps_ExtendedIndex = 230,
    kX86InstIdVfnmadd132sd_ExtendedIndex = 231,
    kX86InstIdVfnmadd132ss_ExtendedIndex = 231,
    kX86InstIdVfnmadd213pd_ExtendedIndex = 230,
    kX86InstIdVfnmadd213ps_ExtendedIndex = 230,
    kX86InstIdVfnmadd213sd_ExtendedIndex = 231,
    kX86InstIdVfnmadd213ss_ExtendedIndex = 231,
    kX86InstIdVfnmadd231pd_ExtendedIndex = 230,
    kX86InstIdVfnmadd231ps_ExtendedIndex = 230,
    kX86InstIdVfnmadd231sd_ExtendedIndex = 231,
    kX86InstIdVfnmadd231ss_ExtendedIndex = 231,
    kX86InstIdVfnmaddpd_ExtendedIndex = 232,
    kX86InstIdVfnmaddps_ExtendedIndex = 232,
    kX86InstIdVfnmaddsd_ExtendedIndex = 233,
    kX86InstIdVfnmaddss_ExtendedIndex = 233,
    kX86InstIdVfnmsub132pd_ExtendedIndex = 230,
    kX86InstIdVfnmsub132ps_ExtendedIndex = 230,
    kX86InstIdVfnmsub132sd_ExtendedIndex = 231,
    kX86InstIdVfnmsub132ss_ExtendedIndex = 231,
    kX86InstIdVfnmsub213pd_ExtendedIndex = 230,
    kX86InstIdVfnmsub213ps_ExtendedIndex = 230,
    kX86InstIdVfnmsub213sd_ExtendedIndex = 231,
    kX86InstIdVfnmsub213ss_ExtendedIndex = 231,
    kX86InstIdVfnmsub231pd_ExtendedIndex = 230,
    kX86InstIdVfnmsub231ps_ExtendedIndex = 230,
    kX86InstIdVfnmsub231sd_ExtendedIndex = 231,
    kX86InstIdVfnmsub231ss_ExtendedIndex = 231,
    kX86InstIdVfnmsubpd_ExtendedIndex = 232,
    kX86InstIdVfnmsubps_ExtendedIndex = 232,
    kX86InstIdVfnmsubsd_ExtendedIndex = 233,
    kX86InstIdVfnmsubss_ExtendedIndex = 233,
    kX86InstIdVfrczpd_ExtendedIndex = 234,
    kX86InstIdVfrczps_ExtendedIndex = 234,
    kX86InstIdVfrczsd_ExtendedIndex = 235,
    kX86InstIdVfrczss_ExtendedIndex = 235,
    kX86InstIdVgatherdpd_ExtendedIndex = 236,
    kX86InstIdVgatherdps_ExtendedIndex = 236,
    kX86InstIdVgatherqpd_ExtendedIndex = 236,
    kX86InstIdVgatherqps_ExtendedIndex = 237,
    kX86InstIdVhaddpd_ExtendedIndex = 211,
    kX86InstIdVhaddps_ExtendedIndex = 211,
    kX86InstIdVhsubpd_ExtendedIndex = 211,
    kX86InstIdVhsubps_ExtendedIndex = 211,
    kX86InstIdVinsertf128_ExtendedIndex = 238,
    kX86InstIdVinserti128_ExtendedIndex = 238,
    kX86InstIdVinsertps_ExtendedIndex = 220,
    kX86InstIdVlddqu_ExtendedIndex = 239,
    kX86InstIdVldmxcsr_ExtendedIndex = 240,
    kX86InstIdVmaskmovdqu_ExtendedIndex = 241,
    kX86InstIdVmaskmovpd_ExtendedIndex = 242,
    kX86InstIdVmaskmovps_ExtendedIndex = 243,
    kX86InstIdVmaxpd_ExtendedIndex = 211,
    kX86InstIdVmaxps_ExtendedIndex = 211,
    kX86InstIdVmaxsd_ExtendedIndex = 211,
    kX86InstIdVmaxss_ExtendedIndex = 211,
    kX86InstIdVminpd_ExtendedIndex = 211,
    kX86InstIdVminps_ExtendedIndex = 211,
    kX86InstIdVminsd_ExtendedIndex = 211,
    kX86InstIdVminss_ExtendedIndex = 211,
    kX86InstIdVmovapd_ExtendedIndex = 244,
    kX86InstIdVmovaps_ExtendedIndex = 245,
    kX86InstIdVmovd_ExtendedIndex = 246,
    kX86InstIdVmovddup_ExtendedIndex = 222,
    kX86InstIdVmovdqa_ExtendedIndex = 247,
    kX86InstIdVmovdqu_ExtendedIndex = 248,
    kX86InstIdVmovhlps_ExtendedIndex = 249,
    kX86InstIdVmovhpd_ExtendedIndex = 250,
    kX86InstIdVmovhps_ExtendedIndex = 251,
    kX86InstIdVmovlhps_ExtendedIndex = 249,
    kX86InstIdVmovlpd_ExtendedIndex = 252,
    kX86InstIdVmovlps_ExtendedIndex = 253,
    kX86InstIdVmovmskpd_ExtendedIndex = 254,
    kX86InstIdVmovmskps_ExtendedIndex = 254,
    kX86InstIdVmovntdq_ExtendedIndex = 255,
    kX86InstIdVmovntdqa_ExtendedIndex = 239,
    kX86InstIdVmovntpd_ExtendedIndex = 255,
    kX86InstIdVmovntps_ExtendedIndex = 255,
    kX86InstIdVmovq_ExtendedIndex = 246,
    kX86InstIdVmovsd_ExtendedIndex = 256,
    kX86InstIdVmovshdup_ExtendedIndex = 222,
    kX86InstIdVmovsldup_ExtendedIndex = 222,
    kX86InstIdVmovss_ExtendedIndex = 257,
    kX86InstIdVmovupd_ExtendedIndex = 258,
    kX86InstIdVmovups_ExtendedIndex = 259,
    kX86InstIdVmpsadbw_ExtendedIndex = 215,
    kX86InstIdVmulpd_ExtendedIndex = 211,
    kX86InstIdVmulps_ExtendedIndex = 211,
    kX86InstIdVmulsd_ExtendedIndex = 211,
    kX86InstIdVmulss_ExtendedIndex = 211,
    kX86InstIdVorpd_ExtendedIndex = 211,
    kX86InstIdVorps_ExtendedIndex = 211,
    kX86InstIdVpabsb_ExtendedIndex = 222,
    kX86InstIdVpabsd_ExtendedIndex = 222,
    kX86InstIdVpabsw_ExtendedIndex = 222,
    kX86InstIdVpackssdw_ExtendedIndex = 211,
    kX86InstIdVpacksswb_ExtendedIndex = 211,
    kX86InstIdVpackusdw_ExtendedIndex = 211,
    kX86InstIdVpackuswb_ExtendedIndex = 211,
    kX86InstIdVpaddb_ExtendedIndex = 211,
    kX86InstIdVpaddd_ExtendedIndex = 211,
    kX86InstIdVpaddq_ExtendedIndex = 211,
    kX86InstIdVpaddsb_ExtendedIndex = 211,
    kX86InstIdVpaddsw_ExtendedIndex = 211,
    kX86InstIdVpaddusb_ExtendedIndex = 211,
    kX86InstIdVpaddusw_ExtendedIndex = 211,
    kX86InstIdVpaddw_ExtendedIndex = 211,
    kX86InstIdVpalignr_ExtendedIndex = 215,
    kX86InstIdVpand_ExtendedIndex = 211,
    kX86InstIdVpandn_ExtendedIndex = 211,
    kX86InstIdVpavgb_ExtendedIndex = 211,
    kX86InstIdVpavgw_ExtendedIndex = 211,
    kX86InstIdVpblendd_ExtendedIndex = 215,
    kX86InstIdVpblendvb_ExtendedIndex = 260,
    kX86InstIdVpblendw_ExtendedIndex = 215,
    kX86InstIdVpbroadcastb_ExtendedIndex = 219,
    kX86InstIdVpbroadcastd_ExtendedIndex = 219,
    kX86InstIdVpbroadcastq_ExtendedIndex = 219,
    kX86InstIdVpbroadcastw_ExtendedIndex = 219,
    kX86InstIdVpclmulqdq_ExtendedIndex = 220,
    kX86InstIdVpcmov_ExtendedIndex = 261,
    kX86InstIdVpcmpeqb_ExtendedIndex = 211,
    kX86InstIdVpcmpeqd_ExtendedIndex = 211,
    kX86InstIdVpcmpeqq_ExtendedIndex = 211,
    kX86InstIdVpcmpeqw_ExtendedIndex = 211,
    kX86InstIdVpcmpestri_ExtendedIndex = 262,
    kX86InstIdVpcmpestrm_ExtendedIndex = 262,
    kX86InstIdVpcmpgtb_ExtendedIndex = 211,
    kX86InstIdVpcmpgtd_ExtendedIndex = 211,
    kX86InstIdVpcmpgtq_ExtendedIndex = 211,
    kX86InstIdVpcmpgtw_ExtendedIndex = 211,
    kX86InstIdVpcmpistri_ExtendedIndex = 262,
    kX86InstIdVpcmpistrm_ExtendedIndex = 262,
    kX86InstIdVpcomb_ExtendedIndex = 263,
    kX86InstIdVpcomd_ExtendedIndex = 263,
    kX86InstIdVpcomq_ExtendedIndex = 263,
    kX86InstIdVpcomub_ExtendedIndex = 263,
    kX86InstIdVpcomud_ExtendedIndex = 263,
    kX86InstIdVpcomuq_ExtendedIndex = 263,
    kX86InstIdVpcomuw_ExtendedIndex = 263,
    kX86InstIdVpcomw_ExtendedIndex = 263,
    kX86InstIdVperm2f128_ExtendedIndex = 264,
    kX86InstIdVperm2i128_ExtendedIndex = 264,
    kX86InstIdVpermd_ExtendedIndex = 265,
    kX86InstIdVpermil2pd_ExtendedIndex = 266,
    kX86InstIdVpermil2ps_ExtendedIndex = 266,
    kX86InstIdVpermilpd_ExtendedIndex = 267,
    kX86InstIdVpermilps_ExtendedIndex = 268,
    kX86InstIdVpermpd_ExtendedIndex = 269,
    kX86InstIdVpermps_ExtendedIndex = 265,
    kX86InstIdVpermq_ExtendedIndex = 269,
    kX86InstIdVpextrb_ExtendedIndex = 270,
    kX86InstIdVpextrd_ExtendedIndex = 229,
    kX86InstIdVpextrq_ExtendedIndex = 271,
    kX86InstIdVpextrw_ExtendedIndex = 272,
    kX86InstIdVpgatherdd_ExtendedIndex = 236,
    kX86InstIdVpgatherdq_ExtendedIndex = 236,
    kX86InstIdVpgatherqd_ExtendedIndex = 237,
    kX86InstIdVpgatherqq_ExtendedIndex = 236,
    kX86InstIdVphaddbd_ExtendedIndex = 235,
    kX86InstIdVphaddbq_ExtendedIndex = 235,
    kX86InstIdVphaddbw_ExtendedIndex = 235,
    kX86InstIdVphaddd_ExtendedIndex = 211,
    kX86InstIdVphadddq_ExtendedIndex = 235,
    kX86InstIdVphaddsw_ExtendedIndex = 211,
    kX86InstIdVphaddubd_ExtendedIndex = 235,
    kX86InstIdVphaddubq_ExtendedIndex = 235,
    kX86InstIdVphaddubw_ExtendedIndex = 235,
    kX86InstIdVphaddudq_ExtendedIndex = 235,
    kX86InstIdVphadduwd_ExtendedIndex = 235,
    kX86InstIdVphadduwq_ExtendedIndex = 235,
    kX86InstIdVphaddw_ExtendedIndex = 211,
    kX86InstIdVphaddwd_ExtendedIndex = 235,
    kX86InstIdVphaddwq_ExtendedIndex = 235,
    kX86InstIdVphminposuw_ExtendedIndex = 213,
    kX86InstIdVphsubbw_ExtendedIndex = 235,
    kX86InstIdVphsubd_ExtendedIndex = 211,
    kX86InstIdVphsubdq_ExtendedIndex = 235,
    kX86InstIdVphsubsw_ExtendedIndex = 211,
    kX86InstIdVphsubw_ExtendedIndex = 211,
    kX86InstIdVphsubwd_ExtendedIndex = 235,
    kX86InstIdVpinsrb_ExtendedIndex = 273,
    kX86InstIdVpinsrd_ExtendedIndex = 274,
    kX86InstIdVpinsrq_ExtendedIndex = 275,
    kX86InstIdVpinsrw_ExtendedIndex = 276,
    kX86InstIdVpmacsdd_ExtendedIndex = 277,
    kX86InstIdVpmacsdqh_ExtendedIndex = 277,
    kX86InstIdVpmacsdql_ExtendedIndex = 277,
    kX86InstIdVpmacssdd_ExtendedIndex = 277,
    kX86InstIdVpmacssdqh_ExtendedIndex = 277,
    kX86InstIdVpmacssdql_ExtendedIndex = 277,
    kX86InstIdVpmacsswd_ExtendedIndex = 277,
    kX86InstIdVpmacssww_ExtendedIndex = 277,
    kX86InstIdVpmacswd_ExtendedIndex = 277,
    kX86InstIdVpmacsww_ExtendedIndex = 277,
    kX86InstIdVpmadcsswd_ExtendedIndex = 277,
    kX86InstIdVpmadcswd_ExtendedIndex = 277,
    kX86InstIdVpmaddubsw_ExtendedIndex = 211,
    kX86InstIdVpmaddwd_ExtendedIndex = 211,
    kX86InstIdVpmaskmovd_ExtendedIndex = 278,
    kX86InstIdVpmaskmovq_ExtendedIndex = 278,
    kX86InstIdVpmaxsb_ExtendedIndex = 211,
    kX86InstIdVpmaxsd_ExtendedIndex = 211,
    kX86InstIdVpmaxsw_ExtendedIndex = 211,
    kX86InstIdVpmaxub_ExtendedIndex = 211,
    kX86InstIdVpmaxud_ExtendedIndex = 211,
    kX86InstIdVpmaxuw_ExtendedIndex = 211,
    kX86InstIdVpminsb_ExtendedIndex = 211,
    kX86InstIdVpminsd_ExtendedIndex = 211,
    kX86InstIdVpminsw_ExtendedIndex = 211,
    kX86InstIdVpminub_ExtendedIndex = 211,
    kX86InstIdVpminud_ExtendedIndex = 211,
    kX86InstIdVpminuw_ExtendedIndex = 211,
    kX86InstIdVpmovmskb_ExtendedIndex = 254,
    kX86InstIdVpmovsxbd_ExtendedIndex = 222,
    kX86InstIdVpmovsxbq_ExtendedIndex = 222,
    kX86InstIdVpmovsxbw_ExtendedIndex = 222,
    kX86InstIdVpmovsxdq_ExtendedIndex = 222,
    kX86InstIdVpmovsxwd_ExtendedIndex = 222,
    kX86InstIdVpmovsxwq_ExtendedIndex = 222,
    kX86InstIdVpmovzxbd_ExtendedIndex = 222,
    kX86InstIdVpmovzxbq_ExtendedIndex = 222,
    kX86InstIdVpmovzxbw_ExtendedIndex = 222,
    kX86InstIdVpmovzxdq_ExtendedIndex = 222,
    kX86InstIdVpmovzxwd_ExtendedIndex = 222,
    kX86InstIdVpmovzxwq_ExtendedIndex = 222,
    kX86InstIdVpmuldq_ExtendedIndex = 211,
    kX86InstIdVpmulhrsw_ExtendedIndex = 211,
    kX86InstIdVpmulhuw_ExtendedIndex = 211,
    kX86InstIdVpmulhw_ExtendedIndex = 211,
    kX86InstIdVpmulld_ExtendedIndex = 211,
    kX86InstIdVpmullw_ExtendedIndex = 211,
    kX86InstIdVpmuludq_ExtendedIndex = 211,
    kX86InstIdVpor_ExtendedIndex = 211,
    kX86InstIdVpperm_ExtendedIndex = 279,
    kX86InstIdVprotb_ExtendedIndex = 280,
    kX86InstIdVprotd_ExtendedIndex = 281,
    kX86InstIdVprotq_ExtendedIndex = 282,
    kX86InstIdVprotw_ExtendedIndex = 283,
    kX86InstIdVpsadbw_ExtendedIndex = 211,
    kX86InstIdVpshab_ExtendedIndex = 284,
    kX86InstIdVpshad_ExtendedIndex = 284,
    kX86InstIdVpshaq_ExtendedIndex = 284,
    kX86InstIdVpshaw_ExtendedIndex = 284,
    kX86InstIdVpshlb_ExtendedIndex = 284,
    kX86InstIdVpshld_ExtendedIndex = 284,
    kX86InstIdVpshlq_ExtendedIndex = 284,
    kX86InstIdVpshlw_ExtendedIndex = 284,
    kX86InstIdVpshufb_ExtendedIndex = 211,
    kX86InstIdVpshufd_ExtendedIndex = 285,
    kX86InstIdVpshufhw_ExtendedIndex = 285,
    kX86InstIdVpshuflw_ExtendedIndex = 285,
    kX86InstIdVpsignb_ExtendedIndex = 211,
    kX86InstIdVpsignd_ExtendedIndex = 211,
    kX86InstIdVpsignw_ExtendedIndex = 211,
    kX86InstIdVpslld_ExtendedIndex = 286,
    kX86InstIdVpslldq_ExtendedIndex = 287,
    kX86InstIdVpsllq_ExtendedIndex = 288,
    kX86InstIdVpsllvd_ExtendedIndex = 211,
    kX86InstIdVpsllvq_ExtendedIndex = 211,
    kX86InstIdVpsllw_ExtendedIndex = 289,
    kX86InstIdVpsrad_ExtendedIndex = 290,
    kX86InstIdVpsravd_ExtendedIndex = 211,
    kX86InstIdVpsraw_ExtendedIndex = 291,
    kX86InstIdVpsrld_ExtendedIndex = 292,
    kX86InstIdVpsrldq_ExtendedIndex = 287,
    kX86InstIdVpsrlq_ExtendedIndex = 293,
    kX86InstIdVpsrlvd_ExtendedIndex = 211,
    kX86InstIdVpsrlvq_ExtendedIndex = 211,
    kX86InstIdVpsrlw_ExtendedIndex = 294,
    kX86InstIdVpsubb_ExtendedIndex = 211,
    kX86InstIdVpsubd_ExtendedIndex = 211,
    kX86InstIdVpsubq_ExtendedIndex = 211,
    kX86InstIdVpsubsb_ExtendedIndex = 211,
    kX86InstIdVpsubsw_ExtendedIndex = 211,
    kX86InstIdVpsubusb_ExtendedIndex = 211,
    kX86InstIdVpsubusw_ExtendedIndex = 211,
    kX86InstIdVpsubw_ExtendedIndex = 211,
    kX86InstIdVptest_ExtendedIndex = 295,
    kX86InstIdVpunpckhbw_ExtendedIndex = 211,
    kX86InstIdVpunpckhdq_ExtendedIndex = 211,
    kX86InstIdVpunpckhqdq_ExtendedIndex = 211,
    kX86InstIdVpunpckhwd_ExtendedIndex = 211,
    kX86InstIdVpunpcklbw_ExtendedIndex = 211,
    kX86InstIdVpunpckldq_ExtendedIndex = 211,
    kX86InstIdVpunpcklqdq_ExtendedIndex = 211,
    kX86InstIdVpunpcklwd_ExtendedIndex = 211,
    kX86InstIdVpxor_ExtendedIndex = 211,
    kX86InstIdVrcpps_ExtendedIndex = 222,
    kX86InstIdVrcpss_ExtendedIndex = 212,
    kX86InstIdVroundpd_ExtendedIndex = 285,
    kX86InstIdVroundps_ExtendedIndex = 285,
    kX86InstIdVroundsd_ExtendedIndex = 220,
    kX86InstIdVroundss_ExtendedIndex = 220,
    kX86InstIdVrsqrtps_ExtendedIndex = 222,
    kX86InstIdVrsqrtss_ExtendedIndex = 212,
    kX86InstIdVshufpd_ExtendedIndex = 215,
    kX86InstIdVshufps_ExtendedIndex = 215,
    kX86InstIdVsqrtpd_ExtendedIndex = 222,
    kX86InstIdVsqrtps_ExtendedIndex = 222,
    kX86InstIdVsqrtsd_ExtendedIndex = 212,
    kX86InstIdVsqrtss_ExtendedIndex = 212,
    kX86InstIdVstmxcsr_ExtendedIndex = 296,
    kX86InstIdVsubpd_ExtendedIndex = 211,
    kX86InstIdVsubps_ExtendedIndex = 211,
    kX86InstIdVsubsd_ExtendedIndex = 212,
    kX86InstIdVsubss_ExtendedIndex = 212,
    kX86InstIdVtestpd_ExtendedIndex = 295,
    kX86InstIdVtestps_ExtendedIndex = 295,
    kX86InstIdVucomisd_ExtendedIndex = 221,
    kX86InstIdVucomiss_ExtendedIndex = 221,
    kX86InstIdVunpckhpd_ExtendedIndex = 211,
    kX86InstIdVunpckhps_ExtendedIndex = 211,
    kX86InstIdVunpcklpd_ExtendedIndex = 211,
    kX86InstIdVunpcklps_ExtendedIndex = 211,
    kX86InstIdVxorpd_ExtendedIndex = 211,
    kX86InstIdVxorps_ExtendedIndex = 211,
    kX86InstIdVzeroall_ExtendedIndex = 297,
    kX86InstIdVzeroupper_ExtendedIndex = 297,
    kX86InstIdWrfsbase_ExtendedIndex = 298,
    kX86InstIdWrgsbase_ExtendedIndex = 298,
    kX86InstIdXadd_ExtendedIndex = 299,
    kX86InstIdXchg_ExtendedIndex = 300,
    kX86InstIdXgetbv_ExtendedIndex = 135,
    kX86InstIdXor_ExtendedIndex = 3,
    kX86InstIdXorpd_ExtendedIndex = 301,
    kX86InstIdXorps_ExtendedIndex = 301,
    kX86InstIdXrstor_ExtendedIndex = 302,
    kX86InstIdXrstor64_ExtendedIndex = 302,
    kX86InstIdXsave_ExtendedIndex = 302,
    kX86InstIdXsave64_ExtendedIndex = 302,
    kX86InstIdXsaveopt_ExtendedIndex = 302,
    kX86InstIdXsaveopt64_ExtendedIndex = 302,
    kX86InstIdXsetbv_ExtendedIndex = 107
};
// ----------------------------------------------------------------------------
// ${X86InstData:End}

// Please run tools/src-gendefs.js (by using just node.js, without any dependencies) to regenerate the code enclosed with ${X86InstData...}.
const X86InstInfo _x86InstInfo[] = {
  // <-----------------+-------------------+----------------------+-------------------------------------------+------------------------------------+-------------+-------+---------------------------------------------------------------------------------------------------+
  //                   |                   |                      |           Instruction Opcodes             |         Instruction Flags          |   E-FLAGS   | Write |              Operands (Gp/Fp/Mm/K/Xmm/Ymm/Zmm Regs, Mem, Imm, Label, None/Undefined)              |
  //  Instruction Id   | Instruction Name  | Instruction Encoding +---------------------+---------------------+---------------+--------------------+-------------+---+---+-------------------+-------------------+-------------------+-------------------+-------------------+
  //                   |                   |                      | 0:PP-MMM OP/O L/W/EW| 1:PP-MMM OP/O L/W/EW| Global Flags  |A512(ID|VL|kz|rnd|b)| EF:OSZAPCDX |Idx| Sz|  [0] 1st Operand  |  [1] 2nd Operand  |  [2] 3rd Operand  |  [3] 4th Operand  |  [4] 5th Operand  |
  // <-----------------+-------------------+----------------------+---------------------+---------------------+---------------+--------------------+-------------+---+---+-------------------+-------------------+-------------------+-------------------+-------------------+
  INST(None            , ""                , Enc(None)            , U                   , U                   , F(None)                            , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Adc             , "adc"             , Enc(X86Arith)        , O_000000(10,2,_,_,_), U                   , F(RW)|F(Lock)                      , EF(WWWWWX__), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(Adcx            , "adcx"            , Enc(X86RegRm)        , O_660F38(F6,U,_,_,_), U                   , F(RW)                              , EF(_____X__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Add             , "add"             , Enc(X86Arith)        , O_000000(00,0,_,_,_), U                   , F(RW)|F(Lock)                      , EF(WWWWWW__), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(Addpd           , "addpd"           , Enc(SimdRm)          , O_660F00(58,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Addps           , "addps"           , Enc(SimdRm)          , O_000F00(58,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Addsd           , "addsd"           , Enc(SimdRm)          , O_F20F00(58,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Addss           , "addss"           , Enc(SimdRm)          , O_F30F00(58,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Addsubpd        , "addsubpd"        , Enc(SimdRm)          , O_660F00(D0,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Addsubps        , "addsubps"        , Enc(SimdRm)          , O_F20F00(D0,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Adox            , "adox"            , Enc(X86RegRm)        , O_F30F38(F6,U,_,_,_), U                   , F(RW)                              , EF(X_______), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Aesdec          , "aesdec"          , Enc(SimdRm)          , O_660F38(DE,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Aesdeclast      , "aesdeclast"      , Enc(SimdRm)          , O_660F38(DF,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Aesenc          , "aesenc"          , Enc(SimdRm)          , O_660F38(DC,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Aesenclast      , "aesenclast"      , Enc(SimdRm)          , O_660F38(DD,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Aesimc          , "aesimc"          , Enc(SimdRm)          , O_660F38(DB,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Aeskeygenassist , "aeskeygenassist" , Enc(SimdRmi)         , O_660F3A(DF,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(And             , "and"             , Enc(X86Arith)        , O_000000(20,4,_,_,_), U                   , F(RW)|F(Lock)                      , EF(WWWUWW__), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(Andn            , "andn"            , Enc(AvxRvm_OptW)     , O_000F38(F2,U,_,_,_), U                   , F(RW)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(Gqd)            , O(GqdMem)         , U                 , U                 ),
  INST(Andnpd          , "andnpd"          , Enc(SimdRm)          , O_660F00(55,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Andnps          , "andnps"          , Enc(SimdRm)          , O_000F00(55,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Andpd           , "andpd"           , Enc(SimdRm)          , O_660F00(54,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Andps           , "andps"           , Enc(SimdRm)          , O_000F00(54,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Bextr           , "bextr"           , Enc(AvxRmv_OptW)     , O_000F38(F7,U,_,_,_), U                   , F(RW)                              , EF(WUWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , O(Gqd)            , U                 , U                 ),
  INST(Blcfill         , "blcfill"         , Enc(XopVm_OptW)      , O_00_M09(01,1,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blci            , "blci"            , Enc(XopVm_OptW)      , O_00_M09(02,6,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blcic           , "blcic"           , Enc(XopVm_OptW)      , O_00_M09(01,5,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blcmsk          , "blcmsk"          , Enc(XopVm_OptW)      , O_00_M09(02,1,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blcs            , "blcs"            , Enc(XopVm_OptW)      , O_00_M09(01,3,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blendpd         , "blendpd"         , Enc(SimdRmi)         , O_660F3A(0D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Blendps         , "blendps"         , Enc(SimdRmi)         , O_660F3A(0C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Blendvpd        , "blendvpd"        , Enc(SimdRm)          , O_660F38(15,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Blendvps        , "blendvps"        , Enc(SimdRm)          , O_660F38(14,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Blsfill         , "blsfill"         , Enc(XopVm_OptW)      , O_00_M09(01,2,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blsi            , "blsi"            , Enc(AvxVm_OptW)      , O_000F38(F3,3,_,_,_), U                   , F(RW)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blsic           , "blsic"           , Enc(XopVm_OptW)      , O_00_M09(01,6,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blsmsk          , "blsmsk"          , Enc(AvxVm_OptW)      , O_000F38(F3,2,_,_,_), U                   , F(RW)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Blsr            , "blsr"            , Enc(AvxVm_OptW)      , O_000F38(F3,1,_,_,_), U                   , F(RW)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Bsf             , "bsf"             , Enc(X86RegRm)        , O_000F00(BC,U,_,_,_), U                   , F(RW)                              , EF(UUWUUU__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Bsr             , "bsr"             , Enc(X86RegRm)        , O_000F00(BD,U,_,_,_), U                   , F(RW)                              , EF(UUWUUU__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Bswap           , "bswap"           , Enc(X86BSwap)        , O_000F00(C8,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Gqd)            , U                 , U                 , U                 , U                 ),
  INST(Bt              , "bt"              , Enc(X86BTest)        , O_000F00(A3,U,_,_,_), O_000F00(BA,4,_,_,_), F(RO)                              , EF(UU_UUW__), 0 , 0 , O(GqdwMem)        , O(Gqdw)|O(Imm)    , U                 , U                 , U                 ),
  INST(Btc             , "btc"             , Enc(X86BTest)        , O_000F00(BB,U,_,_,_), O_000F00(BA,7,_,_,_), F(RW)|F(Lock)                      , EF(UU_UUW__), 0 , 0 , O(GqdwMem)        , O(Gqdw)|O(Imm)    , U                 , U                 , U                 ),
  INST(Btr             , "btr"             , Enc(X86BTest)        , O_000F00(B3,U,_,_,_), O_000F00(BA,6,_,_,_), F(RW)|F(Lock)                      , EF(UU_UUW__), 0 , 0 , O(GqdwMem)        , O(Gqdw)|O(Imm)    , U                 , U                 , U                 ),
  INST(Bts             , "bts"             , Enc(X86BTest)        , O_000F00(AB,U,_,_,_), O_000F00(BA,5,_,_,_), F(RW)|F(Lock)                      , EF(UU_UUW__), 0 , 0 , O(GqdwMem)        , O(Gqdw)|O(Imm)    , U                 , U                 , U                 ),
  INST(Bzhi            , "bzhi"            , Enc(AvxRmv_OptW)     , O_000F38(F5,U,_,_,_), U                   , F(RW)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , O(Gqd)            , U                 , U                 ),
  INST(Call            , "call"            , Enc(X86Call)         , O_000000(FF,2,_,_,_), O_000000(E8,U,_,_,_), F(RW)|F(Flow)|F(Volatile)          , EF(________), 0 , 0 , O(GqdMem)|O(LImm) , U                 , U                 , U                 , U                 ),
  INST(Cbw             , "cbw"             , Enc(X86Op)           , O_660000(98,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Cdq             , "cdq"             , Enc(X86Op)           , O_000000(99,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Cdqe            , "cdqe"            , Enc(X86Op)           , O_000000(98,U,_,W,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Clc             , "clc"             , Enc(X86Op)           , O_000000(F8,U,_,_,_), U                   , F(Volatile)                        , EF(_____W__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Cld             , "cld"             , Enc(X86Op)           , O_000000(FC,U,_,_,_), U                   , F(Volatile)                        , EF(______W_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Clflush         , "clflush"         , Enc(X86M)            , O_000F00(AE,7,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Clflushopt      , "clflushopt"      , Enc(X86M)            , O_660F00(AE,7,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Cmc             , "cmc"             , Enc(X86Op)           , O_000000(F5,U,_,_,_), U                   , F(None)                            , EF(_____X__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Cmova           , "cmova"           , Enc(X86RegRm)        , O_000F00(47,U,_,_,_), U                   , F(RW)                              , EF(__R__R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovae          , "cmovae"          , Enc(X86RegRm)        , O_000F00(43,U,_,_,_), U                   , F(RW)                              , EF(_____R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovb           , "cmovb"           , Enc(X86RegRm)        , O_000F00(42,U,_,_,_), U                   , F(RW)                              , EF(_____R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovbe          , "cmovbe"          , Enc(X86RegRm)        , O_000F00(46,U,_,_,_), U                   , F(RW)                              , EF(__R__R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovc           , "cmovc"           , Enc(X86RegRm)        , O_000F00(42,U,_,_,_), U                   , F(RW)                              , EF(_____R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmove           , "cmove"           , Enc(X86RegRm)        , O_000F00(44,U,_,_,_), U                   , F(RW)                              , EF(__R_____), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovg           , "cmovg"           , Enc(X86RegRm)        , O_000F00(4F,U,_,_,_), U                   , F(RW)                              , EF(RRR_____), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovge          , "cmovge"          , Enc(X86RegRm)        , O_000F00(4D,U,_,_,_), U                   , F(RW)                              , EF(RR______), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovl           , "cmovl"           , Enc(X86RegRm)        , O_000F00(4C,U,_,_,_), U                   , F(RW)                              , EF(RR______), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovle          , "cmovle"          , Enc(X86RegRm)        , O_000F00(4E,U,_,_,_), U                   , F(RW)                              , EF(RRR_____), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovna          , "cmovna"          , Enc(X86RegRm)        , O_000F00(46,U,_,_,_), U                   , F(RW)                              , EF(__R__R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnae         , "cmovnae"         , Enc(X86RegRm)        , O_000F00(42,U,_,_,_), U                   , F(RW)                              , EF(_____R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnb          , "cmovnb"          , Enc(X86RegRm)        , O_000F00(43,U,_,_,_), U                   , F(RW)                              , EF(_____R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnbe         , "cmovnbe"         , Enc(X86RegRm)        , O_000F00(47,U,_,_,_), U                   , F(RW)                              , EF(__R__R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnc          , "cmovnc"          , Enc(X86RegRm)        , O_000F00(43,U,_,_,_), U                   , F(RW)                              , EF(_____R__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovne          , "cmovne"          , Enc(X86RegRm)        , O_000F00(45,U,_,_,_), U                   , F(RW)                              , EF(__R_____), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovng          , "cmovng"          , Enc(X86RegRm)        , O_000F00(4E,U,_,_,_), U                   , F(RW)                              , EF(RRR_____), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnge         , "cmovnge"         , Enc(X86RegRm)        , O_000F00(4C,U,_,_,_), U                   , F(RW)                              , EF(RR______), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnl          , "cmovnl"          , Enc(X86RegRm)        , O_000F00(4D,U,_,_,_), U                   , F(RW)                              , EF(RR______), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnle         , "cmovnle"         , Enc(X86RegRm)        , O_000F00(4F,U,_,_,_), U                   , F(RW)                              , EF(RRR_____), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovno          , "cmovno"          , Enc(X86RegRm)        , O_000F00(41,U,_,_,_), U                   , F(RW)                              , EF(R_______), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnp          , "cmovnp"          , Enc(X86RegRm)        , O_000F00(4B,U,_,_,_), U                   , F(RW)                              , EF(____R___), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovns          , "cmovns"          , Enc(X86RegRm)        , O_000F00(49,U,_,_,_), U                   , F(RW)                              , EF(_R______), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovnz          , "cmovnz"          , Enc(X86RegRm)        , O_000F00(45,U,_,_,_), U                   , F(RW)                              , EF(__R_____), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovo           , "cmovo"           , Enc(X86RegRm)        , O_000F00(40,U,_,_,_), U                   , F(RW)                              , EF(R_______), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovp           , "cmovp"           , Enc(X86RegRm)        , O_000F00(4A,U,_,_,_), U                   , F(RW)                              , EF(____R___), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovpe          , "cmovpe"          , Enc(X86RegRm)        , O_000F00(4A,U,_,_,_), U                   , F(RW)                              , EF(____R___), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovpo          , "cmovpo"          , Enc(X86RegRm)        , O_000F00(4B,U,_,_,_), U                   , F(RW)                              , EF(____R___), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovs           , "cmovs"           , Enc(X86RegRm)        , O_000F00(48,U,_,_,_), U                   , F(RW)                              , EF(_R______), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmovz           , "cmovz"           , Enc(X86RegRm)        , O_000F00(44,U,_,_,_), U                   , F(RW)                              , EF(__R_____), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Cmp             , "cmp"             , Enc(X86Arith)        , O_000000(38,7,_,_,_), U                   , F(RO)                              , EF(WWWWWW__), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(Cmppd           , "cmppd"           , Enc(SimdRmi)         , O_660F00(C2,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Cmpps           , "cmpps"           , Enc(SimdRmi)         , O_000F00(C2,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(CmpsB           , "cmps_b"          , Enc(X86Op)           , O_000000(A6,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(CmpsD           , "cmps_d"          , Enc(X86Op)           , O_000000(A7,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(CmpsQ           , "cmps_q"          , Enc(X86Op)           , O_000000(A7,U,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(CmpsW           , "cmps_w"          , Enc(X86Op_66H)       , O_000000(A7,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Cmpsd           , "cmpsd"           , Enc(SimdRmi)         , O_F20F00(C2,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Cmpss           , "cmpss"           , Enc(SimdRmi)         , O_F30F00(C2,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Cmpxchg         , "cmpxchg"         , Enc(X86RmReg)        , O_000F00(B0,U,_,_,_), U                   , F(RW)|F(Lock)|F(Special)           , EF(WWWWWW__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Cmpxchg16b      , "cmpxchg16b"      , Enc(X86M)            , O_000F00(C7,1,_,W,_), U                   , F(RW)|F(Lock)|F(Special)           , EF(__W_____), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Cmpxchg8b       , "cmpxchg8b"       , Enc(X86M)            , O_000F00(C7,1,_,_,_), U                   , F(RW)|F(Lock)|F(Special)           , EF(__W_____), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Comisd          , "comisd"          , Enc(SimdRm)          , O_660F00(2F,U,_,_,_), U                   , F(RO)                              , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Comiss          , "comiss"          , Enc(SimdRm)          , O_000F00(2F,U,_,_,_), U                   , F(RO)                              , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cpuid           , "cpuid"           , Enc(X86Op)           , O_000F00(A2,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Cqo             , "cqo"             , Enc(X86Op)           , O_000000(99,U,_,W,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Crc32           , "crc32"           , Enc(X86Crc)          , O_F20F38(F0,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Gqd)            , O(GqdwbMem)       , U                 , U                 , U                 ),
  INST(Cvtdq2pd        , "cvtdq2pd"        , Enc(SimdRm)          , O_F30F00(E6,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtdq2ps        , "cvtdq2ps"        , Enc(SimdRm)          , O_000F00(5B,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtpd2dq        , "cvtpd2dq"        , Enc(SimdRm)          , O_F20F00(E6,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtpd2pi        , "cvtpd2pi"        , Enc(SimdRm)          , O_660F00(2D,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtpd2ps        , "cvtpd2ps"        , Enc(SimdRm)          , O_660F00(5A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtpi2pd        , "cvtpi2pd"        , Enc(SimdRm)          , O_660F00(2A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(MmMem)          , U                 , U                 , U                 ),
  INST(Cvtpi2ps        , "cvtpi2ps"        , Enc(SimdRm)          , O_000F00(2A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Xmm)            , O(MmMem)          , U                 , U                 , U                 ),
  INST(Cvtps2dq        , "cvtps2dq"        , Enc(SimdRm)          , O_660F00(5B,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtps2pd        , "cvtps2pd"        , Enc(SimdRm)          , O_000F00(5A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtps2pi        , "cvtps2pi"        , Enc(SimdRm)          , O_000F00(2D,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtsd2si        , "cvtsd2si"        , Enc(SimdRm_Q)        , O_F20F00(2D,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtsd2ss        , "cvtsd2ss"        , Enc(SimdRm)          , O_F20F00(5A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 4 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtsi2sd        , "cvtsi2sd"        , Enc(SimdRm_Q)        , O_F20F00(2A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Xmm)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Cvtsi2ss        , "cvtsi2ss"        , Enc(SimdRm_Q)        , O_F30F00(2A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 4 , O(Xmm)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Cvtss2sd        , "cvtss2sd"        , Enc(SimdRm)          , O_F30F00(5A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvtss2si        , "cvtss2si"        , Enc(SimdRm_Q)        , O_F30F00(2D,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvttpd2dq       , "cvttpd2dq"       , Enc(SimdRm)          , O_660F00(E6,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvttpd2pi       , "cvttpd2pi"       , Enc(SimdRm)          , O_660F00(2C,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvttps2dq       , "cvttps2dq"       , Enc(SimdRm)          , O_F30F00(5B,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvttps2pi       , "cvttps2pi"       , Enc(SimdRm)          , O_000F00(2C,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvttsd2si       , "cvttsd2si"       , Enc(SimdRm_Q)        , O_F20F00(2C,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cvttss2si       , "cvttss2si"       , Enc(SimdRm_Q)        , O_F30F00(2C,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Cwd             , "cwd"             , Enc(X86Op)           , O_660000(99,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Cwde            , "cwde"            , Enc(X86Op)           , O_000000(98,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Daa             , "daa"             , Enc(X86Op)           , O_000000(27,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(UWWXWX__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Das             , "das"             , Enc(X86Op)           , O_000000(2F,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(UWWXWX__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Dec             , "dec"             , Enc(X86IncDec)       , O_000000(FE,1,_,_,_), O_000000(48,U,_,_,_), F(RW)|F(Lock)                      , EF(WWWWW___), 0 , 0 , O(GqdwbMem)       , U                 , U                 , U                 , U                 ),
  INST(Div             , "div"             , Enc(X86Rm_B)         , O_000000(F6,6,_,_,_), U                   , F(RW)|F(Special)                   , EF(UUUUUU__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Divpd           , "divpd"           , Enc(SimdRm)          , O_660F00(5E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Divps           , "divps"           , Enc(SimdRm)          , O_000F00(5E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Divsd           , "divsd"           , Enc(SimdRm)          , O_F20F00(5E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Divss           , "divss"           , Enc(SimdRm)          , O_F30F00(5E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Dppd            , "dppd"            , Enc(SimdRmi)         , O_660F3A(41,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Dpps            , "dpps"            , Enc(SimdRmi)         , O_660F3A(40,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Emms            , "emms"            , Enc(X86Op)           , O_000F00(77,U,_,_,_), U                   , F(Volatile)                        , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Enter           , "enter"           , Enc(X86Enter)        , O_000000(C8,U,_,_,_), U                   , F(Volatile)|F(Special)             , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Extractps       , "extractps"       , Enc(SimdExtract)     , O_660F3A(17,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(GqdMem)         , O(Xmm)            , U                 , U                 , U                 ),
  INST(Extrq           , "extrq"           , Enc(SimdExtrq)       , O_660F00(79,U,_,_,_), O_660F00(78,0,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)|O(Imm)     , O(None)|O(Imm)    , U                 , U                 ),
  INST(F2xm1           , "f2xm1"           , Enc(FpuOp)           , O_00_X(D9F0,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fabs            , "fabs"            , Enc(FpuOp)           , O_00_X(D9E1,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fadd            , "fadd"            , Enc(FpuArith)        , O_00_X(C0C0,0)      , U                   , F(Fp)|F(Mem4_8)                    , EF(________), 0 , 0 , O(FpMem)          , O(Fp)             , U                 , U                 , U                 ),
  INST(Faddp           , "faddp"           , Enc(FpuRDef)         , O_00_X(DEC0,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fbld            , "fbld"            , Enc(X86M)            , O_000000(DF,4,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fbstp           , "fbstp"           , Enc(X86M)            , O_000000(DF,6,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fchs            , "fchs"            , Enc(FpuOp)           , O_00_X(D9E0,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fclex           , "fclex"           , Enc(FpuOp)           , O_9B_X(DBE2,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fcmovb          , "fcmovb"          , Enc(FpuR)            , O_00_X(DAC0,U)      , U                   , F(Fp)                              , EF(_____R__), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcmovbe         , "fcmovbe"         , Enc(FpuR)            , O_00_X(DAD0,U)      , U                   , F(Fp)                              , EF(__R__R__), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcmove          , "fcmove"          , Enc(FpuR)            , O_00_X(DAC8,U)      , U                   , F(Fp)                              , EF(__R_____), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcmovnb         , "fcmovnb"         , Enc(FpuR)            , O_00_X(DBC0,U)      , U                   , F(Fp)                              , EF(_____R__), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcmovnbe        , "fcmovnbe"        , Enc(FpuR)            , O_00_X(DBD0,U)      , U                   , F(Fp)                              , EF(__R__R__), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcmovne         , "fcmovne"         , Enc(FpuR)            , O_00_X(DBC8,U)      , U                   , F(Fp)                              , EF(__R_____), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcmovnu         , "fcmovnu"         , Enc(FpuR)            , O_00_X(DBD8,U)      , U                   , F(Fp)                              , EF(____R___), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcmovu          , "fcmovu"          , Enc(FpuR)            , O_00_X(DAD8,U)      , U                   , F(Fp)                              , EF(____R___), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcom            , "fcom"            , Enc(FpuCom)          , O_00_X(D0D0,2)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)|O(Mem)      , O(Fp)             , U                 , U                 , U                 ),
  INST(Fcomi           , "fcomi"           , Enc(FpuR)            , O_00_X(DBF0,U)      , U                   , F(Fp)                              , EF(WWWWWW__), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcomip          , "fcomip"          , Enc(FpuR)            , O_00_X(DFF0,U)      , U                   , F(Fp)                              , EF(WWWWWW__), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fcomp           , "fcomp"           , Enc(FpuCom)          , O_00_X(D8D8,3)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)|O(Mem)      , O(Fp)             , U                 , U                 , U                 ),
  INST(Fcompp          , "fcompp"          , Enc(FpuOp)           , O_00_X(DED9,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fcos            , "fcos"            , Enc(FpuOp)           , O_00_X(D9FF,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fdecstp         , "fdecstp"         , Enc(FpuOp)           , O_00_X(D9F6,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fdiv            , "fdiv"            , Enc(FpuArith)        , O_00_X(F0F8,6)      , U                   , F(Fp)|F(Mem4_8)                    , EF(________), 0 , 0 , O(FpMem)          , O(Fp)             , U                 , U                 , U                 ),
  INST(Fdivp           , "fdivp"           , Enc(FpuRDef)         , O_00_X(DEF8,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fdivr           , "fdivr"           , Enc(FpuArith)        , O_00_X(F8F0,7)      , U                   , F(Fp)|F(Mem4_8)                    , EF(________), 0 , 0 , O(FpMem)          , O(Fp)             , U                 , U                 , U                 ),
  INST(Fdivrp          , "fdivrp"          , Enc(FpuRDef)         , O_00_X(DEF0,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Femms           , "femms"           , Enc(X86Op)           , O_000F00(0E,U,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Ffree           , "ffree"           , Enc(FpuR)            , O_00_X(DDC0,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fiadd           , "fiadd"           , Enc(FpuM)            , O_000000(DA,0,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Ficom           , "ficom"           , Enc(FpuM)            , O_000000(DA,2,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Ficomp          , "ficomp"          , Enc(FpuM)            , O_000000(DA,3,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fidiv           , "fidiv"           , Enc(FpuM)            , O_000000(DA,6,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fidivr          , "fidivr"          , Enc(FpuM)            , O_000000(DA,7,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fild            , "fild"            , Enc(FpuM)            , O_000000(DB,0,_,_,_), O_000000(DF,5,_,_,_), F(Fp)|F(Mem2_4_8)                  , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fimul           , "fimul"           , Enc(FpuM)            , O_000000(DA,1,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fincstp         , "fincstp"         , Enc(FpuOp)           , O_00_X(D9F7,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Finit           , "finit"           , Enc(FpuOp)           , O_9B_X(DBE3,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fist            , "fist"            , Enc(FpuM)            , O_000000(DB,2,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fistp           , "fistp"           , Enc(FpuM)            , O_000000(DB,3,_,_,_), O_000000(DF,7,_,_,_), F(Fp)|F(Mem2_4_8)                  , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fisttp          , "fisttp"          , Enc(FpuM)            , O_000000(DB,1,_,_,_), O_000000(DD,1,_,_,_), F(Fp)|F(Mem2_4_8)                  , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fisub           , "fisub"           , Enc(FpuM)            , O_000000(DA,4,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fisubr          , "fisubr"          , Enc(FpuM)            , O_000000(DA,5,_,_,_), U                   , F(Fp)|F(Mem2_4)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fld             , "fld"             , Enc(FpuFldFst)       , O_000000(D9,0,_,_,_), O_000000(DB,5,_,_,_), F(Fp)|F(Mem4_8_10)                 , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fld1            , "fld1"            , Enc(FpuOp)           , O_00_X(D9E8,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fldcw           , "fldcw"           , Enc(X86M)            , O_000000(D9,5,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fldenv          , "fldenv"          , Enc(X86M)            , O_000000(D9,4,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fldl2e          , "fldl2e"          , Enc(FpuOp)           , O_00_X(D9EA,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fldl2t          , "fldl2t"          , Enc(FpuOp)           , O_00_X(D9E9,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fldlg2          , "fldlg2"          , Enc(FpuOp)           , O_00_X(D9EC,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fldln2          , "fldln2"          , Enc(FpuOp)           , O_00_X(D9ED,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fldpi           , "fldpi"           , Enc(FpuOp)           , O_00_X(D9EB,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fldz            , "fldz"            , Enc(FpuOp)           , O_00_X(D9EE,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fmul            , "fmul"            , Enc(FpuArith)        , O_00_X(C8C8,1)      , U                   , F(Fp)|F(Mem4_8)                    , EF(________), 0 , 0 , O(FpMem)          , O(Fp)             , U                 , U                 , U                 ),
  INST(Fmulp           , "fmulp"           , Enc(FpuRDef)         , O_00_X(DEC8,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fnclex          , "fnclex"          , Enc(FpuOp)           , O_00_X(DBE2,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fninit          , "fninit"          , Enc(FpuOp)           , O_00_X(DBE3,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fnop            , "fnop"            , Enc(FpuOp)           , O_00_X(D9D0,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fnsave          , "fnsave"          , Enc(X86M)            , O_000000(DD,6,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fnstcw          , "fnstcw"          , Enc(X86M)            , O_000000(D9,7,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fnstenv         , "fnstenv"         , Enc(X86M)            , O_000000(D9,6,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fnstsw          , "fnstsw"          , Enc(FpuStsw)         , O_000000(DD,7,_,_,_), O_00_X(DFE0,U)      , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fpatan          , "fpatan"          , Enc(FpuOp)           , O_00_X(D9F3,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fprem           , "fprem"           , Enc(FpuOp)           , O_00_X(D9F8,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fprem1          , "fprem1"          , Enc(FpuOp)           , O_00_X(D9F5,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fptan           , "fptan"           , Enc(FpuOp)           , O_00_X(D9F2,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Frndint         , "frndint"         , Enc(FpuOp)           , O_00_X(D9FC,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Frstor          , "frstor"          , Enc(X86M)            , O_000000(DD,4,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fsave           , "fsave"           , Enc(X86M)            , O_9B0000(DD,6,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fscale          , "fscale"          , Enc(FpuOp)           , O_00_X(D9FD,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fsin            , "fsin"            , Enc(FpuOp)           , O_00_X(D9FE,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fsincos         , "fsincos"         , Enc(FpuOp)           , O_00_X(D9FB,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fsqrt           , "fsqrt"           , Enc(FpuOp)           , O_00_X(D9FA,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fst             , "fst"             , Enc(FpuFldFst)       , O_000000(D9,2,_,_,_), U                   , F(Fp)|F(Mem4_8)                    , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fstcw           , "fstcw"           , Enc(X86M)            , O_9B0000(D9,7,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fstenv          , "fstenv"          , Enc(X86M)            , O_9B0000(D9,6,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fstp            , "fstp"            , Enc(FpuFldFst)       , O_000000(D9,3,_,_,_), O_000000(DB,7,_,_,_), F(Fp)|F(Mem4_8_10)                 , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fstsw           , "fstsw"           , Enc(FpuStsw)         , O_9B0000(DD,7,_,_,_), O_9B_X(DFE0,U)      , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fsub            , "fsub"            , Enc(FpuArith)        , O_00_X(E0E8,4)      , U                   , F(Fp)|F(Mem4_8)                    , EF(________), 0 , 0 , O(FpMem)          , O(Fp)             , U                 , U                 , U                 ),
  INST(Fsubp           , "fsubp"           , Enc(FpuRDef)         , O_00_X(DEE8,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fsubr           , "fsubr"           , Enc(FpuArith)        , O_00_X(E8E0,5)      , U                   , F(Fp)|F(Mem4_8)                    , EF(________), 0 , 0 , O(FpMem)          , O(Fp)             , U                 , U                 , U                 ),
  INST(Fsubrp          , "fsubrp"          , Enc(FpuRDef)         , O_00_X(DEE0,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Ftst            , "ftst"            , Enc(FpuOp)           , O_00_X(D9E4,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fucom           , "fucom"           , Enc(FpuRDef)         , O_00_X(DDE0,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fucomi          , "fucomi"          , Enc(FpuR)            , O_00_X(DBE8,U)      , U                   , F(Fp)                              , EF(WWWWWW__), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fucomip         , "fucomip"         , Enc(FpuR)            , O_00_X(DFE8,U)      , U                   , F(Fp)                              , EF(WWWWWW__), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fucomp          , "fucomp"          , Enc(FpuRDef)         , O_00_X(DDE8,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fucompp         , "fucompp"         , Enc(FpuOp)           , O_00_X(DAE9,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fwait           , "fwait"           , Enc(X86Op)           , O_000000(DB,U,_,_,_), U                   , F(Fp)|F(Volatile)                  , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fxam            , "fxam"            , Enc(FpuOp)           , O_00_X(D9E5,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fxch            , "fxch"            , Enc(FpuR)            , O_00_X(D9C8,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , O(Fp)             , U                 , U                 , U                 , U                 ),
  INST(Fxrstor         , "fxrstor"         , Enc(X86M)            , O_000F00(AE,1,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fxrstor64       , "fxrstor64"       , Enc(X86M)            , O_000F00(AE,1,_,W,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fxsave          , "fxsave"          , Enc(X86M)            , O_000F00(AE,0,_,_,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fxsave64        , "fxsave64"        , Enc(X86M)            , O_000F00(AE,0,_,W,_), U                   , F(Fp)                              , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Fxtract         , "fxtract"         , Enc(FpuOp)           , O_00_X(D9F4,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fyl2x           , "fyl2x"           , Enc(FpuOp)           , O_00_X(D9F1,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Fyl2xp1         , "fyl2xp1"         , Enc(FpuOp)           , O_00_X(D9F9,U)      , U                   , F(Fp)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Haddpd          , "haddpd"          , Enc(SimdRm)          , O_660F00(7C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Haddps          , "haddps"          , Enc(SimdRm)          , O_F20F00(7C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Hsubpd          , "hsubpd"          , Enc(SimdRm)          , O_660F00(7D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Hsubps          , "hsubps"          , Enc(SimdRm)          , O_F20F00(7D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Idiv            , "idiv"            , Enc(X86Rm_B)         , O_000000(F6,7,_,_,_), U                   , F(RW)|F(Special)                   , EF(UUUUUU__), 0 , 0 , 0                 , 0                 , U                 , U                 , U                 ),
  INST(Imul            , "imul"            , Enc(X86Imul)         , U                   , U                   , F(RW)|F(Special)                   , EF(WUUUUW__), 0 , 0 , 0                 , 0                 , U                 , U                 , U                 ),
  INST(Inc             , "inc"             , Enc(X86IncDec)       , O_000000(FE,0,_,_,_), O_000000(40,U,_,_,_), F(RW)|F(Lock)                      , EF(WWWWW___), 0 , 0 , O(GqdwbMem)       , U                 , U                 , U                 , U                 ),
  INST(Insertps        , "insertps"        , Enc(SimdRmi)         , O_660F3A(21,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Insertq         , "insertq"         , Enc(SimdInsertq)     , O_F20F00(79,U,_,_,_), O_F20F00(78,U,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(None)|O(Imm)    , O(None)|O(Imm)    , U                 ),
  INST(Int             , "int"             , Enc(X86Int)          , O_000000(CC,U,_,_,_), U                   , F(Volatile)                        , EF(_______W), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Ja              , "ja"              , Enc(X86Jcc)          , O_000000(77,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(__R__R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jae             , "jae"             , Enc(X86Jcc)          , O_000000(73,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(_____R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jb              , "jb"              , Enc(X86Jcc)          , O_000000(72,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(_____R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jbe             , "jbe"             , Enc(X86Jcc)          , O_000000(76,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(__R__R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jc              , "jc"              , Enc(X86Jcc)          , O_000000(72,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(_____R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Je              , "je"              , Enc(X86Jcc)          , O_000000(74,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(__R_____), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jg              , "jg"              , Enc(X86Jcc)          , O_000000(7F,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(RRR_____), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jge             , "jge"             , Enc(X86Jcc)          , O_000000(7D,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(RR______), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jl              , "jl"              , Enc(X86Jcc)          , O_000000(7C,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(RR______), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jle             , "jle"             , Enc(X86Jcc)          , O_000000(7E,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(RRR_____), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jna             , "jna"             , Enc(X86Jcc)          , O_000000(76,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(__R__R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnae            , "jnae"            , Enc(X86Jcc)          , O_000000(72,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(_____R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnb             , "jnb"             , Enc(X86Jcc)          , O_000000(73,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(_____R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnbe            , "jnbe"            , Enc(X86Jcc)          , O_000000(77,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(__R__R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnc             , "jnc"             , Enc(X86Jcc)          , O_000000(73,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(_____R__), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jne             , "jne"             , Enc(X86Jcc)          , O_000000(75,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(__R_____), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jng             , "jng"             , Enc(X86Jcc)          , O_000000(7E,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(RRR_____), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnge            , "jnge"            , Enc(X86Jcc)          , O_000000(7C,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(RR______), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnl             , "jnl"             , Enc(X86Jcc)          , O_000000(7D,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(RR______), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnle            , "jnle"            , Enc(X86Jcc)          , O_000000(7F,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(RRR_____), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jno             , "jno"             , Enc(X86Jcc)          , O_000000(71,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(R_______), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnp             , "jnp"             , Enc(X86Jcc)          , O_000000(7B,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(____R___), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jns             , "jns"             , Enc(X86Jcc)          , O_000000(79,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(_R______), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jnz             , "jnz"             , Enc(X86Jcc)          , O_000000(75,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(__R_____), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jo              , "jo"              , Enc(X86Jcc)          , O_000000(70,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(R_______), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jp              , "jp"              , Enc(X86Jcc)          , O_000000(7A,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(____R___), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jpe             , "jpe"             , Enc(X86Jcc)          , O_000000(7A,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(____R___), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jpo             , "jpo"             , Enc(X86Jcc)          , O_000000(7B,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(____R___), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Js              , "js"              , Enc(X86Jcc)          , O_000000(78,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(_R______), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jz              , "jz"              , Enc(X86Jcc)          , O_000000(74,U,_,_,_), U                   , F(Flow)|F(Volatile)                , EF(__R_____), 0 , 0 , O(Label)          , U                 , U                 , U                 , U                 ),
  INST(Jecxz           , "jecxz"           , Enc(X86Jecxz)        , O_000000(E3,U,_,_,_), U                   , F(Flow)|F(Volatile)|F(Special)     , EF(________), 0 , 0 , O(Gqdw)           , O(Label)          , U                 , U                 , U                 ),
  INST(Jmp             , "jmp"             , Enc(X86Jmp)          , O_000000(FF,4,_,_,_), O_000000(E9,U,_,_,_), F(Flow)|F(Volatile)                , EF(________), 0 , 0 , O(Label)|O(Imm)   , U                 , U                 , U                 , U                 ),
  INST(Lahf            , "lahf"            , Enc(X86Op)           , O_000000(9F,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(_RRRRR__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Lddqu           , "lddqu"           , Enc(SimdRm)          , O_F20F00(F0,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(Mem)            , U                 , U                 , U                 ),
  INST(Ldmxcsr         , "ldmxcsr"         , Enc(X86M)            , O_000F00(AE,2,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Lea             , "lea"             , Enc(X86Lea)          , O_000000(8D,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqd)            , O(Mem)            , U                 , U                 , U                 ),
  INST(Leave           , "leave"           , Enc(X86Op)           , O_000000(C9,U,_,_,_), U                   , F(Volatile)|F(Special)             , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Lfence          , "lfence"          , Enc(X86Fence)        , O_000F00(AE,5,_,_,_), U                   , F(Volatile)                        , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(LodsB           , "lods_b"          , Enc(X86Op)           , O_000000(AC,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(______R_), 0 , 1 , U                 , U                 , U                 , U                 , U                 ),
  INST(LodsD           , "lods_d"          , Enc(X86Op)           , O_000000(AD,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(______R_), 0 , 4 , U                 , U                 , U                 , U                 , U                 ),
  INST(LodsQ           , "lods_q"          , Enc(X86Op)           , O_000000(AD,U,_,W,_), U                   , F(WO)|F(Special)                   , EF(______R_), 0 , 8 , U                 , U                 , U                 , U                 , U                 ),
  INST(LodsW           , "lods_w"          , Enc(X86Op_66H)       , O_000000(AD,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(______R_), 0 , 2 , U                 , U                 , U                 , U                 , U                 ),
  INST(Lzcnt           , "lzcnt"           , Enc(X86RegRm)        , O_F30F00(BD,U,_,_,_), U                   , F(RW)                              , EF(UUWUUW__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Maskmovdqu      , "maskmovdqu"      , Enc(SimdRm)          , O_660F00(57,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Maskmovq        , "maskmovq"        , Enc(SimdRm)          , O_000F00(F7,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , O(Mm)             , O(Mm)             , U                 , U                 , U                 ),
  INST(Maxpd           , "maxpd"           , Enc(SimdRm)          , O_660F00(5F,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Maxps           , "maxps"           , Enc(SimdRm)          , O_000F00(5F,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Maxsd           , "maxsd"           , Enc(SimdRm)          , O_F20F00(5F,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Maxss           , "maxss"           , Enc(SimdRm)          , O_F30F00(5F,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Mfence          , "mfence"          , Enc(X86Fence)        , O_000F00(AE,6,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Minpd           , "minpd"           , Enc(SimdRm)          , O_660F00(5D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Minps           , "minps"           , Enc(SimdRm)          , O_000F00(5D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Minsd           , "minsd"           , Enc(SimdRm)          , O_F20F00(5D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Minss           , "minss"           , Enc(SimdRm)          , O_F30F00(5D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Monitor         , "monitor"         , Enc(X86Op)           , O_000F01(C8,U,_,_,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Mov             , "mov"             , Enc(X86Mov)          , U                   , U                   , F(WO)                              , EF(________), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(MovPtr          , "mov_ptr"         , Enc(X86MovPtr)       , O_000000(A0,U,_,_,_), O_000000(A2,U,_,_,_), F(WO)|F(Special)                   , EF(________), 0 , 0 , O(Gqdwb)          , O(Imm)            , U                 , U                 , U                 ),
  INST(Movapd          , "movapd"          , Enc(SimdMov)         , O_660F00(28,U,_,_,_), O_660F00(29,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movaps          , "movaps"          , Enc(SimdMov)         , O_000F00(28,U,_,_,_), O_000F00(29,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movbe           , "movbe"           , Enc(SimdMovBe)       , O_000F38(F0,U,_,_,_), O_000F38(F1,U,_,_,_), F(WO)                              , EF(________), 0 , 0 , O(GqdwMem)        , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Movd            , "movd"            , Enc(SimdMovD)        , O_000F00(6E,U,_,_,_), O_000F00(7E,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(Gd)|O(MmXmmMem) , O(Gd)|O(MmXmmMem) , U                 , U                 , U                 ),
  INST(Movddup         , "movddup"         , Enc(SimdMov)         , O_F20F00(12,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movdq2q         , "movdq2q"         , Enc(SimdMov)         , O_F20F00(D6,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movdqa          , "movdqa"          , Enc(SimdMov)         , O_660F00(6F,U,_,_,_), O_660F00(7F,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movdqu          , "movdqu"          , Enc(SimdMov)         , O_F30F00(6F,U,_,_,_), O_F30F00(7F,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movhlps         , "movhlps"         , Enc(SimdMov)         , O_000F00(12,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Xmm)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movhpd          , "movhpd"          , Enc(SimdMov)         , O_660F00(16,U,_,_,_), O_660F00(17,U,_,_,_), F(RW)                              , EF(________), 8 , 8 , O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movhps          , "movhps"          , Enc(SimdMov)         , O_000F00(16,U,_,_,_), O_000F00(17,U,_,_,_), F(RW)                              , EF(________), 8 , 8 , O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movlhps         , "movlhps"         , Enc(SimdMov)         , O_000F00(16,U,_,_,_), U                   , F(RW)                              , EF(________), 8 , 8 , O(Xmm)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movlpd          , "movlpd"          , Enc(SimdMov)         , O_660F00(12,U,_,_,_), O_660F00(13,U,_,_,_), F(WO)                              , EF(________), 0 , 8 , O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movlps          , "movlps"          , Enc(SimdMov)         , O_000F00(12,U,_,_,_), O_000F00(13,U,_,_,_), F(WO)                              , EF(________), 0 , 8 , O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movmskpd        , "movmskpd"        , Enc(SimdMovNoRexW)   , O_660F00(50,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movmskps        , "movmskps"        , Enc(SimdMovNoRexW)   , O_000F00(50,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movntdq         , "movntdq"         , Enc(SimdMov)         , U                   , O_660F00(E7,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(Mem)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movntdqa        , "movntdqa"        , Enc(SimdMov)         , O_660F38(2A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(Mem)            , U                 , U                 , U                 ),
  INST(Movnti          , "movnti"          , Enc(SimdMov)         , U                   , O_000F00(C3,U,_,_,_), F(WO)                              , EF(________), 0 , 8 , O(Mem)            , O(Gqd)            , U                 , U                 , U                 ),
  INST(Movntpd         , "movntpd"         , Enc(SimdMov)         , U                   , O_660F00(2B,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(Mem)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movntps         , "movntps"         , Enc(SimdMov)         , U                   , O_000F00(2B,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(Mem)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movntq          , "movntq"          , Enc(SimdMov)         , U                   , O_000F00(E7,U,_,_,_), F(WO)                              , EF(________), 0 , 8 , O(Mem)            , O(Mm)             , U                 , U                 , U                 ),
  INST(Movntsd         , "movntsd"         , Enc(SimdMov)         , U                   , O_F20F00(2B,U,_,_,_), F(WO)                              , EF(________), 0 , 8 , O(Mem)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movntss         , "movntss"         , Enc(SimdMov)         , U                   , O_F30F00(2B,U,_,_,_), F(WO)                              , EF(________), 0 , 4 , O(Mem)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Movq            , "movq"            , Enc(SimdMovQ)        , O_000F00(6E,U,_,W,_), O_000F00(7E,U,_,W,_), F(WO)                              , EF(________), 0 , 16, O(Gq)|O(MmXmmMem) , O(Gq)|O(MmXmmMem) , U                 , U                 , U                 ),
  INST(Movq2dq         , "movq2dq"         , Enc(SimdRm)          , O_F30F00(D6,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(Mm)             , U                 , U                 , U                 ),
  INST(MovsB           , "movs_b"          , Enc(X86Op)           , O_000000(A4,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(MovsD           , "movs_d"          , Enc(X86Op)           , O_000000(A5,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(MovsQ           , "movs_q"          , Enc(X86Op)           , O_000000(A5,U,_,W,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(MovsW           , "movs_w"          , Enc(X86Op_66H)       , O_000000(A5,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Movsd           , "movsd"           , Enc(SimdMov)         , O_F20F00(10,U,_,_,_), O_F20F00(11,U,_,_,_), F(WO)|F(ZeroIfMem)                 , EF(________), 0 , 8 , O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movshdup        , "movshdup"        , Enc(SimdRm)          , O_F30F00(16,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movsldup        , "movsldup"        , Enc(SimdRm)          , O_F30F00(12,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movss           , "movss"           , Enc(SimdMov)         , O_F30F00(10,U,_,_,_), O_F30F00(11,U,_,_,_), F(WO)|F(ZeroIfMem)                 , EF(________), 0 , 4 , O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movsx           , "movsx"           , Enc(X86MovsxMovzx)   , O_000F00(BE,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqdw)           , O(GwbMem)         , U                 , U                 , U                 ),
  INST(Movsxd          , "movsxd"          , Enc(X86Movsxd)       , O_000000(63,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gq)             , O(GdMem)          , U                 , U                 , U                 ),
  INST(Movupd          , "movupd"          , Enc(SimdMov)         , O_660F00(10,U,_,_,_), O_660F00(11,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movups          , "movups"          , Enc(SimdMov)         , O_000F00(10,U,_,_,_), O_000F00(11,U,_,_,_), F(WO)                              , EF(________), 0 , 16, O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Movzx           , "movzx"           , Enc(X86MovsxMovzx)   , O_000F00(B6,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqdw)           , O(GwbMem)         , U                 , U                 , U                 ),
  INST(Mpsadbw         , "mpsadbw"         , Enc(SimdRmi)         , O_660F3A(42,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Mul             , "mul"             , Enc(X86Rm_B)         , O_000000(F6,4,_,_,_), U                   , F(RW)|F(Special)                   , EF(WUUUUW__), 0 , 0 , 0                 , 0                 , U                 , U                 , U                 ),
  INST(Mulpd           , "mulpd"           , Enc(SimdRm)          , O_660F00(59,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Mulps           , "mulps"           , Enc(SimdRm)          , O_000F00(59,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Mulsd           , "mulsd"           , Enc(SimdRm)          , O_F20F00(59,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Mulss           , "mulss"           , Enc(SimdRm)          , O_F30F00(59,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Mulx            , "mulx"            , Enc(AvxRvm_OptW)     , O_F20F38(F6,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Gqd)            , O(Gqd)            , O(GqdMem)         , U                 , U                 ),
  INST(Mwait           , "mwait"           , Enc(X86Op)           , O_000F01(C9,U,_,_,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Neg             , "neg"             , Enc(X86Rm_B)         , O_000000(F6,3,_,_,_), U                   , F(RW)|F(Lock)                      , EF(WWWWWW__), 0 , 0 , O(GqdwbMem)       , U                 , U                 , U                 , U                 ),
  INST(Nop             , "nop"             , Enc(X86Op)           , O_000000(90,U,_,_,_), U                   , F(None)                            , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Not             , "not"             , Enc(X86Rm_B)         , O_000000(F6,2,_,_,_), U                   , F(RW)|F(Lock)                      , EF(________), 0 , 0 , O(GqdwbMem)       , U                 , U                 , U                 , U                 ),
  INST(Or              , "or"              , Enc(X86Arith)        , O_000000(08,1,_,_,_), U                   , F(RW)|F(Lock)                      , EF(WWWUWW__), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(Orpd            , "orpd"            , Enc(SimdRm)          , O_660F00(56,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Orps            , "orps"            , Enc(SimdRm)          , O_000F00(56,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pabsb           , "pabsb"           , Enc(SimdRm_P)        , O_000F38(1C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pabsd           , "pabsd"           , Enc(SimdRm_P)        , O_000F38(1E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pabsw           , "pabsw"           , Enc(SimdRm_P)        , O_000F38(1D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Packssdw        , "packssdw"        , Enc(SimdRm_P)        , O_000F00(6B,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Packsswb        , "packsswb"        , Enc(SimdRm_P)        , O_000F00(63,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Packusdw        , "packusdw"        , Enc(SimdRm)          , O_660F38(2B,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Packuswb        , "packuswb"        , Enc(SimdRm_P)        , O_000F00(67,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Paddb           , "paddb"           , Enc(SimdRm_P)        , O_000F00(FC,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Paddd           , "paddd"           , Enc(SimdRm_P)        , O_000F00(FE,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Paddq           , "paddq"           , Enc(SimdRm_P)        , O_000F00(D4,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Paddsb          , "paddsb"          , Enc(SimdRm_P)        , O_000F00(EC,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Paddsw          , "paddsw"          , Enc(SimdRm_P)        , O_000F00(ED,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Paddusb         , "paddusb"         , Enc(SimdRm_P)        , O_000F00(DC,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Paddusw         , "paddusw"         , Enc(SimdRm_P)        , O_000F00(DD,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Paddw           , "paddw"           , Enc(SimdRm_P)        , O_000F00(FD,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Palignr         , "palignr"         , Enc(SimdRmi_P)       , O_000F3A(0F,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , O(Imm)            , U                 , U                 ),
  INST(Pand            , "pand"            , Enc(SimdRm_P)        , O_000F00(DB,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pandn           , "pandn"           , Enc(SimdRm_P)        , O_000F00(DF,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pause           , "pause"           , Enc(X86Op)           , O_F30000(90,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Pavgb           , "pavgb"           , Enc(SimdRm_P)        , O_000F00(E0,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pavgusb         , "pavgusb"         , Enc(Simd3dNow)       , O_000F0F(BF,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pavgw           , "pavgw"           , Enc(SimdRm_P)        , O_000F00(E3,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pblendvb        , "pblendvb"        , Enc(SimdRm)          , O_660F38(10,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pblendw         , "pblendw"         , Enc(SimdRmi)         , O_660F3A(0E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pclmulqdq       , "pclmulqdq"       , Enc(SimdRmi)         , O_660F3A(44,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pcmpeqb         , "pcmpeqb"         , Enc(SimdRm_P)        , O_000F00(74,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pcmpeqd         , "pcmpeqd"         , Enc(SimdRm_P)        , O_000F00(76,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pcmpeqq         , "pcmpeqq"         , Enc(SimdRm)          , O_660F38(29,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pcmpeqw         , "pcmpeqw"         , Enc(SimdRm_P)        , O_000F00(75,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pcmpestri       , "pcmpestri"       , Enc(SimdRmi)         , O_660F3A(61,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pcmpestrm       , "pcmpestrm"       , Enc(SimdRmi)         , O_660F3A(60,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pcmpgtb         , "pcmpgtb"         , Enc(SimdRm_P)        , O_000F00(64,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pcmpgtd         , "pcmpgtd"         , Enc(SimdRm_P)        , O_000F00(66,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pcmpgtq         , "pcmpgtq"         , Enc(SimdRm)          , O_660F38(37,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pcmpgtw         , "pcmpgtw"         , Enc(SimdRm_P)        , O_000F00(65,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pcmpistri       , "pcmpistri"       , Enc(SimdRmi)         , O_660F3A(63,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pcmpistrm       , "pcmpistrm"       , Enc(SimdRmi)         , O_660F3A(62,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pdep            , "pdep"            , Enc(AvxRvm_OptW)     , O_F20F38(F5,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqd)            , O(Gqd)            , O(GqdMem)         , U                 , U                 ),
  INST(Pext            , "pext"            , Enc(AvxRvm_OptW)     , O_F30F38(F5,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqd)            , O(Gqd)            , O(GqdMem)         , U                 , U                 ),
  INST(Pextrb          , "pextrb"          , Enc(SimdExtract)     , O_000F3A(14,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gd)|O(Gb)|O(Mem), O(Xmm)            , U                 , U                 , U                 ),
  INST(Pextrd          , "pextrd"          , Enc(SimdExtract)     , O_000F3A(16,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(GdMem)          , O(Xmm)            , U                 , U                 , U                 ),
  INST(Pextrq          , "pextrq"          , Enc(SimdExtract)     , O_000F3A(16,U,_,W,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(GqdMem)         , O(Xmm)            , U                 , U                 , U                 ),
  INST(Pextrw          , "pextrw"          , Enc(SimdPextrw)      , O_000F00(C5,U,_,_,_), O_000F3A(15,U,_,_,_), F(WO)                              , EF(________), 0 , 8 , O(GdMem)          , O(MmXmm)          , U                 , U                 , U                 ),
  INST(Pf2id           , "pf2id"           , Enc(Simd3dNow)       , O_000F0F(1D,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pf2iw           , "pf2iw"           , Enc(Simd3dNow)       , O_000F0F(1C,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfacc           , "pfacc"           , Enc(Simd3dNow)       , O_000F0F(AE,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfadd           , "pfadd"           , Enc(Simd3dNow)       , O_000F0F(9E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfcmpeq         , "pfcmpeq"         , Enc(Simd3dNow)       , O_000F0F(B0,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfcmpge         , "pfcmpge"         , Enc(Simd3dNow)       , O_000F0F(90,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfcmpgt         , "pfcmpgt"         , Enc(Simd3dNow)       , O_000F0F(A0,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfmax           , "pfmax"           , Enc(Simd3dNow)       , O_000F0F(A4,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfmin           , "pfmin"           , Enc(Simd3dNow)       , O_000F0F(94,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfmul           , "pfmul"           , Enc(Simd3dNow)       , O_000F0F(B4,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfnacc          , "pfnacc"          , Enc(Simd3dNow)       , O_000F0F(8A,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfpnacc         , "pfpnacc"         , Enc(Simd3dNow)       , O_000F0F(8E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfrcp           , "pfrcp"           , Enc(Simd3dNow)       , O_000F0F(96,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfrcpit1        , "pfrcpit1"        , Enc(Simd3dNow)       , O_000F0F(A6,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfrcpit2        , "pfrcpit2"        , Enc(Simd3dNow)       , O_000F0F(B6,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfrsqit1        , "pfrsqit1"        , Enc(Simd3dNow)       , O_000F0F(A7,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfrsqrt         , "pfrsqrt"         , Enc(Simd3dNow)       , O_000F0F(97,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfsub           , "pfsub"           , Enc(Simd3dNow)       , O_000F0F(9A,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pfsubr          , "pfsubr"          , Enc(Simd3dNow)       , O_000F0F(AA,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Phaddd          , "phaddd"          , Enc(SimdRm_P)        , O_000F38(02,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Phaddsw         , "phaddsw"         , Enc(SimdRm_P)        , O_000F38(03,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Phaddw          , "phaddw"          , Enc(SimdRm_P)        , O_000F38(01,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Phminposuw      , "phminposuw"      , Enc(SimdRm)          , O_660F38(41,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Phsubd          , "phsubd"          , Enc(SimdRm_P)        , O_000F38(06,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Phsubsw         , "phsubsw"         , Enc(SimdRm_P)        , O_000F38(07,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Phsubw          , "phsubw"          , Enc(SimdRm_P)        , O_000F38(05,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pi2fd           , "pi2fd"           , Enc(Simd3dNow)       , O_000F0F(0D,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pi2fw           , "pi2fw"           , Enc(Simd3dNow)       , O_000F0F(0C,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pinsrb          , "pinsrb"          , Enc(SimdRmi)         , O_660F3A(20,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(GdMem)          , O(Imm)            , U                 , U                 ),
  INST(Pinsrd          , "pinsrd"          , Enc(SimdRmi)         , O_660F3A(22,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(GdMem)          , O(Imm)            , U                 , U                 ),
  INST(Pinsrq          , "pinsrq"          , Enc(SimdRmi)         , O_660F3A(22,U,_,W,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(GqMem)          , O(Imm)            , U                 , U                 ),
  INST(Pinsrw          , "pinsrw"          , Enc(SimdRmi_P)       , O_000F00(C4,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(GdMem)          , O(Imm)            , U                 , U                 ),
  INST(Pmaddubsw       , "pmaddubsw"       , Enc(SimdRm_P)        , O_000F38(04,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pmaddwd         , "pmaddwd"         , Enc(SimdRm_P)        , O_000F00(F5,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pmaxsb          , "pmaxsb"          , Enc(SimdRm)          , O_660F38(3C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmaxsd          , "pmaxsd"          , Enc(SimdRm)          , O_660F38(3D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmaxsw          , "pmaxsw"          , Enc(SimdRm_P)        , O_000F00(EE,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pmaxub          , "pmaxub"          , Enc(SimdRm_P)        , O_000F00(DE,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pmaxud          , "pmaxud"          , Enc(SimdRm)          , O_660F38(3F,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmaxuw          , "pmaxuw"          , Enc(SimdRm)          , O_660F38(3E,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pminsb          , "pminsb"          , Enc(SimdRm)          , O_660F38(38,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pminsd          , "pminsd"          , Enc(SimdRm)          , O_660F38(39,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pminsw          , "pminsw"          , Enc(SimdRm_P)        , O_000F00(EA,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pminub          , "pminub"          , Enc(SimdRm_P)        , O_000F00(DA,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pminud          , "pminud"          , Enc(SimdRm)          , O_660F38(3B,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pminuw          , "pminuw"          , Enc(SimdRm)          , O_660F38(3A,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovmskb        , "pmovmskb"        , Enc(SimdRm_PQ)       , O_000F00(D7,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , O(MmXmm)          , U                 , U                 , U                 ),
  INST(Pmovsxbd        , "pmovsxbd"        , Enc(SimdRm)          , O_660F38(21,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovsxbq        , "pmovsxbq"        , Enc(SimdRm)          , O_660F38(22,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovsxbw        , "pmovsxbw"        , Enc(SimdRm)          , O_660F38(20,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovsxdq        , "pmovsxdq"        , Enc(SimdRm)          , O_660F38(25,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovsxwd        , "pmovsxwd"        , Enc(SimdRm)          , O_660F38(23,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovsxwq        , "pmovsxwq"        , Enc(SimdRm)          , O_660F38(24,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovzxbd        , "pmovzxbd"        , Enc(SimdRm)          , O_660F38(31,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovzxbq        , "pmovzxbq"        , Enc(SimdRm)          , O_660F38(32,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovzxbw        , "pmovzxbw"        , Enc(SimdRm)          , O_660F38(30,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovzxdq        , "pmovzxdq"        , Enc(SimdRm)          , O_660F38(35,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovzxwd        , "pmovzxwd"        , Enc(SimdRm)          , O_660F38(33,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmovzxwq        , "pmovzxwq"        , Enc(SimdRm)          , O_660F38(34,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmuldq          , "pmuldq"          , Enc(SimdRm)          , O_660F38(28,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmulhrsw        , "pmulhrsw"        , Enc(SimdRm_P)        , O_000F38(0B,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pmulhrw         , "pmulhrw"         , Enc(Simd3dNow)       , O_000F0F(B7,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Pmulhuw         , "pmulhuw"         , Enc(SimdRm_P)        , O_000F00(E4,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pmulhw          , "pmulhw"          , Enc(SimdRm_P)        , O_000F00(E5,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pmulld          , "pmulld"          , Enc(SimdRm)          , O_660F38(40,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Pmullw          , "pmullw"          , Enc(SimdRm_P)        , O_000F00(D5,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pmuludq         , "pmuludq"         , Enc(SimdRm_P)        , O_000F00(F4,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pop             , "pop"             , Enc(X86Pop)          , O_000000(8F,0,_,_,_), O_000000(58,U,_,_,_), F(WO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , 0                 , U                 , U                 , U                 , U                 ),
  INST(Popa            , "popa"            , Enc(X86Op)           , O_000000(61,U,_,_,_), U                   , F(Volatile)|F(Special)             , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Popcnt          , "popcnt"          , Enc(X86RegRm)        , O_F30F00(B8,U,_,_,_), U                   , F(WO)                              , EF(WWWWWW__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Popf            , "popf"            , Enc(X86Op)           , O_000000(9D,U,_,_,_), U                   , F(Volatile)|F(Special)             , EF(WWWWWWWW), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Por             , "por"             , Enc(SimdRm_P)        , O_000F00(EB,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Prefetch        , "prefetch"        , Enc(X86Prefetch)     , O_000F00(18,U,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(________), 0 , 0 , O(Mem)            , O(Imm)            , U                 , U                 , U                 ),
  INST(Prefetch3dNow   , "prefetch3dnow"   , Enc(X86M)            , O_000F00(0D,0,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Prefetchw       , "prefetchw"       , Enc(X86M)            , O_000F00(0D,1,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(UUUUUU__), 0 , 0 , O(Mem)            , O(Imm)            , U                 , U                 , U                 ),
  INST(Prefetchwt1     , "prefetchwt1"     , Enc(X86M)            , O_000F00(0D,2,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(UUUUUU__), 0 , 0 , O(Mem)            , O(Imm)            , U                 , U                 , U                 ),
  INST(Psadbw          , "psadbw"          , Enc(SimdRm_P)        , O_000F00(F6,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pshufb          , "pshufb"          , Enc(SimdRm_P)        , O_000F38(00,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pshufd          , "pshufd"          , Enc(SimdRmi)         , O_660F00(70,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pshufhw         , "pshufhw"         , Enc(SimdRmi)         , O_F30F00(70,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pshuflw         , "pshuflw"         , Enc(SimdRmi)         , O_F20F00(70,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Pshufw          , "pshufw"          , Enc(SimdRmi_P)       , O_000F00(70,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(MmMem)          , O(Imm)            , U                 , U                 ),
  INST(Psignb          , "psignb"          , Enc(SimdRm_P)        , O_000F38(08,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psignd          , "psignd"          , Enc(SimdRm_P)        , O_000F38(0A,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psignw          , "psignw"          , Enc(SimdRm_P)        , O_000F38(09,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pslld           , "pslld"           , Enc(SimdRmRi_P)      , O_000F00(F2,U,_,_,_), O_000F00(72,6,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                 ),
  INST(Pslldq          , "pslldq"          , Enc(SimdRmRi)        , U                   , O_660F00(73,7,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(Imm)            , U                 , U                 , U                 ),
  INST(Psllq           , "psllq"           , Enc(SimdRmRi_P)      , O_000F00(F3,U,_,_,_), O_000F00(73,6,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                 ),
  INST(Psllw           , "psllw"           , Enc(SimdRmRi_P)      , O_000F00(F1,U,_,_,_), O_000F00(71,6,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                 ),
  INST(Psrad           , "psrad"           , Enc(SimdRmRi_P)      , O_000F00(E2,U,_,_,_), O_000F00(72,4,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                 ),
  INST(Psraw           , "psraw"           , Enc(SimdRmRi_P)      , O_000F00(E1,U,_,_,_), O_000F00(71,4,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                 ),
  INST(Psrld           , "psrld"           , Enc(SimdRmRi_P)      , O_000F00(D2,U,_,_,_), O_000F00(72,2,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                 ),
  INST(Psrldq          , "psrldq"          , Enc(SimdRmRi)        , U                   , O_660F00(73,3,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(Imm)            , U                 , U                 , U                 ),
  INST(Psrlq           , "psrlq"           , Enc(SimdRmRi_P)      , O_000F00(D3,U,_,_,_), O_000F00(73,2,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                 ),
  INST(Psrlw           , "psrlw"           , Enc(SimdRmRi_P)      , O_000F00(D1,U,_,_,_), O_000F00(71,2,_,_,_), F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)|O(Imm), U                 , U                 , U                 ),
  INST(Psubb           , "psubb"           , Enc(SimdRm_P)        , O_000F00(F8,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psubd           , "psubd"           , Enc(SimdRm_P)        , O_000F00(FA,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psubq           , "psubq"           , Enc(SimdRm_P)        , O_000F00(FB,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psubsb          , "psubsb"          , Enc(SimdRm_P)        , O_000F00(E8,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psubsw          , "psubsw"          , Enc(SimdRm_P)        , O_000F00(E9,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psubusb         , "psubusb"         , Enc(SimdRm_P)        , O_000F00(D8,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psubusw         , "psubusw"         , Enc(SimdRm_P)        , O_000F00(D9,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Psubw           , "psubw"           , Enc(SimdRm_P)        , O_000F00(F9,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Pswapd          , "pswapd"          , Enc(Simd3dNow)       , O_000F0F(BB,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Mm)             , O(MmMem)          , U                 , U                 , U                 ),
  INST(Ptest           , "ptest"           , Enc(SimdRm)          , O_660F38(17,U,_,_,_), U                   , F(RO)                              , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Punpckhbw       , "punpckhbw"       , Enc(SimdRm_P)        , O_000F00(68,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Punpckhdq       , "punpckhdq"       , Enc(SimdRm_P)        , O_000F00(6A,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Punpckhqdq      , "punpckhqdq"      , Enc(SimdRm)          , O_660F00(6D,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Punpckhwd       , "punpckhwd"       , Enc(SimdRm_P)        , O_000F00(69,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Punpcklbw       , "punpcklbw"       , Enc(SimdRm_P)        , O_000F00(60,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Punpckldq       , "punpckldq"       , Enc(SimdRm_P)        , O_000F00(62,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Punpcklqdq      , "punpcklqdq"      , Enc(SimdRm)          , O_660F00(6C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Punpcklwd       , "punpcklwd"       , Enc(SimdRm_P)        , O_000F00(61,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Push            , "push"            , Enc(X86Push)         , O_000000(FF,6,_,_,_), O_000000(50,U,_,_,_), F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , 0                 , U                 , U                 , U                 , U                 ),
  INST(Pusha           , "pusha"           , Enc(X86Op)           , O_000000(60,U,_,_,_), U                   , F(Volatile)|F(Special)             , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Pushf           , "pushf"           , Enc(X86Op)           , O_000000(9C,U,_,_,_), U                   , F(Volatile)|F(Special)             , EF(RRRRRRRR), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Pxor            , "pxor"            , Enc(SimdRm_P)        , O_000F00(EF,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(MmXmm)          , O(MmXmmMem)       , U                 , U                 , U                 ),
  INST(Rcl             , "rcl"             , Enc(X86Rot)          , O_000000(D0,2,_,_,_), U                   , F(RW)|F(Special)                   , EF(W____X__), 0 , 0 , O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                 ),
  INST(Rcpps           , "rcpps"           , Enc(SimdRm)          , O_000F00(53,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Rcpss           , "rcpss"           , Enc(SimdRm)          , O_F30F00(53,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 4 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Rcr             , "rcr"             , Enc(X86Rot)          , O_000000(D0,3,_,_,_), U                   , F(RW)|F(Special)                   , EF(W____X__), 0 , 0 , O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                 ),
  INST(Rdfsbase        , "rdfsbase"        , Enc(X86Rm)           , O_F30F00(AE,0,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , U                 , U                 , U                 , U                 ),
  INST(Rdgsbase        , "rdgsbase"        , Enc(X86Rm)           , O_F30F00(AE,1,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Gqd)            , U                 , U                 , U                 , U                 ),
  INST(Rdrand          , "rdrand"          , Enc(X86Rm)           , O_000F00(C7,6,_,_,_), U                   , F(WO)                              , EF(WWWWWW__), 0 , 8 , O(Gqdw)           , U                 , U                 , U                 , U                 ),
  INST(Rdseed          , "rdseed"          , Enc(X86Rm)           , O_000F00(C7,7,_,_,_), U                   , F(WO)                              , EF(WWWWWW__), 0 , 8 , O(Gqdw)           , U                 , U                 , U                 , U                 ),
  INST(Rdtsc           , "rdtsc"           , Enc(X86Op)           , O_000F00(31,U,_,_,_), U                   , F(WO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Rdtscp          , "rdtscp"          , Enc(X86Op)           , O_000F01(F9,U,_,_,_), U                   , F(WO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(RepLodsB        , "rep lods_b"      , Enc(X86Rep)          , O_000000(AC,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(RepLodsD        , "rep lods_d"      , Enc(X86Rep)          , O_000000(AD,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(RepLodsQ        , "rep lods_q"      , Enc(X86Rep)          , O_000000(AD,1,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(RepLodsW        , "rep lods_w"      , Enc(X86Rep)          , O_660000(AD,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(RepMovsB        , "rep movs_b"      , Enc(X86Rep)          , O_000000(A4,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepMovsD        , "rep movs_d"      , Enc(X86Rep)          , O_000000(A5,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepMovsQ        , "rep movs_q"      , Enc(X86Rep)          , O_000000(A5,1,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepMovsW        , "rep movs_w"      , Enc(X86Rep)          , O_660000(A5,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepStosB        , "rep stos_b"      , Enc(X86Rep)          , O_000000(AA,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(RepStosD        , "rep stos_d"      , Enc(X86Rep)          , O_000000(AB,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(RepStosQ        , "rep stos_q"      , Enc(X86Rep)          , O_000000(AB,1,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(RepStosW        , "rep stos_w"      , Enc(X86Rep)          , O_660000(AB,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(RepeCmpsB       , "repe cmps_b"     , Enc(X86Rep)          , O_000000(A6,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepeCmpsD       , "repe cmps_d"     , Enc(X86Rep)          , O_000000(A7,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepeCmpsQ       , "repe cmps_q"     , Enc(X86Rep)          , O_000000(A7,1,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepeCmpsW       , "repe cmps_w"     , Enc(X86Rep)          , O_660000(A7,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepeScasB       , "repe scas_b"     , Enc(X86Rep)          , O_000000(AE,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepeScasD       , "repe scas_d"     , Enc(X86Rep)          , O_000000(AF,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepeScasQ       , "repe scas_q"     , Enc(X86Rep)          , O_000000(AF,1,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepeScasW       , "repe scas_w"     , Enc(X86Rep)          , O_660000(AF,1,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepneCmpsB      , "repne cmps_b"    , Enc(X86Rep)          , O_000000(A6,0,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepneCmpsD      , "repne cmps_d"    , Enc(X86Rep)          , O_000000(A7,0,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepneCmpsQ      , "repne cmps_q"    , Enc(X86Rep)          , O_000000(A7,0,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepneCmpsW      , "repne cmps_w"    , Enc(X86Rep)          , O_660000(A7,0,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepneScasB      , "repne scas_b"    , Enc(X86Rep)          , O_000000(AE,0,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepneScasD      , "repne scas_d"    , Enc(X86Rep)          , O_000000(AF,0,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepneScasQ      , "repne scas_q"    , Enc(X86Rep)          , O_000000(AF,0,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(RepneScasW      , "repne scas_w"    , Enc(X86Rep)          , O_660000(AF,0,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , O(Mem)            , O(Mem)            , U                 , U                 , U                 ),
  INST(Ret             , "ret"             , Enc(X86Ret)          , O_000000(C2,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Rol             , "rol"             , Enc(X86Rot)          , O_000000(D0,0,_,_,_), U                   , F(RW)|F(Special)                   , EF(W____W__), 0 , 0 , O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                 ),
  INST(Ror             , "ror"             , Enc(X86Rot)          , O_000000(D0,1,_,_,_), U                   , F(RW)|F(Special)                   , EF(W____W__), 0 , 0 , O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                 ),
  INST(Rorx            , "rorx"            , Enc(AvxRmi_OptW)     , O_F20F3A(F0,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqd)            , O(GqdMem)         , O(Imm)            , U                 , U                 ),
  INST(Roundpd         , "roundpd"         , Enc(SimdRmi)         , O_660F3A(09,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Roundps         , "roundps"         , Enc(SimdRmi)         , O_660F3A(08,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Roundsd         , "roundsd"         , Enc(SimdRmi)         , O_660F3A(0B,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Roundss         , "roundss"         , Enc(SimdRmi)         , O_660F3A(0A,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 4 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Rsqrtps         , "rsqrtps"         , Enc(SimdRm)          , O_000F00(52,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Rsqrtss         , "rsqrtss"         , Enc(SimdRm)          , O_F30F00(52,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 4 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sahf            , "sahf"            , Enc(X86Op)           , O_000000(9E,U,_,_,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(_WWWWW__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Sal             , "sal"             , Enc(X86Rot)          , O_000000(D0,4,_,_,_), U                   , F(RW)|F(Special)                   , EF(WWWUWW__), 0 , 0 , O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                 ),
  INST(Sar             , "sar"             , Enc(X86Rot)          , O_000000(D0,7,_,_,_), U                   , F(RW)|F(Special)                   , EF(WWWUWW__), 0 , 0 , O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                 ),
  INST(Sarx            , "sarx"            , Enc(AvxRmv_OptW)     , O_F30F38(F7,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqd)            , O(GqdMem)         , O(Gqd)            , U                 , U                 ),
  INST(Sbb             , "sbb"             , Enc(X86Arith)        , O_000000(18,3,_,_,_), U                   , F(RW)|F(Lock)                      , EF(WWWWWX__), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(ScasB           , "scas_b"          , Enc(X86Op)           , O_000000(AE,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(ScasD           , "scas_d"          , Enc(X86Op)           , O_000000(AF,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(ScasQ           , "scas_q"          , Enc(X86Op)           , O_000000(AF,U,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(ScasW           , "scas_w"          , Enc(X86Op_66H)       , O_000000(AF,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(WWWWWWR_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Seta            , "seta"            , Enc(X86Set)          , O_000F00(97,U,_,_,_), U                   , F(WO)                              , EF(__R__R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setae           , "setae"           , Enc(X86Set)          , O_000F00(93,U,_,_,_), U                   , F(WO)                              , EF(_____R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setb            , "setb"            , Enc(X86Set)          , O_000F00(92,U,_,_,_), U                   , F(WO)                              , EF(_____R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setbe           , "setbe"           , Enc(X86Set)          , O_000F00(96,U,_,_,_), U                   , F(WO)                              , EF(__R__R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setc            , "setc"            , Enc(X86Set)          , O_000F00(92,U,_,_,_), U                   , F(WO)                              , EF(_____R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Sete            , "sete"            , Enc(X86Set)          , O_000F00(94,U,_,_,_), U                   , F(WO)                              , EF(__R_____), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setg            , "setg"            , Enc(X86Set)          , O_000F00(9F,U,_,_,_), U                   , F(WO)                              , EF(RRR_____), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setge           , "setge"           , Enc(X86Set)          , O_000F00(9D,U,_,_,_), U                   , F(WO)                              , EF(RR______), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setl            , "setl"            , Enc(X86Set)          , O_000F00(9C,U,_,_,_), U                   , F(WO)                              , EF(RR______), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setle           , "setle"           , Enc(X86Set)          , O_000F00(9E,U,_,_,_), U                   , F(WO)                              , EF(RRR_____), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setna           , "setna"           , Enc(X86Set)          , O_000F00(96,U,_,_,_), U                   , F(WO)                              , EF(__R__R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnae          , "setnae"          , Enc(X86Set)          , O_000F00(92,U,_,_,_), U                   , F(WO)                              , EF(_____R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnb           , "setnb"           , Enc(X86Set)          , O_000F00(93,U,_,_,_), U                   , F(WO)                              , EF(_____R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnbe          , "setnbe"          , Enc(X86Set)          , O_000F00(97,U,_,_,_), U                   , F(WO)                              , EF(__R__R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnc           , "setnc"           , Enc(X86Set)          , O_000F00(93,U,_,_,_), U                   , F(WO)                              , EF(_____R__), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setne           , "setne"           , Enc(X86Set)          , O_000F00(95,U,_,_,_), U                   , F(WO)                              , EF(__R_____), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setng           , "setng"           , Enc(X86Set)          , O_000F00(9E,U,_,_,_), U                   , F(WO)                              , EF(RRR_____), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnge          , "setnge"          , Enc(X86Set)          , O_000F00(9C,U,_,_,_), U                   , F(WO)                              , EF(RR______), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnl           , "setnl"           , Enc(X86Set)          , O_000F00(9D,U,_,_,_), U                   , F(WO)                              , EF(RR______), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnle          , "setnle"          , Enc(X86Set)          , O_000F00(9F,U,_,_,_), U                   , F(WO)                              , EF(RRR_____), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setno           , "setno"           , Enc(X86Set)          , O_000F00(91,U,_,_,_), U                   , F(WO)                              , EF(R_______), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnp           , "setnp"           , Enc(X86Set)          , O_000F00(9B,U,_,_,_), U                   , F(WO)                              , EF(____R___), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setns           , "setns"           , Enc(X86Set)          , O_000F00(99,U,_,_,_), U                   , F(WO)                              , EF(_R______), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setnz           , "setnz"           , Enc(X86Set)          , O_000F00(95,U,_,_,_), U                   , F(WO)                              , EF(__R_____), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Seto            , "seto"            , Enc(X86Set)          , O_000F00(90,U,_,_,_), U                   , F(WO)                              , EF(R_______), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setp            , "setp"            , Enc(X86Set)          , O_000F00(9A,U,_,_,_), U                   , F(WO)                              , EF(____R___), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setpe           , "setpe"           , Enc(X86Set)          , O_000F00(9A,U,_,_,_), U                   , F(WO)                              , EF(____R___), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setpo           , "setpo"           , Enc(X86Set)          , O_000F00(9B,U,_,_,_), U                   , F(WO)                              , EF(____R___), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Sets            , "sets"            , Enc(X86Set)          , O_000F00(98,U,_,_,_), U                   , F(WO)                              , EF(_R______), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Setz            , "setz"            , Enc(X86Set)          , O_000F00(94,U,_,_,_), U                   , F(WO)                              , EF(__R_____), 0 , 1 , O(GbMem)          , U                 , U                 , U                 , U                 ),
  INST(Sfence          , "sfence"          , Enc(X86Fence)        , O_000F00(AE,7,_,_,_), U                   , F(Volatile)                        , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Sha1msg1        , "sha1msg1"        , Enc(SimdRm)          , O_000F38(C9,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sha1msg2        , "sha1msg2"        , Enc(SimdRm)          , O_000F38(CA,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sha1nexte       , "sha1nexte"       , Enc(SimdRm)          , O_000F38(C8,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sha1rnds4       , "sha1rnds4"       , Enc(SimdRmi)         , O_000F3A(CC,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Sha256msg1      , "sha256msg1"      , Enc(SimdRm)          , O_000F38(CC,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sha256msg2      , "sha256msg2"      , Enc(SimdRm)          , O_000F38(CD,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sha256rnds2     , "sha256rnds2"     , Enc(SimdRm)          , O_000F38(CB,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Shl             , "shl"             , Enc(X86Rot)          , O_000000(D0,4,_,_,_), U                   , F(RW)|F(Special)                   , EF(WWWUWW__), 0 , 0 , O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                 ),
  INST(Shld            , "shld"            , Enc(X86ShldShrd)     , O_000F00(A4,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(UWWUWW__), 0 , 0 , O(GqdwbMem)       , O(Gb)             , U                 , U                 , U                 ),
  INST(Shlx            , "shlx"            , Enc(AvxRmv_OptW)     , O_660F38(F7,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqd)            , O(GqdMem)         , O(Gqd)            , U                 , U                 ),
  INST(Shr             , "shr"             , Enc(X86Rot)          , O_000000(D0,5,_,_,_), U                   , F(RW)|F(Special)                   , EF(WWWUWW__), 0 , 0 , O(GqdwbMem)       , O(Gb)|O(Imm)      , U                 , U                 , U                 ),
  INST(Shrd            , "shrd"            , Enc(X86ShldShrd)     , O_000F00(AC,U,_,_,_), U                   , F(RW)|F(Special)                   , EF(UWWUWW__), 0 , 0 , O(GqdwbMem)       , O(Gqdwb)          , U                 , U                 , U                 ),
  INST(Shrx            , "shrx"            , Enc(AvxRmv_OptW)     , O_F20F38(F7,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 0 , O(Gqd)            , O(GqdMem)         , O(Gqd)            , U                 , U                 ),
  INST(Shufpd          , "shufpd"          , Enc(SimdRmi)         , O_660F00(C6,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Shufps          , "shufps"          , Enc(SimdRmi)         , O_000F00(C6,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Sqrtpd          , "sqrtpd"          , Enc(SimdRm)          , O_660F00(51,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sqrtps          , "sqrtps"          , Enc(SimdRm)          , O_000F00(51,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 16, O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sqrtsd          , "sqrtsd"          , Enc(SimdRm)          , O_F20F00(51,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 8 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Sqrtss          , "sqrtss"          , Enc(SimdRm)          , O_F30F00(51,U,_,_,_), U                   , F(WO)                              , EF(________), 0 , 4 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Stc             , "stc"             , Enc(X86Op)           , O_000000(F9,U,_,_,_), U                   , F(None)                            , EF(_____W__), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Std             , "std"             , Enc(X86Op)           , O_000000(FD,U,_,_,_), U                   , F(None)                            , EF(______W_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Stmxcsr         , "stmxcsr"         , Enc(X86M)            , O_000F00(AE,3,_,_,_), U                   , F(Volatile)                        , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(StosB           , "stos_b"          , Enc(X86Op)           , O_000000(AA,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(StosD           , "stos_d"          , Enc(X86Op)           , O_000000(AB,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(StosQ           , "stos_q"          , Enc(X86Op)           , O_000000(AB,U,_,W,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(StosW           , "stos_w"          , Enc(X86Op_66H)       , O_000000(AB,U,_,_,_), U                   , F(RW)|F(Volatile)|F(Special)       , EF(______R_), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Sub             , "sub"             , Enc(X86Arith)        , O_000000(28,5,_,_,_), U                   , F(RW)|F(Lock)                      , EF(WWWWWW__), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(Subpd           , "subpd"           , Enc(SimdRm)          , O_660F00(5C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Subps           , "subps"           , Enc(SimdRm)          , O_000F00(5C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Subsd           , "subsd"           , Enc(SimdRm)          , O_F20F00(5C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Subss           , "subss"           , Enc(SimdRm)          , O_F30F00(5C,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(T1mskc          , "t1mskc"          , Enc(XopVm_OptW)      , O_00_M09(01,7,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Test            , "test"            , Enc(X86Test)         , O_000000(84,U,_,_,_), O_000000(F6,U,_,_,_), F(RO)                              , EF(WWWUWW__), 0 , 0 , O(GqdwbMem)       , O(Gqdwb)|O(Imm)   , U                 , U                 , U                 ),
  INST(Tzcnt           , "tzcnt"           , Enc(X86RegRm)        , O_F30F00(BC,U,_,_,_), U                   , F(WO)                              , EF(UUWUUW__), 0 , 0 , O(Gqdw)           , O(GqdwMem)        , U                 , U                 , U                 ),
  INST(Tzmsk           , "tzmsk"           , Enc(XopVm_OptW)      , O_00_M09(01,4,_,_,_), U                   , F(WO)                              , EF(WWWUUW__), 0 , 0 , O(Gqd)            , O(GqdMem)         , U                 , U                 , U                 ),
  INST(Ucomisd         , "ucomisd"         , Enc(SimdRm)          , O_660F00(2E,U,_,_,_), U                   , F(RO)                              , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Ucomiss         , "ucomiss"         , Enc(SimdRm)          , O_000F00(2E,U,_,_,_), U                   , F(RO)                              , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Ud2             , "ud2"             , Enc(X86Op)           , O_000F00(0B,U,_,_,_), U                   , F(None)                            , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Unpckhpd        , "unpckhpd"        , Enc(SimdRm)          , O_660F00(15,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Unpckhps        , "unpckhps"        , Enc(SimdRm)          , O_000F00(15,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Unpcklpd        , "unpcklpd"        , Enc(SimdRm)          , O_660F00(14,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Unpcklps        , "unpcklps"        , Enc(SimdRm)          , O_000F00(14,U,_,_,_), U                   , F(RW)                              , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vaddpd          , "vaddpd"          , Enc(AvxRvm_OptL)     , O_660F00(58,U,_,I,1), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vaddps          , "vaddps"          , Enc(AvxRvm_OptL)     , O_000F00(58,U,_,I,0), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vaddsd          , "vaddsd"          , Enc(AvxRvm)          , O_F20F00(58,U,0,I,1), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vaddss          , "vaddss"          , Enc(AvxRvm)          , O_F30F00(58,U,0,I,0), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vaddsubpd       , "vaddsubpd"       , Enc(AvxRvm_OptL)     , O_660F00(D0,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vaddsubps       , "vaddsubps"       , Enc(AvxRvm_OptL)     , O_F20F00(D0,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vaesdec         , "vaesdec"         , Enc(AvxRvm)          , O_660F38(DE,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vaesdeclast     , "vaesdeclast"     , Enc(AvxRvm)          , O_660F38(DF,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vaesenc         , "vaesenc"         , Enc(AvxRvm)          , O_660F38(DC,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vaesenclast     , "vaesenclast"     , Enc(AvxRvm)          , O_660F38(DD,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vaesimc         , "vaesimc"         , Enc(AvxRm)           , O_660F38(DB,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vaeskeygenassist, "vaeskeygenassist", Enc(AvxRmi)          , O_660F3A(DF,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Vandnpd         , "vandnpd"         , Enc(AvxRvm_OptL)     , O_660F00(55,U,_,_,1), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vandnps         , "vandnps"         , Enc(AvxRvm_OptL)     , O_000F00(55,U,_,_,0), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vandpd          , "vandpd"          , Enc(AvxRvm_OptL)     , O_660F00(54,U,_,_,1), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vandps          , "vandps"          , Enc(AvxRvm_OptL)     , O_000F00(54,U,_,_,0), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vblendpd        , "vblendpd"        , Enc(AvxRvmi_OptL)    , O_660F3A(0D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vblendps        , "vblendps"        , Enc(AvxRvmi_OptL)    , O_660F3A(0C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vblendvpd       , "vblendvpd"       , Enc(AvxRvmr_OptL)    , O_660F3A(4B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Xy)             , U                 ),
  INST(Vblendvps       , "vblendvps"       , Enc(AvxRvmr_OptL)    , O_660F3A(4A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Xy)             , U                 ),
  INST(Vbroadcastf128  , "vbroadcastf128"  , Enc(AvxRm)           , O_660F38(1A,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(Mem)            , U                 , U                 , U                 ),
  INST(Vbroadcasti128  , "vbroadcasti128"  , Enc(AvxRm)           , O_660F38(5A,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(Mem)            , U                 , U                 , U                 ),
  INST(Vbroadcastsd    , "vbroadcastsd"    , Enc(AvxRm)           , O_660F38(19,U,L,0,1), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vbroadcastss    , "vbroadcastss"    , Enc(AvxRm_OptL)      , O_660F38(18,U,_,0,0), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcmppd          , "vcmppd"          , Enc(AvxRvmi_OptL)    , O_660F00(C2,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vcmpps          , "vcmpps"          , Enc(AvxRvmi_OptL)    , O_000F00(C2,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vcmpsd          , "vcmpsd"          , Enc(AvxRvmi)         , O_F20F00(C2,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vcmpss          , "vcmpss"          , Enc(AvxRvmi)         , O_F30F00(C2,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vcomisd         , "vcomisd"         , Enc(AvxRm)           , O_660F00(2F,U,_,_,_), U                   , F(RO)|F(Avx)                       , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcomiss         , "vcomiss"         , Enc(AvxRm)           , O_000F00(2F,U,_,_,_), U                   , F(RO)|F(Avx)                       , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcvtdq2pd       , "vcvtdq2pd"       , Enc(AvxRm_OptL)      , O_F30F00(E6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcvtdq2ps       , "vcvtdq2ps"       , Enc(AvxRm_OptL)      , O_000F00(5B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vcvtpd2dq       , "vcvtpd2dq"       , Enc(AvxRm)           , O_F20F00(E6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vcvtpd2ps       , "vcvtpd2ps"       , Enc(AvxRm)           , O_660F00(5A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vcvtph2ps       , "vcvtph2ps"       , Enc(AvxRm_OptL)      , O_660F38(13,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcvtps2dq       , "vcvtps2dq"       , Enc(AvxRm_OptL)      , O_660F00(5B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vcvtps2pd       , "vcvtps2pd"       , Enc(AvxRm_OptL)      , O_000F00(5A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcvtps2ph       , "vcvtps2ph"       , Enc(AvxMri_OptL)     , O_660F3A(1D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(Xy)             , O(Imm)            , U                 , U                 ),
  INST(Vcvtsd2si       , "vcvtsd2si"       , Enc(AvxRm)           , O_F20F00(2D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Gqd)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcvtsd2ss       , "vcvtsd2ss"       , Enc(AvxRvm)          , O_F20F00(5A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vcvtsi2sd       , "vcvtsi2sd"       , Enc(AvxRvm)          , O_F20F00(2A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(GqdMem)         , U                 , U                 ),
  INST(Vcvtsi2ss       , "vcvtsi2ss"       , Enc(AvxRvm)          , O_F30F00(2A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(GqdMem)         , U                 , U                 ),
  INST(Vcvtss2sd       , "vcvtss2sd"       , Enc(AvxRvm)          , O_F30F00(5A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vcvtss2si       , "vcvtss2si"       , Enc(AvxRm)           , O_F20F00(2D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Gqd)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcvttpd2dq      , "vcvttpd2dq"      , Enc(AvxRm_OptL)      , O_660F00(E6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vcvttps2dq      , "vcvttps2dq"      , Enc(AvxRm_OptL)      , O_F30F00(5B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vcvttsd2si      , "vcvttsd2si"      , Enc(AvxRm)           , O_F20F00(2C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Gqd)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vcvttss2si      , "vcvttss2si"      , Enc(AvxRm)           , O_F30F00(2C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Gqd)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vdivpd          , "vdivpd"          , Enc(AvxRvm_OptL)     , O_660F00(5E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vdivps          , "vdivps"          , Enc(AvxRvm_OptL)     , O_000F00(5E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vdivsd          , "vdivsd"          , Enc(AvxRvm)          , O_F20F00(5E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vdivss          , "vdivss"          , Enc(AvxRvm)          , O_F30F00(5E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vdppd           , "vdppd"           , Enc(AvxRvmi)         , O_660F3A(41,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vdpps           , "vdpps"           , Enc(AvxRvmi_OptL)    , O_660F3A(40,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vextractf128    , "vextractf128"    , Enc(AvxMri)          , O_660F3A(19,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(Ymm)            , O(Imm)            , U                 , U                 ),
  INST(Vextracti128    , "vextracti128"    , Enc(AvxMri)          , O_660F3A(39,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(Ymm)            , O(Imm)            , U                 , U                 ),
  INST(Vextractps      , "vextractps"      , Enc(AvxMri)          , O_660F3A(17,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(GqdMem)         , O(Xmm)            , O(Imm)            , U                 , U                 ),
  INST(Vfmadd132pd     , "vfmadd132pd"     , Enc(AvxRvm_OptL)     , O_660F38(98,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmadd132ps     , "vfmadd132ps"     , Enc(AvxRvm_OptL)     , O_660F38(98,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmadd132sd     , "vfmadd132sd"     , Enc(AvxRvm)          , O_660F38(99,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmadd132ss     , "vfmadd132ss"     , Enc(AvxRvm)          , O_660F38(99,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmadd213pd     , "vfmadd213pd"     , Enc(AvxRvm_OptL)     , O_660F38(A8,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmadd213ps     , "vfmadd213ps"     , Enc(AvxRvm_OptL)     , O_660F38(A8,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmadd213sd     , "vfmadd213sd"     , Enc(AvxRvm)          , O_660F38(A9,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmadd213ss     , "vfmadd213ss"     , Enc(AvxRvm)          , O_660F38(A9,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmadd231pd     , "vfmadd231pd"     , Enc(AvxRvm_OptL)     , O_660F38(B8,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmadd231ps     , "vfmadd231ps"     , Enc(AvxRvm_OptL)     , O_660F38(B8,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmadd231sd     , "vfmadd231sd"     , Enc(AvxRvm)          , O_660F38(B9,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmadd231ss     , "vfmadd231ss"     , Enc(AvxRvm)          , O_660F38(B9,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmaddpd        , "vfmaddpd"        , Enc(Fma4_OptL)       , O_660F3A(69,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfmaddps        , "vfmaddps"        , Enc(Fma4_OptL)       , O_660F3A(68,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfmaddsd        , "vfmaddsd"        , Enc(Fma4)            , O_660F3A(6B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vfmaddss        , "vfmaddss"        , Enc(Fma4)            , O_660F3A(6A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vfmaddsub132pd  , "vfmaddsub132pd"  , Enc(AvxRvm_OptL)     , O_660F38(96,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmaddsub132ps  , "vfmaddsub132ps"  , Enc(AvxRvm_OptL)     , O_660F38(96,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmaddsub213pd  , "vfmaddsub213pd"  , Enc(AvxRvm_OptL)     , O_660F38(A6,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmaddsub213ps  , "vfmaddsub213ps"  , Enc(AvxRvm_OptL)     , O_660F38(A6,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmaddsub231pd  , "vfmaddsub231pd"  , Enc(AvxRvm_OptL)     , O_660F38(B6,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmaddsub231ps  , "vfmaddsub231ps"  , Enc(AvxRvm_OptL)     , O_660F38(B6,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmaddsubpd     , "vfmaddsubpd"     , Enc(Fma4_OptL)       , O_660F3A(5D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfmaddsubps     , "vfmaddsubps"     , Enc(Fma4_OptL)       , O_660F3A(5C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfmsub132pd     , "vfmsub132pd"     , Enc(AvxRvm_OptL)     , O_660F38(9A,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsub132ps     , "vfmsub132ps"     , Enc(AvxRvm_OptL)     , O_660F38(9A,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsub132sd     , "vfmsub132sd"     , Enc(AvxRvm)          , O_660F38(9B,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmsub132ss     , "vfmsub132ss"     , Enc(AvxRvm)          , O_660F38(9B,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmsub213pd     , "vfmsub213pd"     , Enc(AvxRvm_OptL)     , O_660F38(AA,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsub213ps     , "vfmsub213ps"     , Enc(AvxRvm_OptL)     , O_660F38(AA,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsub213sd     , "vfmsub213sd"     , Enc(AvxRvm)          , O_660F38(AB,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmsub213ss     , "vfmsub213ss"     , Enc(AvxRvm)          , O_660F38(AB,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmsub231pd     , "vfmsub231pd"     , Enc(AvxRvm_OptL)     , O_660F38(BA,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsub231ps     , "vfmsub231ps"     , Enc(AvxRvm_OptL)     , O_660F38(BA,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsub231sd     , "vfmsub231sd"     , Enc(AvxRvm)          , O_660F38(BB,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmsub231ss     , "vfmsub231ss"     , Enc(AvxRvm)          , O_660F38(BB,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfmsubadd132pd  , "vfmsubadd132pd"  , Enc(AvxRvm_OptL)     , O_660F38(97,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsubadd132ps  , "vfmsubadd132ps"  , Enc(AvxRvm_OptL)     , O_660F38(97,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsubadd213pd  , "vfmsubadd213pd"  , Enc(AvxRvm_OptL)     , O_660F38(A7,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsubadd213ps  , "vfmsubadd213ps"  , Enc(AvxRvm_OptL)     , O_660F38(A7,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsubadd231pd  , "vfmsubadd231pd"  , Enc(AvxRvm_OptL)     , O_660F38(B7,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsubadd231ps  , "vfmsubadd231ps"  , Enc(AvxRvm_OptL)     , O_660F38(B7,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfmsubaddpd     , "vfmsubaddpd"     , Enc(Fma4_OptL)       , O_660F3A(5F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfmsubaddps     , "vfmsubaddps"     , Enc(Fma4_OptL)       , O_660F3A(5E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfmsubpd        , "vfmsubpd"        , Enc(Fma4_OptL)       , O_660F3A(6D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfmsubps        , "vfmsubps"        , Enc(Fma4_OptL)       , O_660F3A(6C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfmsubsd        , "vfmsubsd"        , Enc(Fma4)            , O_660F3A(6F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vfmsubss        , "vfmsubss"        , Enc(Fma4)            , O_660F3A(6E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vfnmadd132pd    , "vfnmadd132pd"    , Enc(AvxRvm_OptL)     , O_660F38(9C,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmadd132ps    , "vfnmadd132ps"    , Enc(AvxRvm_OptL)     , O_660F38(9C,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmadd132sd    , "vfnmadd132sd"    , Enc(AvxRvm)          , O_660F38(9D,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmadd132ss    , "vfnmadd132ss"    , Enc(AvxRvm)          , O_660F38(9D,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmadd213pd    , "vfnmadd213pd"    , Enc(AvxRvm_OptL)     , O_660F38(AC,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmadd213ps    , "vfnmadd213ps"    , Enc(AvxRvm_OptL)     , O_660F38(AC,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmadd213sd    , "vfnmadd213sd"    , Enc(AvxRvm)          , O_660F38(AD,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmadd213ss    , "vfnmadd213ss"    , Enc(AvxRvm)          , O_660F38(AD,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmadd231pd    , "vfnmadd231pd"    , Enc(AvxRvm_OptL)     , O_660F38(BC,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmadd231ps    , "vfnmadd231ps"    , Enc(AvxRvm_OptL)     , O_660F38(BC,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmadd231sd    , "vfnmadd231sd"    , Enc(AvxRvm)          , O_660F38(BC,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmadd231ss    , "vfnmadd231ss"    , Enc(AvxRvm)          , O_660F38(BC,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmaddpd       , "vfnmaddpd"       , Enc(Fma4_OptL)       , O_660F3A(79,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfnmaddps       , "vfnmaddps"       , Enc(Fma4_OptL)       , O_660F3A(78,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfnmaddsd       , "vfnmaddsd"       , Enc(Fma4)            , O_660F3A(7B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vfnmaddss       , "vfnmaddss"       , Enc(Fma4)            , O_660F3A(7A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vfnmsub132pd    , "vfnmsub132pd"    , Enc(AvxRvm_OptL)     , O_660F38(9E,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmsub132ps    , "vfnmsub132ps"    , Enc(AvxRvm_OptL)     , O_660F38(9E,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmsub132sd    , "vfnmsub132sd"    , Enc(AvxRvm)          , O_660F38(9F,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmsub132ss    , "vfnmsub132ss"    , Enc(AvxRvm)          , O_660F38(9F,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmsub213pd    , "vfnmsub213pd"    , Enc(AvxRvm_OptL)     , O_660F38(AE,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmsub213ps    , "vfnmsub213ps"    , Enc(AvxRvm_OptL)     , O_660F38(AE,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmsub213sd    , "vfnmsub213sd"    , Enc(AvxRvm)          , O_660F38(AF,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmsub213ss    , "vfnmsub213ss"    , Enc(AvxRvm)          , O_660F38(AF,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmsub231pd    , "vfnmsub231pd"    , Enc(AvxRvm_OptL)     , O_660F38(BE,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmsub231ps    , "vfnmsub231ps"    , Enc(AvxRvm_OptL)     , O_660F38(BE,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vfnmsub231sd    , "vfnmsub231sd"    , Enc(AvxRvm)          , O_660F38(BF,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmsub231ss    , "vfnmsub231ss"    , Enc(AvxRvm)          , O_660F38(BF,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vfnmsubpd       , "vfnmsubpd"       , Enc(Fma4_OptL)       , O_660F3A(7D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfnmsubps       , "vfnmsubps"       , Enc(Fma4_OptL)       , O_660F3A(7C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vfnmsubsd       , "vfnmsubsd"       , Enc(Fma4)            , O_660F3A(7F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vfnmsubss       , "vfnmsubss"       , Enc(Fma4)            , O_660F3A(7E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vfrczpd         , "vfrczpd"         , Enc(XopRm_OptL)      , O_00_M09(81,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vfrczps         , "vfrczps"         , Enc(XopRm_OptL)      , O_00_M09(80,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vfrczsd         , "vfrczsd"         , Enc(XopRm)           , O_00_M09(83,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vfrczss         , "vfrczss"         , Enc(XopRm)           , O_00_M09(82,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vgatherdpd      , "vgatherdpd"      , Enc(AvxGather)       , O_660F38(92,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Mem)            , O(Xy)             , U                 , U                 ),
  INST(Vgatherdps      , "vgatherdps"      , Enc(AvxGather)       , O_660F38(92,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Mem)            , O(Xy)             , U                 , U                 ),
  INST(Vgatherqpd      , "vgatherqpd"      , Enc(AvxGather)       , O_660F38(93,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Mem)            , O(Xy)             , U                 , U                 ),
  INST(Vgatherqps      , "vgatherqps"      , Enc(AvxGatherEx)     , O_660F38(93,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Mem)            , O(Xmm)            , U                 , U                 ),
  INST(Vhaddpd         , "vhaddpd"         , Enc(AvxRvm_OptL)     , O_660F00(7C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vhaddps         , "vhaddps"         , Enc(AvxRvm_OptL)     , O_F20F00(7C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vhsubpd         , "vhsubpd"         , Enc(AvxRvm_OptL)     , O_660F00(7D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vhsubps         , "vhsubps"         , Enc(AvxRvm_OptL)     , O_F20F00(7D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vinsertf128     , "vinsertf128"     , Enc(AvxRvmi)         , O_660F3A(18,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(Ymm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vinserti128     , "vinserti128"     , Enc(AvxRvmi)         , O_660F3A(38,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(Ymm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vinsertps       , "vinsertps"       , Enc(AvxRvmi)         , O_660F3A(21,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vlddqu          , "vlddqu"          , Enc(AvxRm_OptL)      , O_F20F00(F0,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Mem)            , U                 , U                 , U                 ),
  INST(Vldmxcsr        , "vldmxcsr"        , Enc(AvxM)            , O_000F00(AE,2,_,_,_), U                   , F(RO)|F(Avx)|F(Volatile)           , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Vmaskmovdqu     , "vmaskmovdqu"     , Enc(AvxRm)           , O_660F00(F7,U,_,_,_), U                   , F(RO)|F(Avx)|F(Special)            , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , U                 , U                 , U                 ),
  INST(Vmaskmovpd      , "vmaskmovpd"      , Enc(AvxRvmMvr_OptL)  , O_660F38(2D,U,_,_,_), O_660F38(2F,U,_,_,_), F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmaskmovps      , "vmaskmovps"      , Enc(AvxRvmMvr_OptL)  , O_660F38(2C,U,_,_,_), O_660F38(2E,U,_,_,_), F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmaxpd          , "vmaxpd"          , Enc(AvxRvm_OptL)     , O_660F00(5F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmaxps          , "vmaxps"          , Enc(AvxRvm_OptL)     , O_000F00(5F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmaxsd          , "vmaxsd"          , Enc(AvxRvm_OptL)     , O_F20F00(5F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmaxss          , "vmaxss"          , Enc(AvxRvm_OptL)     , O_F30F00(5F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vminpd          , "vminpd"          , Enc(AvxRvm_OptL)     , O_660F00(5D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vminps          , "vminps"          , Enc(AvxRvm_OptL)     , O_000F00(5D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vminsd          , "vminsd"          , Enc(AvxRvm_OptL)     , O_F20F00(5D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vminss          , "vminss"          , Enc(AvxRvm_OptL)     , O_F30F00(5D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmovapd         , "vmovapd"         , Enc(AvxRmMr_OptL)    , O_660F00(28,U,_,_,_), O_660F00(29,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmovaps         , "vmovaps"         , Enc(AvxRmMr_OptL)    , O_000F00(28,U,_,_,_), O_000F00(29,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmovd           , "vmovd"           , Enc(AvxMovDQ)        , O_660F00(6E,U,_,_,_), O_660F00(7E,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vmovddup        , "vmovddup"        , Enc(AvxRm_OptL)      , O_F20F00(12,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmovdqa         , "vmovdqa"         , Enc(AvxRmMr_OptL)    , O_660F00(6F,U,_,_,_), O_660F00(7F,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmovdqu         , "vmovdqu"         , Enc(AvxRmMr_OptL)    , O_F30F00(6F,U,_,_,_), O_F30F00(7F,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmovhlps        , "vmovhlps"        , Enc(AvxRvm)          , O_000F00(12,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(Xmm)            , U                 , U                 ),
  INST(Vmovhpd         , "vmovhpd"         , Enc(AvxRvmMr)        , O_660F00(16,U,_,_,_), O_660F00(17,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(Xmm)            , O(Mem)            , U                 , U                 ),
  INST(Vmovhps         , "vmovhps"         , Enc(AvxRvmMr)        , O_000F00(16,U,_,_,_), O_000F00(17,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(Xmm)            , O(Mem)            , U                 , U                 ),
  INST(Vmovlhps        , "vmovlhps"        , Enc(AvxRvm)          , O_000F00(16,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(Xmm)            , U                 , U                 ),
  INST(Vmovlpd         , "vmovlpd"         , Enc(AvxRvmMr)        , O_660F00(12,U,_,_,_), O_660F00(13,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(Xmm)            , O(Mem)            , U                 , U                 ),
  INST(Vmovlps         , "vmovlps"         , Enc(AvxRvmMr)        , O_000F00(12,U,_,_,_), O_000F00(13,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(Xmm)            , O(Mem)            , U                 , U                 ),
  INST(Vmovmskpd       , "vmovmskpd"       , Enc(AvxRm_OptL)      , O_660F00(50,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Gqd)            , O(Xy)             , U                 , U                 , U                 ),
  INST(Vmovmskps       , "vmovmskps"       , Enc(AvxRm_OptL)      , O_000F00(50,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Gqd)            , O(Xy)             , U                 , U                 , U                 ),
  INST(Vmovntdq        , "vmovntdq"        , Enc(AvxMr_OptL)      , O_660F00(E7,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Mem)            , O(Xy)             , U                 , U                 , U                 ),
  INST(Vmovntdqa       , "vmovntdqa"       , Enc(AvxRm_OptL)      , O_660F38(2A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Mem)            , U                 , U                 , U                 ),
  INST(Vmovntpd        , "vmovntpd"        , Enc(AvxMr_OptL)      , O_660F00(2B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Mem)            , O(Xy)             , U                 , U                 , U                 ),
  INST(Vmovntps        , "vmovntps"        , Enc(AvxMr_OptL)      , O_000F00(2B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Mem)            , O(Xy)             , U                 , U                 , U                 ),
  INST(Vmovq           , "vmovq"           , Enc(AvxMovDQ)        , O_660F00(6E,U,_,W,_), O_660F00(7E,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vmovsd          , "vmovsd"          , Enc(AvxMovSsSd)      , O_F20F00(10,U,_,_,_), O_F20F00(11,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(XmmMem)         , O(Xmm)            , U                 , U                 ),
  INST(Vmovshdup       , "vmovshdup"       , Enc(AvxRm_OptL)      , O_F30F00(16,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmovsldup       , "vmovsldup"       , Enc(AvxRm_OptL)      , O_F30F00(12,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmovss          , "vmovss"          , Enc(AvxMovSsSd)      , O_F30F00(10,U,_,_,_), O_F30F00(11,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XmmMem)         , O(Xmm)            , O(Xmm)            , U                 , U                 ),
  INST(Vmovupd         , "vmovupd"         , Enc(AvxRmMr_OptL)    , O_660F00(10,U,_,_,_), O_660F00(11,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmovups         , "vmovups"         , Enc(AvxRmMr_OptL)    , O_000F00(10,U,_,_,_), O_000F00(11,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vmpsadbw        , "vmpsadbw"        , Enc(AvxRvmi_OptL)    , O_660F3A(42,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vmulpd          , "vmulpd"          , Enc(AvxRvm_OptL)     , O_660F00(59,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmulps          , "vmulps"          , Enc(AvxRvm_OptL)     , O_000F00(59,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmulsd          , "vmulsd"          , Enc(AvxRvm_OptL)     , O_F20F00(59,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vmulss          , "vmulss"          , Enc(AvxRvm_OptL)     , O_F30F00(59,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vorpd           , "vorpd"           , Enc(AvxRvm_OptL)     , O_660F00(56,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vorps           , "vorps"           , Enc(AvxRvm_OptL)     , O_000F00(56,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpabsb          , "vpabsb"          , Enc(AvxRm_OptL)      , O_660F38(1C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpabsd          , "vpabsd"          , Enc(AvxRm_OptL)      , O_660F38(1E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpabsw          , "vpabsw"          , Enc(AvxRm_OptL)      , O_660F38(1D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpackssdw       , "vpackssdw"       , Enc(AvxRvm_OptL)     , O_660F00(6B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpacksswb       , "vpacksswb"       , Enc(AvxRvm_OptL)     , O_660F00(63,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpackusdw       , "vpackusdw"       , Enc(AvxRvm_OptL)     , O_660F38(2B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpackuswb       , "vpackuswb"       , Enc(AvxRvm_OptL)     , O_660F00(67,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpaddb          , "vpaddb"          , Enc(AvxRvm_OptL)     , O_660F00(FC,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpaddd          , "vpaddd"          , Enc(AvxRvm_OptL)     , O_660F00(FE,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpaddq          , "vpaddq"          , Enc(AvxRvm_OptL)     , O_660F00(D4,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpaddsb         , "vpaddsb"         , Enc(AvxRvm_OptL)     , O_660F00(EC,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpaddsw         , "vpaddsw"         , Enc(AvxRvm_OptL)     , O_660F00(ED,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpaddusb        , "vpaddusb"        , Enc(AvxRvm_OptL)     , O_660F00(DC,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpaddusw        , "vpaddusw"        , Enc(AvxRvm_OptL)     , O_660F00(DD,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpaddw          , "vpaddw"          , Enc(AvxRvm_OptL)     , O_660F00(FD,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpalignr        , "vpalignr"        , Enc(AvxRvmi_OptL)    , O_660F3A(0F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vpand           , "vpand"           , Enc(AvxRvm_OptL)     , O_660F00(DB,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpandn          , "vpandn"          , Enc(AvxRvm_OptL)     , O_660F00(DF,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpavgb          , "vpavgb"          , Enc(AvxRvm_OptL)     , O_660F00(E0,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpavgw          , "vpavgw"          , Enc(AvxRvm_OptL)     , O_660F00(E3,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpblendd        , "vpblendd"        , Enc(AvxRvmi_OptL)    , O_660F3A(02,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vpblendvb       , "vpblendvb"       , Enc(AvxRvmr)         , O_660F3A(4C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Xy)             , U                 ),
  INST(Vpblendw        , "vpblendw"        , Enc(AvxRvmi_OptL)    , O_660F3A(0E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vpbroadcastb    , "vpbroadcastb"    , Enc(AvxRm_OptL)      , O_660F38(78,U,_,_,0), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vpbroadcastd    , "vpbroadcastd"    , Enc(AvxRm_OptL)      , O_660F38(58,U,_,_,0), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vpbroadcastq    , "vpbroadcastq"    , Enc(AvxRm_OptL)      , O_660F38(59,U,_,_,1), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vpbroadcastw    , "vpbroadcastw"    , Enc(AvxRm_OptL)      , O_660F38(79,U,_,_,0), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vpclmulqdq      , "vpclmulqdq"      , Enc(AvxRvmi)         , O_660F3A(44,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vpcmov          , "vpcmov"          , Enc(XopRvrmRvmr_OptL), O_00_M08(A2,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vpcmpeqb        , "vpcmpeqb"        , Enc(AvxRvm_OptL)     , O_660F00(74,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpcmpeqd        , "vpcmpeqd"        , Enc(AvxRvm_OptL)     , O_660F00(76,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpcmpeqq        , "vpcmpeqq"        , Enc(AvxRvm_OptL)     , O_660F38(29,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpcmpeqw        , "vpcmpeqw"        , Enc(AvxRvm_OptL)     , O_660F00(75,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpcmpestri      , "vpcmpestri"      , Enc(AvxRmi)          , O_660F3A(61,U,_,_,_), U                   , F(WO)|F(Avx)|F(Special)            , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Vpcmpestrm      , "vpcmpestrm"      , Enc(AvxRmi)          , O_660F3A(60,U,_,_,_), U                   , F(WO)|F(Avx)|F(Special)            , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Vpcmpgtb        , "vpcmpgtb"        , Enc(AvxRvm_OptL)     , O_660F00(64,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpcmpgtd        , "vpcmpgtd"        , Enc(AvxRvm_OptL)     , O_660F00(66,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpcmpgtq        , "vpcmpgtq"        , Enc(AvxRvm_OptL)     , O_660F38(37,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpcmpgtw        , "vpcmpgtw"        , Enc(AvxRvm_OptL)     , O_660F00(65,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpcmpistri      , "vpcmpistri"      , Enc(AvxRmi)          , O_660F3A(63,U,_,_,_), U                   , F(WO)|F(Avx)|F(Special)            , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Vpcmpistrm      , "vpcmpistrm"      , Enc(AvxRmi)          , O_660F3A(62,U,_,_,_), U                   , F(WO)|F(Avx)|F(Special)            , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Vpcomb          , "vpcomb"          , Enc(XopRvmi)         , O_00_M08(CC,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vpcomd          , "vpcomd"          , Enc(XopRvmi)         , O_00_M08(CE,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vpcomq          , "vpcomq"          , Enc(XopRvmi)         , O_00_M08(CF,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vpcomub         , "vpcomub"         , Enc(XopRvmi)         , O_00_M08(EC,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vpcomud         , "vpcomud"         , Enc(XopRvmi)         , O_00_M08(EE,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vpcomuq         , "vpcomuq"         , Enc(XopRvmi)         , O_00_M08(EF,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vpcomuw         , "vpcomuw"         , Enc(XopRvmi)         , O_00_M08(ED,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vpcomw          , "vpcomw"          , Enc(XopRvmi)         , O_00_M08(CD,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vperm2f128      , "vperm2f128"      , Enc(AvxRvmi)         , O_660F3A(06,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(Ymm)            , O(YmmMem)         , O(Imm)            , U                 ),
  INST(Vperm2i128      , "vperm2i128"      , Enc(AvxRvmi)         , O_660F3A(46,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(Ymm)            , O(YmmMem)         , O(Imm)            , U                 ),
  INST(Vpermd          , "vpermd"          , Enc(AvxRvm)          , O_660F38(36,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(Ymm)            , O(YmmMem)         , U                 , U                 ),
  INST(Vpermil2pd      , "vpermil2pd"      , Enc(AvxRvrmRvmr_OptL), O_66_M03(49,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vpermil2ps      , "vpermil2ps"      , Enc(AvxRvrmRvmr_OptL), O_66_M03(48,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(XyMem)          , U                 ),
  INST(Vpermilpd       , "vpermilpd"       , Enc(AvxRvmRmi_OptL)  , O_660F38(0D,U,_,_,_), O_660F3A(05,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpermilps       , "vpermilps"       , Enc(AvxRvmRmi_OptL)  , O_660F38(0C,U,_,_,_), O_660F3A(04,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpermpd         , "vpermpd"         , Enc(AvxRmi)          , O_660F3A(01,U,L,W,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(YmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Vpermps         , "vpermps"         , Enc(AvxRvm)          , O_660F38(16,U,L,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(Ymm)            , O(YmmMem)         , U                 , U                 ),
  INST(Vpermq          , "vpermq"          , Enc(AvxRmi)          , O_660F3A(00,U,L,W,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Ymm)            , O(YmmMem)         , O(Imm)            , U                 , U                 ),
  INST(Vpextrb         , "vpextrb"         , Enc(AvxMri)          , O_660F3A(14,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(GqdwbMem)       , O(Xmm)            , O(Imm)            , U                 , U                 ),
  INST(Vpextrd         , "vpextrd"         , Enc(AvxMri)          , O_660F3A(16,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(GqdMem)         , O(Xmm)            , O(Imm)            , U                 , U                 ),
  INST(Vpextrq         , "vpextrq"         , Enc(AvxMri)          , O_660F3A(16,U,_,W,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(GqMem)          , O(Xmm)            , O(Imm)            , U                 , U                 ),
  INST(Vpextrw         , "vpextrw"         , Enc(AvxMri)          , O_660F3A(15,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(GqdwMem)        , O(Xmm)            , O(Imm)            , U                 , U                 ),
  INST(Vpgatherdd      , "vpgatherdd"      , Enc(AvxGather)       , O_660F38(90,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Mem)            , O(Xy)             , U                 , U                 ),
  INST(Vpgatherdq      , "vpgatherdq"      , Enc(AvxGather)       , O_660F38(90,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Mem)            , O(Xy)             , U                 , U                 ),
  INST(Vpgatherqd      , "vpgatherqd"      , Enc(AvxGatherEx)     , O_660F38(91,U,_,_,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Mem)            , O(Xmm)            , U                 , U                 ),
  INST(Vpgatherqq      , "vpgatherqq"      , Enc(AvxGather)       , O_660F38(91,U,_,W,_), U                   , F(RW)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Mem)            , O(Xy)             , U                 , U                 ),
  INST(Vphaddbd        , "vphaddbd"        , Enc(XopRm)           , O_00_M09(C2,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddbq        , "vphaddbq"        , Enc(XopRm)           , O_00_M09(C3,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddbw        , "vphaddbw"        , Enc(XopRm)           , O_00_M09(C1,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddd         , "vphaddd"         , Enc(AvxRvm_OptL)     , O_660F38(02,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vphadddq        , "vphadddq"        , Enc(XopRm)           , O_00_M09(CB,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddsw        , "vphaddsw"        , Enc(AvxRvm_OptL)     , O_660F38(03,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vphaddubd       , "vphaddubd"       , Enc(XopRm)           , O_00_M09(D2,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddubq       , "vphaddubq"       , Enc(XopRm)           , O_00_M09(D3,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddubw       , "vphaddubw"       , Enc(XopRm)           , O_00_M09(D1,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddudq       , "vphaddudq"       , Enc(XopRm)           , O_00_M09(DB,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphadduwd       , "vphadduwd"       , Enc(XopRm)           , O_00_M09(D6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphadduwq       , "vphadduwq"       , Enc(XopRm)           , O_00_M09(D7,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddw         , "vphaddw"         , Enc(AvxRvm_OptL)     , O_660F38(01,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vphaddwd        , "vphaddwd"        , Enc(XopRm)           , O_00_M09(C6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphaddwq        , "vphaddwq"        , Enc(XopRm)           , O_00_M09(C7,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphminposuw     , "vphminposuw"     , Enc(AvxRm)           , O_660F38(41,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphsubbw        , "vphsubbw"        , Enc(XopRm)           , O_00_M09(E1,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphsubd         , "vphsubd"         , Enc(AvxRvm_OptL)     , O_660F38(06,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vphsubdq        , "vphsubdq"        , Enc(XopRm)           , O_00_M09(E3,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vphsubsw        , "vphsubsw"        , Enc(AvxRvm_OptL)     , O_660F38(07,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vphsubw         , "vphsubw"         , Enc(AvxRvm_OptL)     , O_660F38(05,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vphsubwd        , "vphsubwd"        , Enc(XopRm)           , O_00_M09(E2,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vpinsrb         , "vpinsrb"         , Enc(AvxRvmi)         , O_660F3A(20,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(GqdwbMem)       , O(Imm)            , U                 ),
  INST(Vpinsrd         , "vpinsrd"         , Enc(AvxRvmi)         , O_660F3A(22,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(GqdMem)         , O(Imm)            , U                 ),
  INST(Vpinsrq         , "vpinsrq"         , Enc(AvxRvmi)         , O_660F3A(22,U,_,W,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(GqMem)          , O(Imm)            , U                 ),
  INST(Vpinsrw         , "vpinsrw"         , Enc(AvxRvmi)         , O_660F00(C4,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(GqdwMem)        , O(Imm)            , U                 ),
  INST(Vpmacsdd        , "vpmacsdd"        , Enc(XopRvmr)         , O_00_M08(9E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacsdqh       , "vpmacsdqh"       , Enc(XopRvmr)         , O_00_M08(9F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacsdql       , "vpmacsdql"       , Enc(XopRvmr)         , O_00_M08(97,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacssdd       , "vpmacssdd"       , Enc(XopRvmr)         , O_00_M08(8E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacssdqh      , "vpmacssdqh"      , Enc(XopRvmr)         , O_00_M08(8F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacssdql      , "vpmacssdql"      , Enc(XopRvmr)         , O_00_M08(87,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacsswd       , "vpmacsswd"       , Enc(XopRvmr)         , O_00_M08(86,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacssww       , "vpmacssww"       , Enc(XopRvmr)         , O_00_M08(85,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacswd        , "vpmacswd"        , Enc(XopRvmr)         , O_00_M08(96,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmacsww        , "vpmacsww"        , Enc(XopRvmr)         , O_00_M08(95,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmadcsswd      , "vpmadcsswd"      , Enc(XopRvmr)         , O_00_M08(A6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmadcswd       , "vpmadcswd"       , Enc(XopRvmr)         , O_00_M08(B6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Xmm)            , U                 ),
  INST(Vpmaddubsw      , "vpmaddubsw"      , Enc(AvxRvm_OptL)     , O_660F38(04,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaddwd        , "vpmaddwd"        , Enc(AvxRvm_OptL)     , O_660F00(F5,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaskmovd      , "vpmaskmovd"      , Enc(AvxRvmMvr_OptL)  , O_660F38(8C,U,_,_,_), O_660F38(8E,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaskmovq      , "vpmaskmovq"      , Enc(AvxRvmMvr_OptL)  , O_660F38(8C,U,_,W,_), O_660F38(8E,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(XyMem)          , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaxsb         , "vpmaxsb"         , Enc(AvxRvm_OptL)     , O_660F38(3C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaxsd         , "vpmaxsd"         , Enc(AvxRvm_OptL)     , O_660F38(3D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaxsw         , "vpmaxsw"         , Enc(AvxRvm_OptL)     , O_660F00(EE,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaxub         , "vpmaxub"         , Enc(AvxRvm_OptL)     , O_660F00(DE,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaxud         , "vpmaxud"         , Enc(AvxRvm_OptL)     , O_660F38(3F,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmaxuw         , "vpmaxuw"         , Enc(AvxRvm_OptL)     , O_660F38(3E,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpminsb         , "vpminsb"         , Enc(AvxRvm_OptL)     , O_660F38(38,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpminsd         , "vpminsd"         , Enc(AvxRvm_OptL)     , O_660F38(39,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpminsw         , "vpminsw"         , Enc(AvxRvm_OptL)     , O_660F00(EA,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpminub         , "vpminub"         , Enc(AvxRvm_OptL)     , O_660F00(DA,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpminud         , "vpminud"         , Enc(AvxRvm_OptL)     , O_660F38(3B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpminuw         , "vpminuw"         , Enc(AvxRvm_OptL)     , O_660F38(3A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmovmskb       , "vpmovmskb"       , Enc(AvxRm_OptL)      , O_660F00(D7,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Gqd)            , O(Xy)             , U                 , U                 , U                 ),
  INST(Vpmovsxbd       , "vpmovsxbd"       , Enc(AvxRm_OptL)      , O_660F38(21,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovsxbq       , "vpmovsxbq"       , Enc(AvxRm_OptL)      , O_660F38(22,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovsxbw       , "vpmovsxbw"       , Enc(AvxRm_OptL)      , O_660F38(20,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovsxdq       , "vpmovsxdq"       , Enc(AvxRm_OptL)      , O_660F38(25,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovsxwd       , "vpmovsxwd"       , Enc(AvxRm_OptL)      , O_660F38(23,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovsxwq       , "vpmovsxwq"       , Enc(AvxRm_OptL)      , O_660F38(24,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovzxbd       , "vpmovzxbd"       , Enc(AvxRm_OptL)      , O_660F38(31,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovzxbq       , "vpmovzxbq"       , Enc(AvxRm_OptL)      , O_660F38(32,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovzxbw       , "vpmovzxbw"       , Enc(AvxRm_OptL)      , O_660F38(30,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovzxdq       , "vpmovzxdq"       , Enc(AvxRm_OptL)      , O_660F38(35,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovzxwd       , "vpmovzxwd"       , Enc(AvxRm_OptL)      , O_660F38(33,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmovzxwq       , "vpmovzxwq"       , Enc(AvxRm_OptL)      , O_660F38(34,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpmuldq         , "vpmuldq"         , Enc(AvxRvm_OptL)     , O_660F38(28,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmulhrsw       , "vpmulhrsw"       , Enc(AvxRvm_OptL)     , O_660F38(0B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmulhuw        , "vpmulhuw"        , Enc(AvxRvm_OptL)     , O_660F00(E4,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmulhw         , "vpmulhw"         , Enc(AvxRvm_OptL)     , O_660F00(E5,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmulld         , "vpmulld"         , Enc(AvxRvm_OptL)     , O_660F38(40,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmullw         , "vpmullw"         , Enc(AvxRvm_OptL)     , O_660F00(D5,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpmuludq        , "vpmuludq"        , Enc(AvxRvm_OptL)     , O_660F00(F4,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpor            , "vpor"            , Enc(AvxRvm_OptL)     , O_660F00(EB,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpperm          , "vpperm"          , Enc(XopRvrmRvmr)     , O_00_M08(A3,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 ),
  INST(Vprotb          , "vprotb"          , Enc(XopRvmRmi)       , O_00_M09(90,U,_,_,_), O_00_M08(C0,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)|O(Imm)  , U                 , U                 ),
  INST(Vprotd          , "vprotd"          , Enc(XopRvmRmi)       , O_00_M09(92,U,_,_,_), O_00_M08(C2,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)|O(Imm)  , U                 , U                 ),
  INST(Vprotq          , "vprotq"          , Enc(XopRvmRmi)       , O_00_M09(93,U,_,_,_), O_00_M08(C3,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)|O(Imm)  , U                 , U                 ),
  INST(Vprotw          , "vprotw"          , Enc(XopRvmRmi)       , O_00_M09(91,U,_,_,_), O_00_M08(C1,U,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)|O(Imm)  , U                 , U                 ),
  INST(Vpsadbw         , "vpsadbw"         , Enc(AvxRvm_OptL)     , O_660F00(F6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpshab          , "vpshab"          , Enc(XopRvmRmv)       , O_00_M09(98,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                 ),
  INST(Vpshad          , "vpshad"          , Enc(XopRvmRmv)       , O_00_M09(9A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                 ),
  INST(Vpshaq          , "vpshaq"          , Enc(XopRvmRmv)       , O_00_M09(9B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                 ),
  INST(Vpshaw          , "vpshaw"          , Enc(XopRvmRmv)       , O_00_M09(99,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                 ),
  INST(Vpshlb          , "vpshlb"          , Enc(XopRvmRmv)       , O_00_M09(94,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                 ),
  INST(Vpshld          , "vpshld"          , Enc(XopRvmRmv)       , O_00_M09(96,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                 ),
  INST(Vpshlq          , "vpshlq"          , Enc(XopRvmRmv)       , O_00_M09(97,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                 ),
  INST(Vpshlw          , "vpshlw"          , Enc(XopRvmRmv)       , O_00_M09(95,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , O(XmmMem)         , U                 , U                 ),
  INST(Vpshufb         , "vpshufb"         , Enc(AvxRvm_OptL)     , O_660F38(00,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpshufd         , "vpshufd"         , Enc(AvxRmi_OptL)     , O_660F00(70,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                 ),
  INST(Vpshufhw        , "vpshufhw"        , Enc(AvxRmi_OptL)     , O_F30F00(70,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                 ),
  INST(Vpshuflw        , "vpshuflw"        , Enc(AvxRmi_OptL)     , O_F20F00(70,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                 ),
  INST(Vpsignb         , "vpsignb"         , Enc(AvxRvm_OptL)     , O_660F38(08,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsignd         , "vpsignd"         , Enc(AvxRvm_OptL)     , O_660F38(0A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsignw         , "vpsignw"         , Enc(AvxRvm_OptL)     , O_660F38(09,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpslld          , "vpslld"          , Enc(AvxRvmVmi_OptL)  , O_660F00(F2,U,_,_,_), O_660F00(72,6,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpslldq         , "vpslldq"         , Enc(AvxVmi_OptL)     , O_660F00(73,7,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                 ),
  INST(Vpsllq          , "vpsllq"          , Enc(AvxRvmVmi_OptL)  , O_660F00(F3,U,_,_,_), O_660F00(73,6,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpsllvd         , "vpsllvd"         , Enc(AvxRvm_OptL)     , O_660F38(47,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsllvq         , "vpsllvq"         , Enc(AvxRvm_OptL)     , O_660F38(47,U,_,W,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsllw          , "vpsllw"          , Enc(AvxRvmVmi_OptL)  , O_660F00(F1,U,_,_,_), O_660F00(71,6,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpsrad          , "vpsrad"          , Enc(AvxRvmVmi_OptL)  , O_660F00(E2,U,_,_,_), O_660F00(72,4,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpsravd         , "vpsravd"         , Enc(AvxRvm_OptL)     , O_660F38(46,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsraw          , "vpsraw"          , Enc(AvxRvmVmi_OptL)  , O_660F00(E1,U,_,_,_), O_660F00(71,4,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpsrld          , "vpsrld"          , Enc(AvxRvmVmi_OptL)  , O_660F00(D2,U,_,_,_), O_660F00(72,2,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpsrldq         , "vpsrldq"         , Enc(AvxVmi_OptL)     , O_660F00(73,3,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                 ),
  INST(Vpsrlq          , "vpsrlq"          , Enc(AvxRvmVmi_OptL)  , O_660F00(D3,U,_,_,_), O_660F00(73,2,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpsrlvd         , "vpsrlvd"         , Enc(AvxRvm_OptL)     , O_660F38(45,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsrlvq         , "vpsrlvq"         , Enc(AvxRvm_OptL)     , O_660F38(45,U,_,W,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsrlw          , "vpsrlw"          , Enc(AvxRvmVmi_OptL)  , O_660F00(D1,U,_,_,_), O_660F00(71,2,_,_,_), F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(XyMem)|O(Imm)   , U                 , U                 ),
  INST(Vpsubb          , "vpsubb"          , Enc(AvxRvm_OptL)     , O_660F00(F8,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsubd          , "vpsubd"          , Enc(AvxRvm_OptL)     , O_660F00(FA,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsubq          , "vpsubq"          , Enc(AvxRvm_OptL)     , O_660F00(FB,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsubsb         , "vpsubsb"         , Enc(AvxRvm_OptL)     , O_660F00(E8,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsubsw         , "vpsubsw"         , Enc(AvxRvm_OptL)     , O_660F00(E9,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsubusb        , "vpsubusb"        , Enc(AvxRvm_OptL)     , O_660F00(D8,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsubusw        , "vpsubusw"        , Enc(AvxRvm_OptL)     , O_660F00(D9,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpsubw          , "vpsubw"          , Enc(AvxRvm_OptL)     , O_660F00(F9,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vptest          , "vptest"          , Enc(AvxRm_OptL)      , O_660F38(17,U,_,_,_), U                   , F(RO)|F(Avx)                       , EF(WWWWWW__), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vpunpckhbw      , "vpunpckhbw"      , Enc(AvxRvm_OptL)     , O_660F00(68,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpunpckhdq      , "vpunpckhdq"      , Enc(AvxRvm_OptL)     , O_660F00(6A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpunpckhqdq     , "vpunpckhqdq"     , Enc(AvxRvm_OptL)     , O_660F00(6D,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpunpckhwd      , "vpunpckhwd"      , Enc(AvxRvm_OptL)     , O_660F00(69,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpunpcklbw      , "vpunpcklbw"      , Enc(AvxRvm_OptL)     , O_660F00(60,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpunpckldq      , "vpunpckldq"      , Enc(AvxRvm_OptL)     , O_660F00(62,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpunpcklqdq     , "vpunpcklqdq"     , Enc(AvxRvm_OptL)     , O_660F00(6C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpunpcklwd      , "vpunpcklwd"      , Enc(AvxRvm_OptL)     , O_660F00(61,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vpxor           , "vpxor"           , Enc(AvxRvm_OptL)     , O_660F00(EF,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vrcpps          , "vrcpps"          , Enc(AvxRm_OptL)      , O_000F00(53,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vrcpss          , "vrcpss"          , Enc(AvxRvm)          , O_F30F00(53,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vroundpd        , "vroundpd"        , Enc(AvxRmi_OptL)     , O_660F3A(09,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                 ),
  INST(Vroundps        , "vroundps"        , Enc(AvxRmi_OptL)     , O_660F3A(08,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , O(Imm)            , U                 , U                 ),
  INST(Vroundsd        , "vroundsd"        , Enc(AvxRvmi)         , O_660F3A(0B,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vroundss        , "vroundss"        , Enc(AvxRvmi)         , O_660F3A(0A,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , O(Imm)            , U                 ),
  INST(Vrsqrtps        , "vrsqrtps"        , Enc(AvxRm_OptL)      , O_000F00(52,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vrsqrtss        , "vrsqrtss"        , Enc(AvxRvm)          , O_F30F00(52,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vshufpd         , "vshufpd"         , Enc(AvxRvmi_OptL)    , O_660F00(C6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vshufps         , "vshufps"         , Enc(AvxRvmi_OptL)    , O_000F00(C6,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , O(Imm)            , U                 ),
  INST(Vsqrtpd         , "vsqrtpd"         , Enc(AvxRm_OptL)      , O_660F00(51,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vsqrtps         , "vsqrtps"         , Enc(AvxRm_OptL)      , O_000F00(51,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vsqrtsd         , "vsqrtsd"         , Enc(AvxRvm)          , O_F20F00(51,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vsqrtss         , "vsqrtss"         , Enc(AvxRvm)          , O_F30F00(51,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vstmxcsr        , "vstmxcsr"        , Enc(AvxM)            , O_000F00(AE,3,_,_,_), U                   , F(Volatile)|F(Avx)                 , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Vsubpd          , "vsubpd"          , Enc(AvxRvm_OptL)     , O_660F00(5C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vsubps          , "vsubps"          , Enc(AvxRvm_OptL)     , O_000F00(5C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vsubsd          , "vsubsd"          , Enc(AvxRvm)          , O_F20F00(5C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vsubss          , "vsubss"          , Enc(AvxRvm)          , O_F30F00(5C,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xmm)            , O(Xmm)            , O(XmmMem)         , U                 , U                 ),
  INST(Vtestpd         , "vtestpd"         , Enc(AvxRm_OptL)      , O_660F38(0F,U,_,_,_), U                   , F(RO)|F(Avx)                       , EF(WWWWWW__), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vtestps         , "vtestps"         , Enc(AvxRm_OptL)      , O_660F38(0E,U,_,_,_), U                   , F(RO)|F(Avx)                       , EF(WWWWWW__), 0 , 0 , O(Xy)             , O(XyMem)          , U                 , U                 , U                 ),
  INST(Vucomisd        , "vucomisd"        , Enc(AvxRm)           , O_660F00(2E,U,_,_,_), U                   , F(RO)|F(Avx)                       , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vucomiss        , "vucomiss"        , Enc(AvxRm)           , O_000F00(2E,U,_,_,_), U                   , F(RO)|F(Avx)                       , EF(WWWWWW__), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Vunpckhpd       , "vunpckhpd"       , Enc(AvxRvm_OptL)     , O_660F00(15,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vunpckhps       , "vunpckhps"       , Enc(AvxRvm_OptL)     , O_000F00(15,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vunpcklpd       , "vunpcklpd"       , Enc(AvxRvm_OptL)     , O_660F00(14,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vunpcklps       , "vunpcklps"       , Enc(AvxRvm_OptL)     , O_000F00(14,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vxorpd          , "vxorpd"          , Enc(AvxRvm_OptL)     , O_660F00(57,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vxorps          , "vxorps"          , Enc(AvxRvm_OptL)     , O_000F00(57,U,_,_,_), U                   , F(WO)|F(Avx)                       , EF(________), 0 , 0 , O(Xy)             , O(Xy)             , O(XyMem)          , U                 , U                 ),
  INST(Vzeroall        , "vzeroall"        , Enc(AvxOp)           , O_000F00(77,U,L,_,_), U                   , F(Volatile)|F(Avx)                 , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Vzeroupper      , "vzeroupper"      , Enc(AvxOp)           , O_000F00(77,U,_,_,_), U                   , F(Volatile)|F(Avx)                 , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Wrfsbase        , "wrfsbase"        , Enc(X86Rm)           , O_F30F00(AE,2,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(________), 0 , 0 , O(Gqd)            , U                 , U                 , U                 , U                 ),
  INST(Wrgsbase        , "wrgsbase"        , Enc(X86Rm)           , O_F30F00(AE,3,_,_,_), U                   , F(RO)|F(Volatile)                  , EF(________), 0 , 0 , O(Gqd)            , U                 , U                 , U                 , U                 ),
  INST(Xadd            , "xadd"            , Enc(X86Xadd)         , O_000F00(C0,U,_,_,_), U                   , F(RW)|F(Xchg)|F(Lock)              , EF(WWWWWW__), 0 , 0 , O(GqdwbMem)       , O(Gqdwb)          , U                 , U                 , U                 ),
  INST(Xchg            , "xchg"            , Enc(X86Xchg)         , O_000000(86,U,_,_,_), U                   , F(RW)|F(Xchg)|F(Lock)              , EF(________), 0 , 0 , O(GqdwbMem)       , O(Gqdwb)          , U                 , U                 , U                 ),
  INST(Xgetbv          , "xgetbv"          , Enc(X86Op)           , O_000F01(D0,U,_,_,_), U                   , F(WO)|F(Special)                   , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 ),
  INST(Xor             , "xor"             , Enc(X86Arith)        , O_000000(30,6,_,_,_), U                   , F(RW)|F(Lock)                      , EF(WWWUWW__), 0 , 0 , O(GqdwbMem)       , O(GqdwbMem)|O(Imm), U                 , U                 , U                 ),
  INST(Xorpd           , "xorpd"           , Enc(SimdRm)          , O_660F00(57,U,_,_,_), U                   , F(RW)|F(None)                      , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Xorps           , "xorps"           , Enc(SimdRm)          , O_000F00(57,U,_,_,_), U                   , F(RW)|F(None)                      , EF(________), 0 , 0 , O(Xmm)            , O(XmmMem)         , U                 , U                 , U                 ),
  INST(Xrstor          , "xrstor"          , Enc(X86M)            , O_000F00(AE,5,_,_,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Xrstor64        , "xrstor64"        , Enc(X86M)            , O_000F00(AE,5,_,W,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Xsave           , "xsave"           , Enc(X86M)            , O_000F00(AE,4,_,_,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Xsave64         , "xsave64"         , Enc(X86M)            , O_000F00(AE,4,_,W,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Xsaveopt        , "xsaveopt"        , Enc(X86M)            , O_000F00(AE,6,_,_,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Xsaveopt64      , "xsaveopt64"      , Enc(X86M)            , O_000F00(AE,6,_,W,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , O(Mem)            , U                 , U                 , U                 , U                 ),
  INST(Xsetbv          , "xsetbv"          , Enc(X86Op)           , O_000F01(D1,U,_,_,_), U                   , F(RO)|F(Volatile)|F(Special)       , EF(________), 0 , 0 , U                 , U                 , U                 , U                 , U                 )
};
#undef INST

#undef O_00_X
#undef O_9B_X

#undef O_66_M03
#undef O_00_M09
#undef O_00_M08

#undef O_F30F3A
#undef O_F30F38
#undef O_F30F00
#undef O_F30000
#undef O_F20F3A
#undef O_F20F38
#undef O_F20F00
#undef O_F20000
#undef O_9B0000
#undef O_660F3A
#undef O_660F38
#undef O_660F00
#undef O_660000
#undef O_000F3A
#undef O_000F38
#undef O_000F0F
#undef O_000F01
#undef O_000F00
#undef O_000000

#undef Enc
#undef EF
#undef O
#undef A
#undef F
#undef U

// ============================================================================
// [asmjit::X86Cond]
// ============================================================================

#define CC_TO_INST(_Inst_) { \
  _Inst_##o  , _Inst_##no , _Inst_##b  , _Inst_##ae , \
  _Inst_##e  , _Inst_##ne , _Inst_##be , _Inst_##a  , \
  _Inst_##s  , _Inst_##ns , _Inst_##pe , _Inst_##po , \
  _Inst_##l  , _Inst_##ge , _Inst_##le , _Inst_##g  , \
  kInstIdNone, kInstIdNone, kInstIdNone, kInstIdNone  \
}

const uint32_t _x86ReverseCond[20] = {
  /* O|NO|B|AE    -> */ kX86CondO, kX86CondNO, kX86CondA , kX86CondBE,
  /* E|NE|BE|A    -> */ kX86CondE, kX86CondNE, kX86CondAE, kX86CondB ,
  /* S|NS|PE|PO   -> */ kX86CondS, kX86CondNS, kX86CondPE, kX86CondPO,
  /* L|GE|LE|G    -> */ kX86CondG, kX86CondLE, kX86CondGE, kX86CondL ,
  /* Unord|!Unord -> */ kX86CondFpuUnordered , kX86CondFpuNotUnordered, 0x12, 0x13
};

const uint32_t _x86CondToCmovcc[20] = CC_TO_INST(kX86InstIdCmov);
const uint32_t _x86CondToJcc   [20] = CC_TO_INST(kX86InstIdJ   );
const uint32_t _x86CondToSetcc [20] = CC_TO_INST(kX86InstIdSet );

#undef CC_TO_INST

// ============================================================================
// [asmjit::X86Util]
// ============================================================================

#if !defined(ASMJIT_DISABLE_TEXT)
//! \internal
//!
//! Compare two instruction names.
//!
//! `a` is null terminated instruction name from `_x86InstNameData[]` table.
//! `b` is non-null terminated instruction name passed to `getInstIdByName()`.
static ASMJIT_INLINE int X86Util_cmpInstName(const char* a, const char* b, size_t len) noexcept {
  for (size_t i = 0; i < len; i++) {
    int c = static_cast<int>(static_cast<uint8_t>(a[i])) -
            static_cast<int>(static_cast<uint8_t>(b[i])) ;
    if (c != 0)
      return c;
  }

  return static_cast<int>(a[len]);
}

uint32_t X86Util::getInstIdByName(const char* name, size_t len) noexcept {
  if (name == nullptr)
    return kInstIdNone;

  if (len == kInvalidIndex)
    len = ::strlen(name);

  if (len == 0)
    return kInstIdNone;

  uint32_t prefix = name[0] - kX86InstAlphaIndexFirst;
  if (prefix > kX86InstAlphaIndexLast - kX86InstAlphaIndexFirst)
    return kInstIdNone;

  uint32_t index = _x86InstAlphaIndex[prefix];
  if (index == kX86InstAlphaIndexInvalid)
    return kInstIdNone;

  const uint16_t* base = _x86InstNameIndex + index;
  const uint16_t* end = _x86InstNameIndex + _kX86InstIdCount;

  // Handle instructions starting with 'j' specially. `jcc` instruction breaks
  // the sorting, because of the suffixes (it's considered as one instruction),
  // so basically `jecxz` and `jmp` are stored after all `jcc` instructions.
  bool linearSearch = prefix == ('j' - kX86InstAlphaIndexFirst);

  while (++prefix <= kX86InstAlphaIndexLast - kX86InstAlphaIndexFirst) {
    index = _x86InstAlphaIndex[prefix];
    if (index == kX86InstAlphaIndexInvalid)
      continue;
    end = _x86InstNameIndex + index;
    break;
  }

  if (linearSearch) {
    while (base != end) {
      if (X86Util_cmpInstName(_x86InstNameData + base[0], name, len) == 0)
        return static_cast<uint32_t>((size_t)(base - _x86InstNameIndex));
      base++;
    }
  }
  else {
    for (size_t lim = (size_t)(end - base); lim != 0; lim >>= 1) {
      const uint16_t* cur = base + (lim >> 1);
      int result = X86Util_cmpInstName(_x86InstNameData + cur[0], name, len);

      if (result < 0) {
        base = cur + 1;
        lim--;
        continue;
      }

      if (result > 0)
        continue;

      return static_cast<uint32_t>((size_t)(cur - _x86InstNameIndex));
    }
  }

  return kInstIdNone;
}

const char* X86Util::getInstNameById(uint32_t id) noexcept {
  if (id >= _kX86InstIdCount)
    return nullptr;
  return _x86InstNameData + _x86InstNameIndex[id];
}
#endif // ASMJIT_DISABLE_TEXT

// ============================================================================
// [asmjit::X86Util - Test]
// ============================================================================

#if defined(ASMJIT_TEST) && !defined(ASMJIT_DISABLE_TEXT)
UNIT(x86_inst_name) {
  // All known instructions should be matched.
  INFO("Matching all X86/X64 instructions.");
  for (uint32_t a = 0; a < _kX86InstIdCount; a++) {
    uint32_t b = X86Util::getInstIdByName(_x86InstNameData + _x86InstNameIndex[a]);
    EXPECT(a == b,
      "Should match existing instruction \"%s\" {id:%u} != \"%s\" {id:%u}.",
        _x86InstNameData + _x86InstNameIndex[a], a,
        _x86InstNameData + _x86InstNameIndex[b], b);
  }

  // Everything else should return `kInstIdNone`.
  INFO("Trying to look-up instructions that don't exist.");
  EXPECT(X86Util::getInstIdByName(nullptr) == kInstIdNone,
    "Should return kInstIdNone for `nullptr` input.");

  EXPECT(X86Util::getInstIdByName("") == kInstIdNone,
    "Should return kInstIdNone for empty string.");

  EXPECT(X86Util::getInstIdByName("_") == kInstIdNone,
    "Should return kInstIdNone for unknown instruction.");

  EXPECT(X86Util::getInstIdByName("123xyz") == kInstIdNone,
    "Should return kInstIdNone for unknown instruction.");
}
#endif // ASMJIT_TEST && !ASMJIT_DISABLE_TEXT

} // asmjit namespace

// [Api-End]
#include "../apiend.h"

// [Guard]
#endif // ASMJIT_BUILD_X86 || ASMJIT_BUILD_X64
